#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Processing tags applied to an agent resource, update the UCD environment globalTagListing
and local properties.

1. Extract the TAGS from a "server",
  - Create a new element [hostname]:[csv list of tags]
  - Create or add to UCD environment property {globalTagList} and populate with the new element

2. Check local tag rules

  1. Only accept a fixed list of tags
  2. There must be WAS or Liberty present for any group containing an APP
  3. If LDX present there must also be JTS
  4. Cannot have DB and any other TAG on the same server
  5. Cannot have IHS and any other TAG on the same server
  6. Cannot have WAS and Liberty on the same server
  7. Only one application server (WAS or Liberty) can exist on the same server

3. Remove the AM tag if the ALM release is less than 6.0.5

4. Create localTagList and localHostname
  - Add or update their entries in the env.properties file
  - Update localTags, APPS, WASRole entries for compatibility (to QD11)

Environment variables:
  - AH_WEB_URL - UCD server
  - AUTH_TOKEN - UCD server authorization token

Usage::

    process_local_resource_tags.py APPLICATION_ID ENVIRONMENT_ID COMPONENT_ID AGENT_ID

(see :func:`main`)
'''
import os
import sys

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))


from AutomationPrep.Tagging import tags_lib
from qdlib import qd, ucd_client, qd_topo, qd_props


def get_server_resource_parent_tags(client, component_id):
    '''
    get a csv list of application TAGS on the current server

    :param client: client to connect to UCD with
    :param component_id: unique id of the current UCD component 'object'
    :return: csv list of TAGS on server
    '''
    json_data = client.get_resource(component_id)
    qd.echo_debug('JSON resource is %s' % json_data)

    parent_value = json_data.get('parent')
    if not isinstance(parent_value, dict):
        qd.echo_error("JSON resource does not contain any [parent] information")
        qd.echo_error("JSON resource is %s" % json_data)
        qd.exit_(1)

    tags_value = parent_value.get('tags')
    if not isinstance(tags_value, list):
        qd.echo_error("JSON resource does not contain any [parent:tags] information")
        qd.echo_error("JSON resource is %s" % json_data)
        qd.exit_(1)

    # cleanup tag names
    tag_list = [t['name'].strip() for t in tags_value if t.get('name') and t['name'].strip()]
    qd.echo_info('tag_list [%s]' % ','.join(tag_list))
    return tag_list


def check_valid_am_tag(env_props, thost):
    '''
    Remove AM tag if present and release less then 605.

    :param env_props: object representing the content of the env.properties file
    :param thost: object representing the current server

    '''
    qd.echo_info("Check AM allowed rule")
    release_version = qd.get_jazz_release_version(env_props.get_values('appMediaRoot'))
    if int(release_version) < 605:
        qd.echo_info('AM tag is not supported in release [%s], it will be removed'
                     % release_version)
        thost.tags.remove(qd_topo.TAG_AM)
    else:
        qd.echo_info('AM tag is supported in release [%s]' % release_version)


# pylint: disable=too-many-branches
def check_tag_rules(thost):
    '''
    Check tag rules and exit with error if they are not met.

    :param thost: object representing the current server

    '''

    qd.echo_info('tagString [%s]' % ','.join(thost.tags))

    is_valid = True

    # 1. only accept a fixed list of tags
    # no longer accept "jrs"
    qd.echo_info("check valid tags rule")
    for tag in thost.tags:
        if tag in qd_topo.ALL_TAGS:
            qd.echo_info("-- " + tag + " in the valid list --> rule validated")
        else:
            qd.echo_error("-- " + tag + " not in valid list --> rule invalidated")
            is_valid = False

    # 2. there must be WAS or Liberty present for any group containing an APP
    # no longer accept "jrs"
    qd.echo_info("check WAS|Liberty and {app} rule")
    for tag in thost.tags:
        if tag in qd_topo.ALL_APP_TAGS:
            if any(appsvr in thost.tags for appsvr in qd_topo.ALL_APPSVR_TAGS):
                qd.echo_info("-- was|liberty and " + tag + " present --> rule validated")
            else:
                qd.echo_error("-- " + tag + " present but no was or liberty --> rule invalidated")
                is_valid = False
        else:
            qd.echo_info("-- " + tag + " not in applist --> no need to test")

    # 3. if LDX present there must also be JTS
    qd.echo_info("check LDX rule")
    if qd_topo.TAG_LDX in thost.tags:
        if qd_topo.TAG_JTS in thost.tags:
            qd.echo_info("-- ldx and jts present --> rule validated")
        else:
            qd.echo_error("-- ldx present but no jts --> rule invalidated")
            is_valid = False
    else:
        qd.echo_info("-- ldx not present --> no need to test")

    def check_no_other_tags(atag):
        '''
        Check that the TAG {atag} does not already exist on the server

        :param atag: TAG to check for
        :return: (boolean) True if TAG does not exist; False if it did
        '''

        result = True  # valid
        if atag in thost.tags:
            is_local_valid = True
            for tag in thost.tags:
                if tag != atag:
                    qd.echo_error("-- " + tag + " and " + atag +
                                  " cannot be on the same server --> rule invalidated")
                    result = False
                    is_local_valid = False
            if is_local_valid:
                qd.echo_info("-- only %s present --> rule validated" % atag)
        else:
            qd.echo_info("-- %s not present --> no need to test" % atag)
        return result

    # 4. unless LIBERTY is present cannot have DB and any other TAG on the same server
    # even then a further check is required in process_global_tags.py once the system
    # is better understood
    if qd_topo.TAG_DB in thost.tags:
        qd.echo_info("check DB database rules")
        if qd_topo.TAG_LIBERTY in thost.tags:
            qd.echo_info("delaying check until we know database vendor")
        else:
            if not check_no_other_tags(qd_topo.TAG_DB):
                is_valid = False

    # 5. cannot have IHS and any other TAG on the same server
    qd.echo_info("check IHS rule")
    if not check_no_other_tags(qd_topo.TAG_IHS):
        is_valid = False

    # 6. Cannot have WAS and Liberty on the same server
    qd.echo_info("check only one application server per server rule")
    if qd_topo.TAG_WAS in thost.tags and qd_topo.TAG_LIBERTY in thost.tags:
        qd.echo_error("-- WAS and LIBERTY on the same server --> rule invalidated")
        is_valid = False
    else:
        qd.echo_info("-- No WAS or Liberty on same server --> rule validated")

    # 7. Only one application server (WAS or Liberty) can exist on the same server
    qd.echo_info("check only one application server Tag per server")
    result = [tag for tag in thost.tags if tag in qd_topo.ALL_APPSVR_TAGS]
    if len(result) > 1:
        qd.echo_error("-- APPSERVER and WAS or LIBERTY tag on same server --> rule invalidated")
        is_valid = False
    else:
        qd.echo_info("-- Only one AppServer Tag --> rule validated")
    # confirm all is valid or exit
    if not is_valid:
        # one or more rules invalid
        sys.exit(1)
    else:
        qd.echo_info("TAG list valid")


# pylint: disable=too-many-locals
# pylint: disable=too-many-statements
def main(argv):
    '''
    Main function; see module documentation.

    :param APPLICATION_ID: unique id of the current UCD application 'object'
    :param ENVIRONMENT_ID: unique id of the current UCD environment 'object'
    :param COMPONENT_ID: unique id of the current UCD component 'object'
    :param AGENT_ID: unique id of the current UCD agent 'object'
    '''

    qd.echo_info("====== process inputs =========")
    arg_parsed = qd.process_inputs(argv, 'APPLICATION_ID,ENVIRONMENT_ID,COMPONENT_ID,AGENT_ID')

    # pylint error unbalanced-tuple-unpacking on this line because no guarantee that arg_parsed,
    # i.e., qd.processInputs() always returns 4 items.  Probably should refactor.  For now, ignore.
    # pylint: disable=unbalanced-tuple-unpacking
    application_id, environment_id, component_id, agent_id = arg_parsed

    qd.echo_info("====== determine operating system =========")
    qd.echo_info('running on [%s] operating system' % qd.os_switch('linux', 'windows'))

    qd.echo_info("====== get environment properties =========")
    qd.echo_debug('LOCALSERVERPROPERTYFILE [%s]' % qd_props.DEFAULT_ENV_PROPS_LOCATION)
    env_props = qd_props.get_default_env_props()

    # get UCD info from env variables
    client = ucd_client.UDClient.create_from_env2('AH_WEB_URL', 'AUTH_TOKEN')
    qd.echo_info("====== check the UCD agent <--> server connection =========")
    tags_lib.test_ucd_agent_connection(client, agent_id, fail_on_error=False)

    qd.echo_info("====== get_server_resource_parent_tags using the ucd_client api ======")
    resource_tags = get_server_resource_parent_tags(client, component_id)

    qd.echo_info("====== get Server Hostname ======")
    hostname = qd.get_host_name()

    # create a host instance
    thost = qd_topo.QDTopoHost(hostname, resource_tags)

    qd.echo_info("====== get Application Server ======")

    app_server_tag = qd_topo.TAG_APPSERVER
    app_server = env_props.get_values(qd_props.PN_LOCAL_APP_SERVER)
    assert app_server in ['was', 'liberty'], '[%s] appServer not supported' % app_server

    qd.echo_info('appserver [%s]' % app_server)
    for tag in thost.tags:
        if tag == app_server_tag:
            if app_server != qd_topo.TAG_CUSTOM_APPSERVER:
                qd.echo_info('change ' + app_server_tag + ' tag to ' + app_server)
                thost.tags.remove(app_server_tag)
                thost.tags.append(app_server)
            else:
                qd.echo_info('[%s] tags selected' % app_server)

    qd.echo_info("====== check_tag_rules ======")
    check_tag_rules(thost)  # check local tag rules

    if qd_topo.TAG_AM in thost.tags:
        qd.echo_info("====== check_valid_am_tag ======")
        check_valid_am_tag(env_props, thost)

    qd.echo_info("====== getAppsFromTags ======")
    apps_list = thost.get_local_apps()
    qd.echo_info("apps_list [%s]" % apps_list)

    qd.echo_info("====== getWasRoleFromTags ======")
    was_role = thost.get_was_role()
    qd.echo_info("was_role [%s]" % was_role)

    # update UCD environment
    qd.echo_info("====== updateUCDEnvironmentProperty using the ucd_client api ======")
    tags_lib.update_ucd_tag_listing_host(env_props, client, application_id,
                                         environment_id, thost)

    if qd_topo.TAG_APPSERVER.upper() in resource_tags:
        resource_tags.remove(qd_topo.TAG_APPSERVER.upper())
        resource_tags.append(was_role.upper())

    # update local server environment
    qd.echo_info("====== update local server property file (localHostname=..) ======")
    env_props.set_value("localHostname", hostname)
    qd.echo_info("====== update local server property file (localTags=..) ======")
    env_props.set_value("localTags", resource_tags)
    qd.echo_info("====== update local server property file (APPS=..) ======")
    env_props.set_value("APPS", ','.join(apps_list))

    if was_role:
        qd.echo_info("====== update local server property file (was_role=..) ======")
        env_props.set_value("WASRole", was_role)

    if was_role == 'was' or was_role == 'liberty':
        qd.echo_info("====== checking file [%s] exists on %s server ======"
                     % ("Custom_LDAPSecurity.properties", was_role.upper()))
        ldap_props = qd_props.get_custom_ldap_props(None)
        assert ldap_props
        qd.echo_info("file exists")


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main(sys.argv)
    exit(0)
