#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Get WAS node name, and then update the UCD environment globalTagListing if necessary.

Environment variables:

- AH_WEB_URL - UCD server
- AUTH_TOKEN - UCD server authorization token

Usage::

    ProcessWASNodeName.py APPLICATION_ID ENVIRONMENT_ID COMPONENT_ID AGENT_ID

'''
import os
import sys

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from AutomationPrep.Tagging import tags_lib
from qdlib import qd, ucd_client, qd_props, install_util, qd_topo, resource


# pylint: disable=too-many-locals
def main(argv):
    '''Main function; see module documentation.'''
    qd.echo_info("====== process inputs =========")
    arg_parsed = qd.process_inputs(argv, 'APPLICATION_ID,ENVIRONMENT_ID,COMPONENT_ID,AGENT_ID')

    # pylint error unbalanced-tuple-unpacking on this line because no guarantee that arg_parsed,
    # i.e., qd.processInputs() always returns 4 items.  Probably should refactor.  For now, ignore.
    # pylint: disable=unbalanced-tuple-unpacking
    application_id, environment_id, _, agent_id = arg_parsed

    qd.echo_info("====== determine operating system =========")
    qd.echo_info('running on [%s] operating system' % qd.os_switch('linux', 'windows'))

    qd.echo_info("====== get environment properties =========")
    qd.echo_debug('LOCALSERVERPROPERTYFILE [%s]' % qd_props.DEFAULT_ENV_PROPS_LOCATION)
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)

    ##
    # get WAS nodeName
    ##
    all_tag_listing = env_props.get_values(qd_props.UCD_PN_GLOBAL_TAG_LISTING)
    topo = qd_topo.QDTopo(all_tag_listing)
    local_host = env_props.get_values('localHostname', localHostname=qd.get_host_name())
    local_server_role = topo.get_host(local_host).get_was_role()

    if local_server_role == 'was':

        qd.echo_info('extract WAS Profile NODENAME')

        was_home = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                                qd_props.DEFAULT_REL_WAS_PROFILE)
        os_user_name, os_user_password = qd_props.get_os_users_creds(user_props)
        jython_script = resource.get_resource('AutomationPrep/Tagging/getNodeName.jy')
        cmd = [qd.path_join(was_home, 'bin', qd.os_switch('wsadmin.sh', 'wsadmin.bat')),
               '-lang', 'jython', '-f', jython_script]
        py_out = install_util.su_call_if_necessary('-', (os_user_name, os_user_password), cmd,
                                                   call_method=qd.check_output_strip)

        qd.echo_info('python output : [%s]' % py_out)

        # strip out the node_name value
        # split at the first "=" for cut -d'=' -f 1 --complement
        _, node_name = py_out.split('=', 1)
    elif local_server_role == 'liberty':
        qd.echo_info('Get Liberty Profile NODENAME')
        node_name, _ = str(local_host).split('.', 1)

    qd.echo_info('node_name=%s' % node_name)
    qd.echo_info('')
    qd.echo_info('update UCD environment with [node_name]')

    ##
    # update the UCD environment
    ##
    # get UCD info from env variables
    client = ucd_client.UDClient.create_from_env2('AH_WEB_URL', 'AUTH_TOKEN')

    qd.echo_info("====== check the UCD agent <--> server connection =========")
    tags_lib.test_ucd_agent_connection(client, agent_id, fail_on_error=False)

    qd.echo_info("====== get Server Hostname ======")
    hostname = qd.get_host_name()

    # Update the existing globalTagListing with the node_name
    qd.echo_info("====== updateUCDEnvironmentProperty using the ucd_client api ======")
    tags_lib.update_ucd_tag_listing_nodename(env_props, client, application_id,
                                             environment_id, hostname, node_name)


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    if qd_props.check_app_server():
        main(sys.argv)
    exit(0)
