#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************'''
'''
Methods used to set / modify the UCD environment property {globalTagListing}
'''
import os
import sys

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from qdlib import qd_topo, qd_props, qd


# pylint: disable=too-many-arguments
def update_ucd_tag_listing(env_props, client, app, env, update_callback):
    '''
    Update UCD environment property {globalTagListing}, also updates the local {env_props} copy

    :param env_props: environment properties
    :param client: client to connect to UCD with
    :param app: unique id of the current UCD application 'object'
    :param env: unique id of the current UCD environment 'object'
    :param update_callback: callback method defined in the calling method
    '''
    prop = qd_props.UCD_PN_GLOBAL_TAG_LISTING
    # load existing
    existing_value = client.get_env_prop(app, env, prop) or ''
    topo = qd_topo.QDTopo()
    topo.load_tag_listing_string(existing_value)

    # do update by calling the given call-back function
    update_callback(topo)

    # write the updated topo back to UCD
    new_value = topo.to_tag_listing_string()
    if existing_value != new_value:
        # no need to examine again
        update_ucd_env_property(client, app, env, prop, new_value, force=True)

    # update env_props
    if env_props:
        env_props.set_value(prop, new_value)


def update_ucd_tag_listing_host(env_props, client, app, env, thost):
    '''
    Update the UCD environment property {globalTagListing} hostname entry

    :param env_props: environment properties
    :param client: client to connect to UCD with
    :param app: unique id of the current UCD application 'object'
    :param env: unique id of the current UCD environment 'object'
    :param thost: object representing the current server
    '''
    def update_thost(topo):
        '''Update host'''
        topo.add_or_update_host(thost)
    # update with the method
    update_ucd_tag_listing(env_props, client, app, env, update_thost)


def update_ucd_tag_listing_nodename(env_props, client, app, env, hostname, node_name):
    '''
    Update UCD environment property {globalTagListing} node name entry

    :param env_props: environment properties
    :param client: client to connect to UCD with
    :param app: unique id of the current UCD application 'object'
    :param env: unique id of the current UCD environment 'object'
    :param hostname: fully qualified hostname of the current server
    :param node_name: WAS node name
    '''
    # callback from _update_ucd_tag_listing
    def update_node_name(topo):
        '''Update host was node name to host from topology'''
        host = topo.get_host(hostname)
        assert host  # hostname must be resolved in the topo
        host.was_node_name = node_name
    # update with the method
    update_ucd_tag_listing(env_props, client, app, env, update_node_name)


def test_ucd_agent_connection(client, agent_id, fail_on_error=False):
    '''
    Test UCD agent connection and exit with error if test fails.
    Can be configured to just log a WARNING if the test fails.

    :param client: client to connect to UCD with
    :param agent_id: unique id of the current UCD agent 'object'
    :param fail_on_error: True - will exit on error; False - will continue
    '''

    # Disable pylint bare-except to allow catching everything.
    # pylint: disable=bare-except
    try:
        qd.echo_info("test agent connection")
        client.test_agent(agent_id)
        qd.echo_info("agent connected")
    except:
        # potentially ignore the exception (non fatal test failure!)
        if fail_on_error:
            raise
        qd.echo_warning('agent connection test failed. skipping...')


def update_ucd_env_property(client, application_id, environment_id,
                            prop, value, force=False):
    '''
    Set a UCD environment property value.
    Can be configured to only update an existing property if the value is changing.

    :param client: client to connect to UCD with
    :param application_id: unique id of the current UCD application 'object'
    :param environment_id: unique id of the current UCD environment 'object'
    :param prop: property to update
    :param value: new property value
    :param force: True - update even if value is not changing;
        False - update only if value is different
    :return: "nothing"
    '''
    qd.echo_debug('Setting UCD environment property [%s] to [%s]...' % (prop, value))
    if not force:
        current_value = client.get_env_prop(application_id, environment_id, prop)
        if current_value == value:
            qd.echo_debug('Skipped as it is already set.')
            return
    client.set_env_prop(application_id, environment_id, prop, value)
