#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Deploy Jazz applications to WebSphere application server. See :func:`main` function.
'''
import os
import sys

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..', '../../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from ConfigureWAS.WASCommon import single_was_start_stop
from qdlib import qd, qd_props, install_util, resource


def main():
    '''
    Deploy Jazz applications to WebSphere application server.

    1. Update LQE-specific JVM properties
    2. Call install_applications.jy to install each application into WebSphere
    3. Restart the WebSphere application server process
    '''
    # pylint: disable=too-many-locals
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)
    os_user_name, os_user_password = qd_props.get_os_users_creds(user_props)

    clm_property_file = resource.get_escaped_resource('ConfigureWAS/was.properties')
    ldap_property_file = resource.get_escaped_resource(qd_props.DEFAULT_LDAP_REL_PATH)

    apps = getattr(env_props.get_values(qd_props.PN_APPS), 'split')(',')

    for app in apps:
        timer = qd.start_clock('Installing App [%s]' % app)
        if app == 'lqe':
            qd.echo_info("Add LQE specific JVM properties to WAS appserver.....")
            lqe_was_property = resource.get_resource('ConfigureAppServer/lqewas.properties')
            escaped_lqe_was_property = resource.get_escaped_resource(
                'ConfigureAppServer/lqewas.properties')
            # we need convert "\\" to "/" in the lqewas.properties,
            # since the path passed into WASadmin need to be c:/IBMQD/JazzTeamServer
            # str(jazz_home) is required to go from the tuple to a string!
            jazz_home = str(env_props.get_values(qd_props.PN_CLM_HOME_PREP))
            replacement_list = [('_CLM_HOME_', jazz_home.replace('\\', '/'))]
            qd.sed_in_place(lqe_was_property, replacement_list)

            cmd = [qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                   qd_props.DEFAULT_REL_WAS_ROOT, 'bin', qd_props.DEFAULT_WSADMIN_EXE), '-lang',
                   'jython', '-f', resource.get_resource('ConfigureAppServer/update_jvm_lqe.jy'),
                   escaped_lqe_was_property]

            install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                              cmd, call_method=qd.check_call)
            qd.echo_info("LQE specific JVM properties update succeeded")

        # Determine product release
        if env_props.get_values('mediaPropertiesType') == 'pipeline':
            release_version = 'pipeline'
        else:
            release_version = qd.get_jazz_release_version(env_props.get_values('appMediaRoot'))

        # Installing each application to WAS
        qd.echo_info("Install Application [%s] to WAS....." % app)
        cmd = [qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
               qd_props.DEFAULT_REL_WAS_ROOT, 'bin', qd_props.DEFAULT_WSADMIN_EXE), '-lang',
               'jython', '-f', resource.get_resource('ConfigureAppServer/install_application.jy'),
               app, clm_property_file, ldap_property_file, release_version]

        install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                          cmd, call_method=qd.check_call)
        qd.echo_info("Successfully installed application [%s] to WAS application server" % app)
        qd.stop_clock(timer)

    # dynamic load the module
    single_was_start_stop.main(['restart'])


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    if qd_props.check_app_server():
        main()
    exit(0)
