#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Configure liberty Evaluation Topology with 2 Instances
'''
import sys
import os
import time
import subprocess
import shutil
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import Comment, tostring
from xml.dom import minidom
import socket

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from qdlib import qd, qd_props, install_util

LIBERTY_INSTANCE = qd.os_switch('liberty.serverx', 'liberty.serverx.bat')
LIBERTY_SERVER = "liberty/wlp/bin/server"
SERVER_STARTUP = qd.os_switch('server.startupx', 'server.startupx.bat')
SERVER_SHUTDOWN = qd.os_switch('server.shutdownx', 'server.shutdownx.bat')
''' Stop and Start Liberty command '''

JAVA_OPTS = qd.os_switch('JAVA_OPTS="$JAVA_OPTS ', 'set JAVA_OPTS=%JAVA_OPTS% ')
''' JAVA Options for server start script'''

REPOTOOLS_SCRIPT = qd.os_switch('repotools-jts.sh', 'repotools-jts.bat')
''' JTS Repotools script '''

# pylint: disable=line-too-long
TIME_OUT = ('com.ibm.team.repository.xserver.timeout=1800',
            qd.os_switch('DEFINE="$DEFINE -Dcom.ibm.team.repository.common.transport.TeamServerConfiguration.connectTimeout=1800000"',
                         'set DEFINE=%DEFINE% "-Dcom.ibm.team.repository.common.transport.TeamServerConfiguration.connectTimeout=1800000"'),
            qd.os_switch('DEFINE="$DEFINE -Dcom.ibm.team.repository.common.transport.TeamServerConfiguration.socketTimeout=1800000"',
                         'set DEFINE=%DEFINE% "-Dcom.ibm.team.repository.common.transport.TeamServerConfiguration.socketTimeout=1800000"'))
''' JTS Time out properties '''
# pylint: enable=line-too-long

DETACHED_PROCESS = 0x00000008
''' Windows process creation Flag '''

SERVER_NAME = qd.os_switch('export CLM_LIBERTY_SERVER_NAME=clm', 'set CLM_LIBERTY_SERVER_NAME=clm')
''' Liberty Server Name property '''

APP_DIR = qd.os_switch('export CLM_LIBERTY_SERVER_TEMPLATE=',
                       'set CLM_LIBERTY_SERVER_TEMPLATE=')
''' Liberty Application directory '''

LTPA = {'keysFileName': 'resources/security/jtsLTPAkey', 'keysPassword': 'WebAS',
        'expiration': '120'}
''' Liberty SSO configuration '''

SERVER_APPS = ('lqe.war', 'dcc.war', 'relm.war', 'rs.war', 'lqe.war.zip', 'dcc.war.zip',
               'relm.war.zip', 'rs.war.zip')
''' Liberty Server Apps '''

SERVER_NUM = 2


def get_jts_dir():
    '''
    Get Liberty server directory

    :return: Returns Liberty server directory
    '''
    env_props = qd_props.get_default_env_props()
    return qd.path_join(env_props.get_values(qd_props.PN_CLM_HOME_PREP), 'server')


JTS_DIR = get_jts_dir()
''' Jazz Team Server Path '''


def start_liberty(server_startup):
    '''
    This module helps start Liberty Server and create the server

    :param server_startup: Name of the startup file of the Liberty Instance
    '''
    for server in range(SERVER_NUM):
        if str(server) in server_startup:
            clm_server = 'clm' + str(server)
    if not verify_server(clm_server):
        qd.echo_info('==== Start Liberty Server %s ====' % clm_server)
        env_props = qd_props.get_default_env_props()
        user_props = qd_props.get_user_props(env_props)
        os_user_name, os_user_password = qd_props.get_os_users_creds(user_props)
        service_cmd = qd.path_join(JTS_DIR, server_startup)
        if qd.is_windows():
            subprocess.Popen(service_cmd, stdin=None, stdout=None,
                             stderr=None, close_fds=True, creationflags=DETACHED_PROCESS)
            timeout = time.time() + 60 * 5
            while True:
                return_code = verify_server(clm_server)
                if return_code:
                    break
                if time.time() > timeout:
                    return_code = 0
                    break
                time.sleep(5)
        elif qd.is_linux():
            return_code = install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                                            service_cmd)
            return_code = not bool(return_code)
        time.sleep(30)
        if return_code:
            qd.echo_info('Liberty Server %s Started' % clm_server)
        else:
            qd.exit_with_err_msg('Failed to start Liberty Server %s' % clm_server)
    else:
        qd.echo_info('==== Liberty Server %s Running ====' % clm_server)


def stop_liberty(clm_server):
    '''
    This module helps stop Liberty Server

    :param clm_server: Liberty instance server name
    '''
    qd.echo_info('==== Stop Liberty Server %s ====' % clm_server)
    server_tool = qd.path_join(JTS_DIR, LIBERTY_SERVER)
    return_code = qd.call(server_tool + ' stop ' + clm_server, shell=True)
    time.sleep(30)
    timeout = time.time() + 60 * 5
    while True:
        return_code = verify_server(clm_server)
        if return_code:
            stop_liberty(clm_server)
        else:
            break
        if time.time() > timeout:
            return_code = 0
            break
        time.sleep(5)
    return_code = not bool(return_code)
    if return_code:
        qd.echo_info('Liberty Server %s Stopped' % clm_server)
    else:
        qd.echo_error('Failed to stop Liberty Server %s' % clm_server)


def verify_server(server):
    '''
    This module helps check server status

    :param server: Liberty instance Server number
    :return: True if Liberty Server is running else return False
    '''
    server_tool = qd.path_join(JTS_DIR, LIBERTY_SERVER)
    return_code = qd.call(server_tool + ' status ' + server, shell=True)
    return not bool(return_code)


def create_liberty_server(server_startup):
    '''
    This module helps create Liberty Server

    :param server_startup: Name of the startup file of the Liberty Instance
    '''
    qd.echo_info('==== Create Liberty Server %s ====' % server_startup)
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)
    os_user_name, os_user_password, _ = qd_props.get_os_users(user_props)
    service_cmd = qd.path_join(JTS_DIR, server_startup + ' -create')
    if qd.is_windows():
        if not install_util.need_user_inst(os_user_name):
            return_code = qd.call(service_cmd)
        else:
            return_code = install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                                            service_cmd)
    elif qd.is_linux():
        return_code = install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                                        service_cmd)
    else:
        qd.exit_with_err_msg('Unsupported OS')

    if return_code == 0:
        qd.echo_info('Liberty Server %s Created' % server_startup)
    else:
        qd.exit_with_err_msg('Failed to create Liberty Server %s' % server_startup)


def config_liberty_server():
    '''
    This module helps configure liberty server so server utility
    handles the server actions
    '''
    qd.echo_info('==== Configure liberty Server Utility ====')
    servers_path = qd.path_join(JTS_DIR, 'liberty/wlp/usr/servers')
    clm_server = qd.path_join(JTS_DIR, 'liberty/servers')
    cmd = qd.os_switch('ln -s ', 'mklink /D ')
    cmd = qd.os_switch(cmd + clm_server + ' ' + servers_path, cmd + servers_path + ' ' + clm_server)
    if not qd.check_call(cmd, shell=True):
        qd.echo_info('Liberty Server Utility Configured correctly')
    else:
        qd.exit_with_err_msg('Fail to configure Liberty Server Utility')


def secure_usr_pass(os_user_password):
    '''
    Run The securityUtility
    command supports plain text encryption and SSL certificate creation for Liberty.

    :param os_user_password: contains the decoded user password
    :return: Return the encrypted user password
    '''
    security_tool = qd.path_join(JTS_DIR, "liberty/wlp/bin/securityUtility")
    cmd_args = 'encode --encoding=aes'
    cmd = security_tool + ' ' + cmd_args + ' ' + os_user_password
    return qd.check_output(cmd, shell=True)


def insert_to_file(filename, new_line, line_search=None, action=False):
    '''
    Insert or replace a line into a file

    :param filename: file to modify
    :param new_line: new line to add or replace to the file
    :param line_search: Line to search if set to None will add to the EOF
    :param action: Defines either its a replace or an insert
                   If set to True insert new line after line to search
                   If set to False replaces the line to search
    '''
    if not os.path.isfile(filename):
        qd.exit_with_err_msg("File doesn't exist")
    else:
        new_file = ''
        with open(filename, 'r') as r_file:
            for line in r_file:
                if line_search is not None:
                    if line_search in line:
                        if action:
                            new_file += new_line + '\n'
                        else:
                            line = new_line
                new_file += line
            if line_search is None:
                new_file += new_line + '\n'
        with open(filename, 'w') as w_file:
            w_file.write(new_file)


def remove_apps(server):
    '''
    Comment applications that were not installed from the application.xml

    :param server: Liberty instance Server number
    '''
    applist = []
    app_dir = qd.path_join(JTS_DIR, 'liberty/servers/' + server + '/apps')
    for dir_file in os.listdir(app_dir):
        if dir_file.endswith('.war'):
            file_name = dir_file.split(".", 1)
            if file_name[0] != 'converter':
                applist.append(file_name[0])
    app_xml = qd.path_join(JTS_DIR, 'liberty/servers/' + server + '/conf/application.xml')
    if not os.path.isfile(app_xml):
        qd.exit_with_err_msg('No application.xml file exists')
    else:
        tree = ET.parse(app_xml, CommentParser())
        root = tree.getroot()
        index = 0
        for child in root:
            index += 1
            if 'Comment' in str(child.tag):
                continue
            app_id = child.get('id')
            if app_id not in applist:
                comment = Comment(tostring(child))
                root.insert(index, comment)
                root.remove(child)
        tree = ET.ElementTree(root)
        tree.write(app_xml, xml_declaration=False, method="xml", encoding="utf-8")


class CommentParser(ET.XMLTreeBuilder):
    '''
    Add parser to get the comments from the XML file
    '''
    def __init__(self):
        ET.XMLTreeBuilder.__init__(self)
        self._parser.CommentHandler = self.handle_comment

    def handle_comment(self, data):
        '''
        Module helper to handle comments of XML files
        '''
        self._target.start(ET.Comment, {})
        self._target.data(data)
        self._target.end(ET.Comment)


def pretty_print_xml(xml_file):
    '''
    Pretty print XML files
    May add some additional whitespaces

    :param xml_file: xml file to apply the pretty print
    '''
    if not os.path.isfile(xml_file):
        qd.exit_with_err_msg("XML file doesn't exist")
    else:
        tree = ET.parse(xml_file, CommentParser())
        root = tree.getroot()
    xml_str = minidom.parseString(ET.tostring(root)).toprettyxml()
    with open(xml_file, "w") as xml:
        xml.write(xml_str.encode('utf-8'))


def enable_sso(server):
    '''
    Enable Liberty SSO

    :param server: Liberty instance Server number
    '''
    server_xml = qd.path_join(JTS_DIR, 'liberty/servers/' + server + '/server.xml')
    if not os.path.isfile(server_xml):
        qd.exit_with_err_msg('Fail to configure SSO server.xml not found')
    else:
        tree = ET.parse(server_xml, CommentParser())
        root = tree.getroot()
        domain = socket.getfqdn().split(".", 1)
        ET.SubElement(root, 'webAppSecurity', dict(singleSignonEnabled="true"))
        ET.SubElement(root, 'webAppSecurity', dict(ssoDomainNames=domain[1]))
        ET.SubElement(root, 'ltpa', LTPA)
        tree = ET.ElementTree(root)
        tree.write(server_xml, xml_declaration=True,
                   encoding='utf-8', method="xml")
        pretty_print_xml(server_xml)


def server_port_config():
    '''
    Server Port Configuration
    '''
    server_xml = qd.path_join(JTS_DIR, 'liberty/servers/clm1/server.xml')
    if not os.path.isfile(server_xml):
        qd.exit_with_err_msg('Fail to configure Liberty Port on server.xml not found')
    else:
        tree = ET.parse(server_xml, CommentParser())
        root = tree.getroot()
        web_cont = root.find('httpEndpoint')
        web_cont.set('httpPort', "9081")
        web_cont.set('httpsPort', "9444")
        tree = ET.ElementTree(root)
        tree.write(server_xml, xml_declaration=True,
                   encoding='utf-8', method="xml")


def copy_ltpakey():
    '''
    Update jtsLTPAkey parameters and copy jtsLTPAkey to liberty instances
    '''
    qd.echo_info('==== Update jtsLTAPKey parameters ====')
    qd.move_file(qd.path_join(JTS_DIR, 'liberty/servers/clm0/resources/security/ltpa.keys'),
                 qd.path_join(JTS_DIR, 'liberty/servers/clm0/resources/security/jtsLTPAkey'))
    config_ltpakey(qd.path_join(JTS_DIR, 'liberty/servers/clm0/resources/security/jtsLTPAkey'))
    qd.chmod_777(qd.path_join(JTS_DIR, 'liberty/servers/clm0/resources/security/jtsLTPAkey'),
                 mode='a+r')
    qd.echo_info('Copy ltpa key')
    qd.copy_file_with_stat(qd.path_join(JTS_DIR,
                                        'liberty/servers/clm0/resources/security/jtsLTPAkey'),
                           qd.path_join(JTS_DIR, 'liberty/servers/clm1/resources/security'))
    if qd.is_linux():
        stats = os.stat(qd.path_join(JTS_DIR, 'liberty/servers/clm0/resources/security/jtsLTPAkey'))
        user = stats.st_uid
        group = stats.st_gid
        for server in range(SERVER_NUM):
            qd.chown(qd.path_join(JTS_DIR,
                                  'liberty/servers/clm' + str(server) +
                                  '/resources/security/jtsLTPAkey'),
                     user, group)


def config_ltpakey(ltpakey):
    '''
    Add host to ltpakey

    :param ltpakey: jtsLTPAkey filename
    '''
    hostname_arg = 'com.ibm.websphere.CreationHost='
    hostname = socket.getfqdn().split(".", 1)
    new_line = hostname_arg + hostname[0] + '\n'
    insert_to_file(ltpakey, new_line, line_search=hostname_arg, action=False)


def create_liberty_instances(instance, liberty_instance, server_startup):
    '''
    Modify all required files to create the liberty instances

    :param instance: Liberty instance number
    :param liberty_instance: Name of the Liberty instances
    :param server_startup: Name of the startup file of the Liberty Instance
    '''
    insert_to_file(qd.path_join(JTS_DIR, liberty_instance),
                   SERVER_NAME + str(instance), line_search=SERVER_NAME)
    qd.sed_in_place(qd.path_join(JTS_DIR, liberty_instance), [('logs', 'logsclm' + str(instance))])
    if qd.is_windows():
        # pylint: disable=anomalous-backslash-in-string
        qd.sed_in_place(qd.path_join(JTS_DIR, liberty_instance),
                        [("\\\\logsclm" + str(instance), '\logs')])
        # pylint: enable=anomalous-backslash-in-string
    elif qd.is_linux():
        stats = os.stat(qd.path_join(JTS_DIR, qd.os_switch('liberty.server', 'liberty.server.bat')))
        user = stats.st_uid
        group = stats.st_gid
        qd.sed_in_place(qd.path_join(JTS_DIR, liberty_instance), [('/logsclm' + str(instance),
                                                                   '/logs')])
        qd.chown(qd.path_join(JTS_DIR, liberty_instance), user, group)
        qd.chown(qd.path_join(JTS_DIR, server_startup), user, group)
        qd.chown(qd.path_join(JTS_DIR, SERVER_SHUTDOWN.replace('x', str(instance))), user, group)

    insert_to_file(qd.path_join(JTS_DIR, server_startup),
                   qd.os_switch('./' + liberty_instance + ' $ACTION;',
                                'call "%START_DIR%"\\' + liberty_instance + ' %ACTION%'),
                   line_search=qd.os_switch('liberty.server $ACTION;',
                                            'liberty.server.bat %ACTION%'))
    insert_to_file(qd.path_join(JTS_DIR, SERVER_SHUTDOWN.replace('x', str(instance))),
                   qd.os_switch('./' + liberty_instance + ' stop;',
                                'call "%START_DIR%"\\' + liberty_instance + ' stop'),
                   line_search=qd.os_switch('liberty.server stop;',
                                            'liberty.server.bat stop'))


def update_instance_app(liberty_instance):
    '''
    Update required applications for every liberty instance
    Liberty Instance 1 JTS, CCM, RM, QM, GC, LDX, DM
    Liberty Instance 2 LQE, RELM, DCC

    :param liberty_instance: List of Liberty instances
    '''
    app_dir = qd.path_join(JTS_DIR, 'liberty/servers/clm1/apps')
    for files in os.listdir(app_dir):
        if files not in SERVER_APPS:
            if 'zip' in files:
                os.remove(qd.path_join(app_dir, files))
            else:
                shutil.rmtree(qd.path_join(app_dir, files))
    for app in SERVER_APPS:
        if 'zip' in app:
            os.remove(qd.path_join(JTS_DIR, 'liberty/servers/clm0/apps/' + app))
        else:
            shutil.rmtree(qd.path_join(JTS_DIR, 'liberty/servers/clm0/apps/' + app))
    for server in range(SERVER_NUM):
        remove_apps('clm' + str(server))
        insert_to_file(qd.path_join(JTS_DIR, liberty_instance[server]),
                       APP_DIR + qd.path_join('liberty/servers', 'clm' + str(server)),
                       line_search=APP_DIR)
        # add_liberty_usr('clm' + str(server))


def main():
    '''
    Configure Evaluation Topology Liberty with 2 Instances
    1. Create two Liberty Instances
    2. Set appropriate heap size for the Liberty Server
    3. Create Liberty servers that will run
    4. Increase Socket timeout
    5. Configure Liberty server to be accessed by service, start, stop tools
    6. Update applications that will manage each Liberty server
    7. Change Liberty Server Port
    8. Start and Stop Liberty Server for the first time
    9. Enable SSO for Liberty
    10. Start Liberty Servers
    '''
    liberty_instance = []
    server_startup = []
    for server in range(SERVER_NUM):
        liberty_instance.append(LIBERTY_INSTANCE.replace('x', str(server)))
        server_startup.append(SERVER_STARTUP.replace('x', str(server)))

    # Copy liberty.service, service.startup and service.shutdown
    qd.echo_info('==== Copy Required Files ====')
    for server in range(SERVER_NUM):
        shutil.copy2(qd.path_join(JTS_DIR, qd.os_switch('liberty.server',
                                                        'liberty.server.bat')),
                     qd.path_join(JTS_DIR, liberty_instance[server]))
        shutil.copy2(qd.path_join(JTS_DIR, qd.os_switch('server.startup',
                                                        'server.startup.bat')),
                     qd.path_join(JTS_DIR, server_startup[server]))
        shutil.copy2(qd.path_join(JTS_DIR, qd.os_switch('server.shutdown',
                                                        'server.shutdown.bat')),
                     qd.path_join(JTS_DIR, SERVER_SHUTDOWN.replace('x', str(server))))
        create_liberty_instances(server, liberty_instance[server], server_startup[server])

    # Adjust Memory
        qd.echo_info('==== Increase Memory to %s ====' % server_startup[server])
        heap_list = qd.get_heapsize_replacement()
        if heap_list > 0:
            heap_list.pop(0)
            new_list = heap_list[0][1].split(' ')
            insert_to_file(qd.path_join(JTS_DIR, server_startup[server]),
                           JAVA_OPTS + qd.os_switch(new_list[0] + '"\n', new_list[0] + '\n'),
                           line_search='-Xmx')
            insert_to_file(qd.path_join(JTS_DIR, server_startup[server]),
                           JAVA_OPTS + qd.os_switch(new_list[1] + '"\n', new_list[1] + '\n'),
                           line_search='-Xms')
            if len(heap_list) > 1:
                insert_to_file(qd.path_join(JTS_DIR, server_startup[server]),
                               JAVA_OPTS + qd.os_switch(heap_list[1][1] + '"\n',
                                                        heap_list[1][1] + '\n'),
                               line_search='-Xmn')

    # Create Server
        create_liberty_server(server_startup[server])
        time.sleep(60)

    # Increase timeout
    qd.echo_info('==== Augment Socket Timeout for Register Applications ====')
    insert_to_file(qd.path_join(JTS_DIR, REPOTOOLS_SCRIPT), TIME_OUT[2],
                   line_search='Patch service', action=True)
    insert_to_file(qd.path_join(JTS_DIR, REPOTOOLS_SCRIPT), TIME_OUT[1],
                   line_search='Patch service', action=True)
    insert_to_file(qd.path_join(JTS_DIR, 'conf/jts/teamserver.properties'), TIME_OUT[0])

    # Configure Liberty Server Path
    config_liberty_server()

    # Modify Applications on each server
    qd.echo_info('==== Update Liberty Applications ====')
    update_instance_app(liberty_instance)

    # Change Liberty Server Port
    qd.echo_info("==== Configure Liberty Server Port ====")
    server_port_config()

    # Start Liberty for the first time
    for server in range(SERVER_NUM):
        start_liberty(server_startup[server])
        time.sleep(60)
        stop_liberty('clm' + str(server))
        time.sleep(60)

    # Set up SSO
    qd.echo_info("==== Configure SSO for Liberty Profile ====")
    for server in range(SERVER_NUM):
        enable_sso('clm' + str(server))
    copy_ltpakey()

    # Start liberty
    for server in range(SERVER_NUM):
        start_liberty(server_startup[server])


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main()
    exit(0)
