#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Build Plugin-cf xml file based on the topology
'''
import os
import shutil
import sys

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..', '../../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from qdlib import qd, qd_props, resource


def get_node_name(env_props, count):
    '''Get the node name based on the number from env.properties, NODE1_NodeName

    :param env_props: environment properties
    :param count: node number
    :return: name of the node'''
    return env_props.get_values(''.join(['NODE', str(count), '_NodeName']))


def get_node_hostname(env_props, count):
    '''Get the host name based on the number from env.properties, NODE1_HostName

    :param env_props: environment properties
    :param count: node number
    :return: hostname of the node'''
    return env_props.get_values(''.join(['NODE', str(count), '_HostName']))


def update_node_section(plugin_template_dir, template_file, replace_list, out_file):
    '''
    Update the template section and write to an output file

    :param plugin_template_dir: plugin-cfg.xml template file directory
    :param template_file: the template file name
    :param replace_list: the list of replacement from the template_file
    :param out_file: output file that will have all changes
    '''
    temp_file = 'temp_update.xml'
    qd.duplicate_file(qd.path_join(plugin_template_dir, template_file), temp_file)
    qd.sed_in_place(qd.path_join(plugin_template_dir, temp_file), replace_list)
    concatenate_file(qd.path_join(plugin_template_dir, temp_file), out_file)


def concatenate_file(file_to_be_read, output_file):
    '''
    Add content of a file to the end of another file

    :param file_to_be_read: the source file will be added to another file
    :param output_file: append content to this file
    '''
    with open(output_file, 'a') as wfd:
        with open(file_to_be_read, 'rb') as input_fd:
            shutil.copyfileobj(input_fd, wfd)


def main():
    '''
    Main function; see module documentation.
    '''
    env_props = qd_props.get_default_env_props()

    output_file = resource.create_resource_path('ConfigureIHS/pluginConfigTemplate/plugin-cfg.xml')
    qd.create_file(output_file)
    plugin_root = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                               qd_props.DEFAULT_REL_PLUGINS_HOME)

    plugin_template_dir = os.path.dirname(output_file)

    qd.echo_info("Building header....")
    concatenate_file(qd.path_join(plugin_template_dir, 'plugin-cfg-head-template.xml'), output_file)
    total_nodes = int(env_props.get_values(qd_props.PN_CUSTOM_NODES))

    qd.echo_info("Building servercluster....")
    count = 1
    while count <= total_nodes:
        replacement_list = [('_WEBSERVER_Node_Name_', get_node_name(env_props, count))]
        replacement_list += [('_WEBSERVER_Host_Name_', get_node_hostname(env_props, count))]
        update_node_section(plugin_template_dir, 'plugin-cfg-ServerCluster-template.xml',
                            replacement_list, output_file)
        count = count + 1

    qd.echo_info("Building UriGroup....")
    count = 1
    while count <= total_nodes:
        replacement_list = [('_WEBSERVER_Node_Name_', get_node_name(env_props, count))]
        update_node_section(plugin_template_dir, 'plugin-cfg-UriGroup-head-template.xml',
                            replacement_list, output_file)
        node_apps = env_props.get_values(''.join(['NODE', str(count), '_AppList']))

        apps = getattr(node_apps, 'split')(',')
        additional_lines = []
        for app in apps:
            # pylint: disable=line-too-long
            additional_lines += [''.join(["      <Uri AffinityCookie=\"JSESSIONID\" AffinityURLIdentifier=\"jsessionid\" Name=\"/", app, "/*\"/>"])]
            if app == 'jts':
                additional_lines += ["      <Uri AffinityCookie=\"JSESSIONID\" AffinityURLIdentifier=\"jsessionid\" Name=\"/clmhelp/*\"/>"]
            # pylint: enable=line-too-long
        additional_lines += ["</UriGroup>"]
        qd.append_content_to_file(output_file, additional_lines)
        count = count + 1

    qd.echo_info("Building RouteServercluster....")
    count = 1
    while count <= total_nodes:
        replacement_list = [('_WEBSERVER_Node_Name_', get_node_name(env_props, count))]
        update_node_section(plugin_template_dir, 'plugin-cfg-RouteServerCluster-template.xml',
                            replacement_list, output_file)
        count = count + 1

    qd.echo_info("Building end section...")
    concatenate_file(qd.path_join(plugin_template_dir, 'plugin-cfg-tail-template.xml'), output_file)
    replacement_list = [('_WebServerMerge_', 'webservermerge')]
    replacement_list += [('_PLUGIN_ROOT_', qd.escape_path(plugin_root))]
    qd.sed_in_place(output_file, replacement_list)


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main()
    exit(0)
