#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Configure Liberty server such as create, start, stop and restart liberty service.
See :func:`main` function.
'''
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import Comment, tostring
import os
import sys
import time

# Due to circular import between liberty_config and liberty_service, imports need to stay relative,
# instead of the from X import Y format.
import liberty_start_stop  # pylint:disable=relative-import
import liberty_service  # pylint:disable=relative-import

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from qdlib import qd, qd_props, install_util, qd_topo

JTS_DIR = liberty_start_stop.get_jts_dir()
''' Jazz Team Server Path '''

SERVER_STARTUP = qd.os_switch('server.startup', 'server.startup.bat')
''' Start Liberty tool '''

JAVA_OPTS = qd.os_switch('JAVA_OPTS="$JAVA_OPTS ', 'set JAVA_OPTS=%JAVA_OPTS% ')
''' JAVA Options for server start script'''

REPOTOOLS_SCRIPT = qd.os_switch('repotools-jts.sh', 'repotools-jts.bat')
''' JTS Repotools script '''

# pylint:disable=line-too-long
TIME_OUT = ('com.ibm.team.repository.xserver.timeout=1800',
            qd.os_switch('DEFINE="$DEFINE -Dcom.ibm.team.repository.common.transport.TeamServerConfiguration.connectTimeout=1800000"',
                         'set DEFINE=%DEFINE% "-Dcom.ibm.team.repository.common.transport.TeamServerConfiguration.connectTimeout=1800000"'),
            qd.os_switch('DEFINE="$DEFINE -Dcom.ibm.team.repository.common.transport.TeamServerConfiguration.socketTimeout=1800000"',
                         'set DEFINE=%DEFINE% "-Dcom.ibm.team.repository.common.transport.TeamServerConfiguration.socketTimeout=1800000"'))
# pylint:enable=line-too-long
''' JTS Time out properties '''

RE_DB_DRIVER = {'oracle': 'ojdbc.*?.jar', 'sqlserver': 'sqljdbc.*?.jar'}
''' Regular Expression Database Driver '''


def create_liberty_server():
    '''
    This module helps create Liberty Server
    '''
    qd.echo_info('==== Create Liberty Server ====')
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)
    os_user_name, os_user_password, _ = qd_props.get_os_users(user_props)
    service_cmd = qd.path_join(JTS_DIR, SERVER_STARTUP + ' -create')
    if qd.is_windows():
        if not install_util.need_user_inst(os_user_name):
            return_code = qd.call(service_cmd)
        else:
            return_code = install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                                            service_cmd)
    elif qd.is_linux():
        return_code = install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                                        service_cmd)
    else:
        qd.exit_with_err_msg('Unsupported OS')

    if return_code == 0:
        qd.echo_info('Liberty Server Created')
    else:
        qd.exit_with_err_msg('Unsupported OS')


def config_liberty_server():
    '''
    This module helps configure liberty server so server utility
    handles the server actions
    '''
    qd.echo_info('==== Configure liberty Server Utility ====')
    servers_path = qd.path_join(JTS_DIR, 'liberty/wlp/usr/servers')
    clm_server = qd.path_join(JTS_DIR, 'liberty/servers')
    cmd = qd.os_switch('ln -s ', 'mklink /D ')
    cmd = qd.os_switch(cmd + clm_server + ' ' + servers_path, cmd + servers_path + ' ' + clm_server)
    if not qd.check_call(cmd, shell=True):
        qd.echo_info('Liberty Server Utility Configured correctly')
    else:
        qd.exit_with_err_msg('Fail to configure Liberty Server Utility')


def set_liberty_env_props(env_props):
    '''
    Set Liberty Config directory to env.properties

    :param env_props: Environmental properties
    '''
    liberty_config_dir = os.getcwdu()
    env_props.set_value('LIBERTY_CONF_DIR', liberty_config_dir)
    return env_props


def add_liberty_usr():
    '''
    Add user to Liberty server
    Need to add a user to the Liberty basic user registry
    in order to run JTS Setup using repotools-jts,
    if the user is not added it will show a connection error due to credentials
    '''
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)
    liberty_user_name, liberty_user_password = qd_props.get_was_admin_users(user_props)
    liberty_user_password = secure_usr_pass(liberty_user_password)
    qd.echo_info('==== Adding Liberty User ====')
    user_xml = qd.path_join(JTS_DIR, "liberty/servers/clm/conf/basicUserRegistry.xml")
    qd.check_file_exists(user_xml)

    tree = ET.parse(user_xml, liberty_start_stop.CommentParser())
    root = tree.getroot()
    user = {'name': liberty_user_name, 'password': liberty_user_password}
    child = root.find('basicRegistry')
    ET.SubElement(child, 'user', user)
    for group in child.findall('group'):
        if group.get('name') == 'JazzAdmins':
            ET.SubElement(group, 'member', name=liberty_user_name)
    tree = ET.ElementTree(root)
    tree.write(user_xml, xml_declaration=True,
               encoding='utf-8', method="xml")
    liberty_start_stop.pretty_print_xml(user_xml)


def secure_usr_pass(os_user_password):
    '''
    Run The securityUtility to encrypt the Liberty Password
    Command supports plain text encryption and SSL certificate creation for Liberty.

    :param os_user_password: contains the decoded user password
    :return: Return the encrypted user password
    '''
    security_tool = qd.path_join(JTS_DIR, "liberty/wlp/bin/securityUtility")
    cmd_args = 'encode --encoding=aes'
    cmd = security_tool + ' ' + cmd_args + ' ' + os_user_password
    return qd.check_output(cmd, shell=True)


def insert_to_file(filename, new_line, line_search=None, action=False):
    '''
    Insert or replace a line into a file

    :param filename: file to modify
    :param new_line: new line to add or replace to the file
    :param line_search: Line to search if set to None will add to the EOF
    :param action: Defines either its a replace or an insert
                   If set to True insert new line after line to search
                   If set to False replaces the line to search
    '''
    qd.check_file_exists(filename)
    new_file = ''
    with open(filename, 'r') as r_file:
        for line in r_file:
            if line_search is not None:
                if line_search in line:
                    if action:
                        new_file += new_line + '\n'
                    else:
                        line = new_line
            new_file += line
        if line_search is None:
            new_file += new_line + '\n'
    with open(filename, 'w') as w_file:
        w_file.write(new_file)


def remove_apps():
    '''
    Remove applications that were not installed from the application.xml
    '''
    applist = []
    app_dir = qd.path_join(JTS_DIR, 'liberty/servers/clm/apps')
    for dir_file in os.listdir(app_dir):
        if dir_file.endswith('.war'):
            file_name = dir_file.split(".", 1)
            if file_name[0] != 'converter':
                applist.append(file_name[0])
    app_xml = qd.path_join(JTS_DIR, "liberty/servers/clm/conf/application.xml")
    qd.check_file_exists(app_xml)

    tree = ET.parse(app_xml, liberty_start_stop.CommentParser())
    root = tree.getroot()
    index = 0
    for child in root:
        index += 1
        if 'Comment' in str(child.tag):
            continue
        app_id = child.get('id')
        if app_id not in applist:
            comment = Comment(tostring(child))
            root.insert(index, comment)
            root.remove(child)
    tree = ET.ElementTree(root)
    tree.write(app_xml, xml_declaration=False, method="xml", encoding="utf-8")


def lqe_jvm_options():
    '''
    Set LQE specific JVM properties
    '''
    java_op = '-Dlqe.config.location=file://%s/conf/lqe' % JTS_DIR
    insert_to_file(qd.path_join(JTS_DIR, SERVER_STARTUP), JAVA_OPTS + java_op +
                   qd.os_switch('"', '') + '\n', line_search='-Dderby.locks.deadlockTimeout',
                   action=True)


def set_heap_size():
    '''
    Based on total physical memory size to determine the appropriate heap size
    '''
    heap_list = qd.get_heapsize_replacement()
    if heap_list > 0:
        heap_list.pop(0)
        new_list = heap_list[0][1].split(' ')
        insert_to_file(qd.path_join(JTS_DIR, SERVER_STARTUP),
                       JAVA_OPTS + qd.os_switch(new_list[0] + '"\n', new_list[0] + '\n'),
                       line_search='-Xmx')
        insert_to_file(qd.path_join(JTS_DIR, SERVER_STARTUP),
                       JAVA_OPTS + qd.os_switch(new_list[1] + '"\n', new_list[1] + '\n'),
                       line_search='-Xms')
        if len(heap_list) > 1:
            insert_to_file(qd.path_join(JTS_DIR, SERVER_STARTUP),
                           JAVA_OPTS + qd.os_switch(heap_list[1][1] + '"\n',
                                                    heap_list[1][1] + '\n'),
                           line_search='-Xmn')


def set_jdbc_driver():
    '''
    Set correct JDBC driver to server_startup script
    '''
    env_props = qd_props.get_default_env_props()
    db_vendor = env_props.get_values(qd_props.PN_DB_VENDOR)
    if db_vendor == 'oracle' or db_vendor == 'sqlserver':
        db_driver = env_props.get_values(qd_props.PN_DB_DRIVER_FILE)
        qd.sed_in_place(qd.path_join(JTS_DIR, SERVER_STARTUP),
                        [(RE_DB_DRIVER[db_vendor], db_driver)])


def main():
    '''
    Configure Liberty Server

    1. Create Liberty server that will run
    2. Configure Liberty server to be accessed by service, start, stop tools
    3. Configure Liberty as service
    4. Set appropriate heap size for the Liberty Server
    5. Set JDBC Driver for Liberty
    6. Update applications that will manage each Liberty server
    7. If LQE present on server set specific JVM properties
    8. Start and Stop Liberty Server for the first time
    '''
    qd.echo_info('==== Configuring Liberty ====')
    env_props = qd_props.get_default_env_props()
    all_tag_listing = env_props.get_values(qd_props.UCD_PN_GLOBAL_TAG_LISTING)
    topo = qd_topo.QDTopo(all_tag_listing)
    if topo.get_user_registry_type_tmp() == 'LDAP':
        create_liberty_server()
        config_liberty_server()
        qd.echo_info('==== Configure Liberty as a Service ====')
        liberty_service.main()
        qd.echo_info('==== Increase Memory to server.startup ====')
        set_heap_size()
        qd.echo_info('==== Set JDBC Driver ====')
        set_jdbc_driver()
        qd.echo_info('==== Update Liberty Applications ====')
        remove_apps()
        env_props = set_liberty_env_props(env_props)
        add_liberty_usr()
        apps = env_props.get_values(qd_props.PN_APPS)
        if 'lqe' in apps:
            qd.echo_info('==== Set LQE specific JVM properties ====')
            lqe_jvm_options()
        liberty_start_stop.start_liberty()
        time.sleep(120)
        liberty_start_stop.stop_liberty()
    qd.echo_info('==== Finish Liberty Configuration ====')


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    if qd_props.check_app_server():
        main()
    exit(0)
