#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Module helps configure LDAP registry with Liberty
See :func:`main` function.
'''
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import Comment, tostring
import os
import sys

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from ConfigureLiberty import liberty_start_stop
from qdlib import qd, qd_props, resource


JTS_DIR = liberty_start_stop.get_jts_dir()
''' Jazz Team Server Path '''
LDAP_TYPE = {'Custom': 'customFilters'}
''' LDAP Type and filter dictionary'''

LDAP_REGISTRY = {'baseDN': 'jazz_ldap_baseDN', 'host': 'jazz_ldap_host', 'realm': 'jazz_userrealm',
                 'port': 'jazz_ldap_port', 'id': 'jazz_userrealm'}
''' LDAP Registry as mapped on LDAP properties file'''

LDAP_FILTER = {'groupFilter': 'jazz_ldap_groupFilter', 'groupIdMap': 'jazz_ldap_groupIdMap',
               'groupMemberIdMap': 'jazz_ldap_groupMemberIdMap',
               'userFilter': 'jazz_ldap_userFilter',
               'userIdMap': 'jazz_ldap_userIdMap'}
''' LDAP Filter as mapped on LDAP properties file'''

LDAP_DEFAULT_REG = {'ignoreCase': 'false', 'recursiveSearch': 'true', 'sslEnabled': 'false'}
''' LDAP Default values can change later'''


def ldap_config():
    '''
    This module Configures LDAP for Liberty
    Changes from basic to LDAP registry
    '''
    qd.echo_info('==== LDAP configuration for Liberty ====')
    clm_xml = qd.path_join(JTS_DIR, "liberty/servers/clm/server.xml")
    qd.check_file_exists(clm_xml)

    tree = ET.parse(clm_xml, liberty_start_stop.CommentParser())
    root = tree.getroot()
    index = 0
    for child in root:
        index += 1
        if child.tag == 'include':
            if child.get('location') == "conf/basicUserRegistry.xml":
                comment = Comment(tostring(child))
                root.insert(index, comment)
                root.remove(child)
        if child.text:
            uncomm_child = str(child.text)
            if 'conf/ldapUserRegistry.xml' in uncomm_child:
                uncomm_child = '<' + uncomm_child + '>'
                root.insert(index, ET.fromstring(uncomm_child))
                root.remove(child)
    tree = ET.ElementTree(root)
    tree.write(clm_xml, xml_declaration=False, method="xml", encoding="utf-8")
    liberty_start_stop.pretty_print_xml(clm_xml)


def create_ldap_xml():
    '''
    This module creates ldapUserRegistry.xml file with LDAP properties
    '''
    qd.echo_info('==== Create LDAP Registry ====')
    env_props = qd_props.get_default_env_props()
    ldap_props = qd_props.get_ldap_props(env_props)
    ldap_temp = resource.get_resource('ConfigureLiberty//ldapUserRegistry_template.xml')
    ldap_xml = qd.path_join(JTS_DIR, "liberty/servers/clm/conf/ldapUserRegistry.xml")
    qd.check_file_exists(ldap_temp)

    tree = ET.parse(ldap_temp, liberty_start_stop.CommentParser())
    root = tree.getroot()
    ldap_vendor = 'Custom'
    ldap_registry = get_ldap_registry(ldap_props, ldapType=ldap_vendor)
    ldap_filter = get_ldap_filter(ldap_props)
    ET.SubElement(root, 'ldapRegistry', ldap_registry)
    child = root.find('ldapRegistry')
    ET.SubElement(child, LDAP_TYPE[ldap_vendor], ldap_filter)
    tree = ET.ElementTree(root)
    tree.write(ldap_xml, xml_declaration=True,
               encoding='utf-8', method="xml")
    liberty_start_stop.pretty_print_xml(ldap_xml)
    create_app_xml(ldap_props)


def create_app_xml(ldap_props):
    '''
    This module maps LDAP groups and roles to the applications
    Liberty version of deploy applications

    :param ldap_props: LDAP Custom properties
    '''
    qd.echo_info('==== Add LDAP Parameters to application.xml ====')
    app_xml = qd.path_join(JTS_DIR, "liberty/servers/clm/conf/application.xml")
    qd.check_file_exists(app_xml)

    tree = ET.parse(app_xml, liberty_start_stop.CommentParser())
    root = tree.getroot()
    ldap_map = get_ldap_mapping(ldap_props)
    for app in root.findall('application'):
        app_bnd = app.find('application-bnd')
        if app_bnd is not None:
            for key, value in ldap_map.iteritems():
                security_role = app_bnd.findall(".//security-role[@name='%s']" % key)
                if security_role:
                    for security in security_role:
                        ET.SubElement(security, 'group', dict(name=value))
                else:
                    ET.SubElement(app_bnd, 'security-role', dict(name=key))
                    child = app_bnd.find(".//*[@name='%s']" % key)
                    ET.SubElement(child, 'group', dict(name=value))
    tree = ET.ElementTree(root)
    tree.write(app_xml, xml_declaration=True,
               encoding='utf-8', method="xml")
    liberty_start_stop.pretty_print_xml(app_xml)


def get_ldap_registry(ldap_props, **kwargs):
    '''
    Get LDAP Registry properties

    :param ldap_props: Get LDAP Custom properties
    :param ignoreCase: Default property needed to create creates ldapUserRegistry.xml
    :param recursiveSearch: Default property needed to create creates ldapUserRegistry.xml
    :param sslEnabled: Default property needed to create creates ldapUserRegistry.xml
    :param kwargs: Adds more LDAP properties e.g. ldapType contains ldap vendor
    :return: Return a dictionary containing LDAP registry information
    '''
    qd.echo_info('==== Get LDAP Registry from LDAP Properties ====')
    ldap_registry = dict(LDAP_REGISTRY)
    ldap_registry.update(LDAP_DEFAULT_REG)
    ldap_registry.update(kwargs)
    for key, value in LDAP_REGISTRY.iteritems():
        ldap_registry[key] = ldap_props.get_values(value)
    return ldap_registry


def get_ldap_filter(ldap_props):
    '''
    Get LDAP Filter Properties

    :param ldap_props: Get LDAP Custom properties
    :return: Return a dictionary containing LDAP filter information
    '''
    qd.echo_info('==== Get LDAP Filter from LDAP Properties ====')
    ldap_filter = LDAP_FILTER
    for key, value in LDAP_FILTER.iteritems():
        ldap_filter[key] = ldap_props.get_values(value)
    return ldap_filter


def get_ldap_mapping(ldap_props):
    '''
    Get LDAP group mapping properties

    :param ldap_props: Get LDAP Custom properties
    '''
    ldap_map = {}
    qd.echo_info('==== Get LDAP group mapping from LDAP Properties ====')
    mapping = ldap_props.get_values('jazz_ldap_groupMapping')
    for i in mapping.split(','):
        ldap_map[i.split('=')[0]] = i.split('=')[1]
    return ldap_map


def set_ldap_props(ldap_user_name, temp_password):
    '''
    Set LDAP properties file

    :param ldap_user_name: LDAP user name
    :param temp_password: LDAP user name password
    :return: Return LDAP properties
    '''
    env_props = qd_props.get_default_env_props()
    qd_props.copy_ldap_property_file()
    ldap_props = qd_props.get_ldap_props(env_props)
    ldap_props.set_value('jazz_ldap_primaryid', ldap_user_name)
    ldap_props.set_value('jazz_ldap_primaryid_Password', temp_password)
    return ldap_props


def find_was_common(env_props):
    '''
    Find was_common folder and add it to the env.properties

    :param env_props: Environmental properties
    '''
    was_properties = resource.get_resource('ConfigureWAS/was.properties')
    was_common_dir = os.path.dirname(was_properties)
    env_props.set_value(qd_props.PN_WAS_COMMON_DIR, was_common_dir)


def main():
    '''
    Configure LDAP registry with Liberty

    1. Remove Liberty Basic Registry and add LDAP Registry to server.xml
    2. Create LDAP XML with the custom LDAP properties
    3. Map applications to use LDAP Registry
    '''
    liberty_start_stop.stop_liberty()
    qd.echo_info("==== Configure LDAP for Liberty Profile ====")
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)
    was_user_name = user_props.get_values(qd_props.PN_USER_WAS_USER_NAME)
    find_was_common(env_props)
    set_ldap_props(was_user_name, user_props.get_values(qd_props.PN_USER_WAS_USER_PASSWORD))
    ldap_config()
    create_ldap_xml()
    qd.echo_info('==== Liberty LDAP Configuration Finished ====')


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    if qd_props.check_app_server():
        main()
    exit(0)
