#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Set Liberty as a Service
See :func:`main` function.
'''
import os
import sys
import xml.etree.ElementTree as ET

# Due to circular import between liberty_config and liberty_service, imports need to stay relative,
# instead of the from X import Y format.
import liberty_start_stop  # pylint:disable=relative-import
import liberty_config  # pylint:disable=relative-import

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from qdlib import qd, qd_props, install_util, resource

WIN_PROPS = ('set DAEMON_EXEC=', 'set SERVER_START_COMMAND=', 'set SERVER_START_COMMAND_PATH=',
             'set SERVER_STOP_COMMAND=', 'set SERVER_STOP_COMMAND_PATH=', 'set LOG_PATH=')

SERVICE_NAME = 'IBM-Jazz-Liberty'


def startup_apps():
    '''
    Modify server.xml to force all applications to load on startup,
    Set a larger a start timeout for applications to start when liberty is started
    '''
    server_xml = qd.path_join(liberty_start_stop.get_jts_dir(), "liberty/servers/clm/server.xml")
    qd.check_file_exists(server_xml)

    tree = ET.parse(server_xml, liberty_start_stop.CommentParser())
    root = tree.getroot()
    web_cont = root.find('webContainer')
    web_cont.attrib = dict(deferServletLoad="false")
    ET.SubElement(root, 'applicationManager', dict(startTimeout="5m"))
    tree = ET.ElementTree(root)
    tree.write(server_xml, xml_declaration=True,
               encoding='utf-8', method="xml")


def reset_permissions(jts_dir):
    '''
    Reset jre/lib/security permisions on Windows

    :param jts_dir: Jazz Team Server Path
    '''
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)
    os_user_name, os_user_password, _ = qd_props.get_os_users(user_props)
    jre_path = qd.path_join(jts_dir, 'jre/lib/security', '*')
    cmd = ['icacls', jre_path, '/reset', '/t']
    install_util.su_call_if_necessary('--', (os_user_name, os_user_password), cmd)


def main():
    '''
    Configure Liberty as a Service

    1. Force all apps to run when the server starts
    2. Create files needed for Windows or Linux service
    3. Set Liberty Service
    '''
    qd.echo_info('==== Create Liberty Service ====')
    startup_apps()
    jts_dir = liberty_start_stop.get_jts_dir()
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)
    if qd.is_linux():
        liberty_path = os.path.dirname(resource.get_resource('ConfigureLiberty/ibm-liberty'))
        os_user_name = user_props.get_values(qd_props.PN_USER_OS_USER_NAME)
        replacement_list = [('_CLM_HOME_', jts_dir)]
        replacement_list += [('_USER_', os_user_name)]
        qd.sed_in_place(qd.path_join(liberty_path, 'ibm-liberty'), replacement_list)
        qd.create_service('ibm-liberty', liberty_path)
    elif qd.is_windows():
        reset_permissions(jts_dir)
        daemon_path = qd.find_a_file(env_props.get_values(qd_props.PN_QD_ROOT), 'prunsrv.exe')
        if daemon_path is not None:
            log_dir = qd.path_join(jts_dir, 'commons/logs')
            qd.check_and_create_folder(log_dir)
            service_file = resource.get_resource('ConfigureLiberty/jts.service.setup.bat')
            qd.sed_in_place(service_file, [('_JTS_DIR_', qd.escape_path(jts_dir))])
            liberty_config.insert_to_file(service_file, 'set DAEMON_EXEC="' + daemon_path + '"\n',
                                          '_DAEMON_DIR_', action=False)
            return_code = qd.call(service_file)
            if return_code == 0:
                os_user_name, os_user_password, _ = qd_props.get_os_users(user_props)
                service_user_name = ''.join(['obj=.\\', os_user_name])
                service_password = ''.join(['password=', os_user_password])
                sc_cmd = ['sc', 'config', SERVICE_NAME, 'start=delayed-auto', service_user_name,
                          service_password]
                install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                                  sc_cmd)
                qd.echo_info('Liberty added as a Service')
        else:
            qd.echo_error("Fail to add Liberty as a service")
            qd.echo_error("Make sure Apache Commons Daemon is installed correctly")
            qd.echo_error("And re-run liberty_service script")
            qd.exit_with_err_msg("Fail to add Liberty as a service")
    else:
        qd.exit_with_err_msg('Unsupported OS')
    qd.echo_info('==== Liberty Service Created ====')


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    if qd_props.check_app_server():
        main()
    exit(0)
