#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Enable SSO on Liberty
See :func:`main` function.
'''
import xml.etree.ElementTree as ET
import sys
import os
import socket

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from ConfigureLiberty import liberty_start_stop, liberty_config
from qdlib import qd, qd_props

JTS_DIR = liberty_start_stop.get_jts_dir()
''' Jazz Team Server Path '''
LTPA = {'keysFileName': 'resources/security/jtsLTPAkey',
        'keysPassword': 'WebAS',
        'expiration': '120'}


def enable_sso():
    '''
    Modify server.xml to enable SSO
    '''
    server_xml = qd.path_join(JTS_DIR, 'liberty/servers/clm/server.xml')
    qd.check_file_exists(server_xml)

    tree = ET.parse(server_xml, liberty_start_stop.CommentParser())
    root = tree.getroot()
    domain = socket.getfqdn().split(".", 1)
    ET.SubElement(root, 'webAppSecurity', dict(singleSignonEnabled="true"))
    ET.SubElement(root, 'webAppSecurity', dict(ssoDomainNames=domain[1]))
    ET.SubElement(root, 'ltpa', LTPA)
    tree = ET.ElementTree(root)
    tree.write(server_xml, xml_declaration=True,
               encoding='utf-8', method="xml")
    liberty_start_stop.pretty_print_xml(server_xml)


def config_ltpakey(ltpakey):
    '''
    Add Ldap parameters and host to ltpakey
    '''
    env_props = qd_props.get_default_env_props()
    ldap_props = qd_props.get_ldap_props(env_props)
    ldap_realm = ldap_props.get_values('jazz_userrealm')
    ldap_arg = 'com.ibm.websphere.ltpa.Realm='
    new_line = ldap_arg + ldap_realm + '\n'
    liberty_config.insert_to_file(ltpakey, new_line, line_search=ldap_arg, action=False)
    hostname_arg = 'com.ibm.websphere.CreationHost='
    hostname = socket.getfqdn().split(".", 1)
    new_line = hostname_arg + hostname[0] + '\n'
    liberty_config.insert_to_file(ltpakey, new_line, line_search=hostname_arg, action=False)


def main():
    '''
    Enable SSO for Liberty

    1. Enable and Configure SSO
    2. Create and update jtsLTPAkey parameters
    3. Copy jtsLTPAkey to temp folder
    4. Change permissions and owner to jtsLTPAkey file
    '''
    liberty_start_stop.stop_liberty()
    qd.echo_info("==== Configure SSO for Liberty Profile ====")
    enable_sso()
    env_props = qd_props.get_default_env_props()
    if 'JTS' in env_props.get_values('localTags'):
        qd.echo_info('Update jtsLTAPKey parameters')
        qd.move_file(qd.path_join(JTS_DIR, 'liberty/servers/clm/resources/security/ltpa.keys'),
                     qd.path_join(JTS_DIR, 'liberty/servers/clm/resources/security/jtsLTPAkey'))
        config_ltpakey(qd.path_join(JTS_DIR, 'liberty/servers/clm/resources/security/jtsLTPAkey'))
        qd.chmod_777(qd.path_join(JTS_DIR, 'liberty/servers/clm/resources/security/jtsLTPAkey'),
                     mode='a+r')
        qd.echo_info('Copy jtsLTAPkey to temp folder')
        qd.copy_file_with_stat(qd.path_join(JTS_DIR,
                                            'liberty/servers/clm/resources/security/jtsLTPAkey'),
                               qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                                            qd_props.DEFAULT_REL_TMP_DIR))
        if qd.is_linux():
            jts_ltpakey_path = qd.path_join(qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                                                         qd_props.DEFAULT_REL_TMP_DIR),
                                            'jtsLTPAkey')
            stats = os.stat(qd.path_join(JTS_DIR,
                                         'liberty/servers/clm/resources/security/jtsLTPAkey'))
            qd.chown(jts_ltpakey_path, stats.st_uid, stats.st_gid)

    qd.echo_info('==== Liberty SSO Configuration Finished ====')


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    if qd_props.check_app_server():
        main()
    exit(0)
