#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Contains common modules for Liberty Configuration
'''
import xml.etree.ElementTree as ET
from xml.dom import minidom
import sys
import os
import time
import subprocess

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from qdlib import qd, qd_props, install_util

LIBERTY_SERVER = "liberty/wlp/bin/server"
''' Liberty server tool '''

SERVER_STARTUP = qd.os_switch('server.startup', 'server.startup.bat')
''' Start Liberty tool '''

WIN_LIBERTY_SERVICE = 'IBM-Jazz-Liberty'
''' Windows Liberty Service '''


def start_liberty():
    '''
    This module helps start Liberty Server
    '''
    if not verify_server():
        qd.echo_info('==== Start Liberty Server ====')
        env_props = qd_props.get_default_env_props()
        user_props = qd_props.get_user_props(env_props)
        os_user_name, os_user_password = qd_props.get_os_users_creds(user_props)
        service_cmd = qd.path_join(get_jts_dir(), SERVER_STARTUP)
        if qd.is_windows():
            service_cmd = 'net start ' + WIN_LIBERTY_SERVICE
            install_util.su_call_if_necessary('--', (os_user_name, os_user_password), service_cmd)
            timeout = time.time() + 60 * 5
            while True:
                return_code = verify_server()
                if return_code:
                    break
                if time.time() > timeout:
                    return_code = 0
                    break
                time.sleep(10)
        elif qd.is_linux():
            return_code = install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                                            service_cmd)
            return_code = not bool(return_code)
        if return_code:
            qd.echo_info('Liberty Server Started')
        else:
            qd.exit_with_err_msg('Failed to start Liberty Server')
    else:
        qd.echo_info('==== Liberty Server Running ====')


def stop_liberty(timeout=60 * 10):
    '''
    This module helps stop Liberty Server

    :param timeout: time (seconds) to wait for liberty to stop before exiting with an error
    '''
    qd.echo_info('==== Stop Liberty Server ====')
    if verify_server():
        server_tool = qd.path_join(get_jts_dir(), LIBERTY_SERVER)
        subprocess.Popen(server_tool + ' stop clm', shell=True, stdin=None, stdout=None,
                         stderr=None, close_fds=True)
        timeout_time = time.time() + timeout
        while True:
            return_code = verify_server()
            if return_code:
                pass
            else:
                qd.echo_info('Liberty Server Stopped')
                break
            if time.time() > timeout_time:
                qd.exit_with_err_msg('Failed to stop Liberty Server with in [%s] seconds'
                                     % str(timeout))
                break
            time.sleep(10)
    else:
        qd.echo_info('Liberty Server already Stopped')


def verify_server():
    '''
    This module helps check server status

    :return: True if Liberty Server is running else return False
    '''
    server_tool = qd.path_join(get_jts_dir(), LIBERTY_SERVER)
    return_code = qd.call(server_tool + ' status clm', shell=True)
    return not bool(return_code)


def restart_liberty():
    '''
    This module helps restart Liberty Server
    '''
    qd.echo_info('==== Restarting Liberty Server ====')
    if verify_server():
        stop_liberty()
    start_liberty()


def get_jts_dir():
    '''
    Get Liberty server directory

    :return: Returns Liberty server directory
    '''
    if qd_props.check_app_server():
        env_props = qd_props.get_default_env_props()
        return qd.path_join(env_props.get_values(qd_props.PN_CLM_HOME_PREP), 'server')


class CommentParser(ET.XMLTreeBuilder):
    '''
    Add parser to get the comments from the XML file
    '''
    def __init__(self):
        ET.XMLTreeBuilder.__init__(self)
        self._parser.CommentHandler = self.handle_comment

    def handle_comment(self, data):
        '''
        Module helper to handle comments of XML files
        '''
        self._target.start(ET.Comment, {})
        self._target.data(data)
        self._target.end(ET.Comment)


def pretty_print_xml(xml_file):
    '''
    Pretty print XML files
    May add some additional white spaces

    :param xml_file: xml file to apply the pretty print
    '''
    qd.check_file_exists(xml_file)
    tree = ET.parse(xml_file, CommentParser())
    root = tree.getroot()
    xml_str = minidom.parseString(ET.tostring(root)).toprettyxml()
    with open(xml_file, "w") as xml:
        xml.write(xml_str.encode('utf-8'))


def main(argv):
    '''
    Start, Stop and Restart Liberty Server

    :param start: Starts Liberty Server
    :param stop: Stops Liberty Server
    :param restart: Restart Liberty Server
    '''
    if len(argv) == 1:
        option = argv[0]
        option = option.lower()
    else:
        qd.exit_with_err_msg("Missing argument [ start, stop, restart ]")
    if option == 'start':
        start_liberty()
    elif option == 'stop':
        stop_liberty()
    elif option == 'restart':
        restart_liberty()
    else:
        qd.exit_with_err_msg("Argument %s not supported" % option)


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    if qd_props.check_app_server():
        main(sys.argv[1:])
    exit(0)
