#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Provides utilities, such as:

- Access shared folders on windows platform
- Copy jtsLTPA key from JTS node to other WAS node
- Copy serverroot.cer from all WAS node to IHS node
'''

import os
import shutil
import sys

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..', '../../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from ConfigureWAS.WASCommon import enable_sso
from qdlib import qd, qd_props


def prepare_ihs(remote_host, count, file_to_be_copy, target_dir):
    '''
    Copy serverroot.cer to IHS node

    :param remote_host: the remote host that has serverroot.cer file
    :param count: the number of node
    :param file_to_be_copy: file to be copied
    :param target_dir: the directory hold the certificates on IHS node
    '''
    net_share_prefix = r'\\'
    qd.check_and_create_folder(target_dir)
    network_share_path = qd.path_join(net_share_prefix, remote_host,
                                      qd_props.DEFAULT_REL_TMP_DIR, file_to_be_copy)
    shutil.copyfile(network_share_path, qd.path_join(target_dir,
                                                     ''.join(['serverroot_', str(count), '.cer'])))


def prepare_sso_was_node(remote_host, file_to_be_copy, target_dir):
    '''Copy jtsLTPA key file

    :param remote_host: the remote host
    :param file_to_be_copy: file to be copied
    :param target_dir: the target directory
    '''
    net_share_prefix = r'\\'
    network_share_path = qd.path_join(net_share_prefix, remote_host,
                                      qd_props.DEFAULT_REL_TMP_DIR, file_to_be_copy)
    key_file = qd.copy_file(network_share_path, target_dir)
    qd.check_file_exists(key_file)


def get_jts_hostname(env_props, total_nodes):
    '''
    Get JTS hostname

    :param env_props: environment properties
    :param total_nodes: number of nodes
    :return: JTS hostname
    '''
    count = 1
    while count <= total_nodes:
        node_app_list = ''.join(['NODE', str(count), '_AppList'])
        if 'jts' in env_props.get_values(node_app_list):
            node_host = ''.join(['NODE', str(count), '_HostName'])
            return env_props.get_values(node_host)
        else:
            pass
        count = count + 1
    qd.exit_with_err_msg("Failed to find jts hostname, exiting... ")


def main():
    '''
    Main function; see module documentation.
    '''
    if qd.is_linux():
        qd.echo_info('Linux platform found, do not need run this script')
        qd.exit_(0)
    env_props = qd_props.get_default_env_props()
    total_nodes = int(env_props.get_values(qd_props.PN_CUSTOM_NODES))
    # get the qd_tmp directory for storing the root certificate files
    qd_temp_dir = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                               qd_props.DEFAULT_REL_TMP_DIR)

    if env_props.get_values('WASRole') == 'ihs':
        if os.path.isdir(qd_temp_dir):
            if len(os.walk(qd_temp_dir).next()[2]) > 0:
                qd.echo_info('Script already ran on this Server')
                qd.exit_(0)

    count = 1
    if env_props.get_values('WASRole') == 'ihs':
        while count <= total_nodes:
            node_host = ''.join(['NODE', str(count), '_HostName'])
            remote_server = env_props.get_values(node_host)
            prepare_ihs(remote_server, count, 'serverroot.cer', qd_temp_dir)
            count = count + 1
    else:
        if 'JTS' in env_props.get_values('localTags'):
            pass
        else:
            jts_node = get_jts_hostname(env_props, total_nodes)
            if env_props.get_values('WASRole') == 'was':
                prepare_sso_was_node(jts_node, 'jtsLTPAkey', qd_temp_dir)
                enable_sso.main(['import'])
            elif env_props.get_values('WASRole') == 'liberty':
                jts_temp_dir = env_props.get_values(qd_props.PN_CLM_HOME_PREP)
                jts_security_dir = qd.path_join(jts_temp_dir,
                                                'server/liberty/servers/clm/resources/security/')
                prepare_sso_was_node(jts_node, 'jtsLTPAkey', jts_security_dir)
                cmd = ['net', 'share', qd_props.DEFAULT_REL_TMP_DIR, '/Y', '/DELETE']
                qd.check_call(cmd)


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    if qd_props.check_app_server(is_ihs=True):
        main()
    exit(0)
