#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Update WebSphere configuration:

1. Set up appropriate heapsize based on the total physical memory size
2. Set up JVM properties and custom properties
3. Set up Threadpool properties
4. Set up session manager properties
5. Set up LDAP authentication
6. Configure single sign on option - export/import jts LTPA key
'''
import os
import sys

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..', '../../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from ConfigureWAS.WASCommon import enable_ldap, enable_sso
from qdlib import qd, qd_props, install_util, resource


def execute_wsadmin_command(envprops, jython_file_full_path):
    '''
    Execute wasadmin.sh or wsadmin.bat to run a jython file

    :param envprops: environment property, to get the QD_Root directory
    :param jython_file_full_path: absolute jython file path
    '''
    user_props = qd_props.get_user_props(envprops)
    os_user_name, os_user_password = qd_props.get_os_users_creds(user_props)
    escaped_was_properties_path = resource.get_escaped_resource('ConfigureWAS/was.properties')

    cmd = [qd.path_join(envprops.get_values(qd_props.PN_QD_ROOT),
                        qd_props.DEFAULT_REL_WAS_ROOT, 'bin',
                        qd_props.DEFAULT_WSADMIN_EXE), '-lang', 'jython',
           '-f', jython_file_full_path, escaped_was_properties_path]
    install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                      cmd, call_method=qd.check_call)


def main():
    '''Main function; see module documentation.'''
    # set up local environment for this script
    was_properties = resource.get_resource('ConfigureWAS/was.properties')
    was_common_dir = os.path.dirname(was_properties)
    # make sure others can execute any file under the folder, especially the shell scripts
    qd.chmod_r_777(was_common_dir, mode='755')
    env_props = qd_props.get_default_env_props()
    env_props.set_value(qd_props.PN_WAS_COMMON_DIR, was_common_dir)
    local_tags = env_props.get_values('localTags')

    qd.echo_info("was_common_dir is [%s]" % was_common_dir)
    # we need convert "\\" to "/" in the was.properties,
    # since the path passed into WASadmin need to be c:/IBMQD/JazzTeamServer
    # str(jazz_home) is required to go from the tuple to a string!
    jazz_home = str(env_props.get_values(qd_props.PN_CLM_HOME_PREP))
    converted_clm_home = jazz_home.replace('\\', '/')
    replacement_list = [('_CLM_HOME_', converted_clm_home)]
    replacement_list += qd.get_heapsize_replacement()

    if env_props.get_values(qd_props.PN_DB_VENDOR) == 'oracle':
        qd.QDEnv.set_value_to_prop_file(was_properties, 'ORACLE_JAR_NAME',
                                        env_props.get_values(qd_props.PN_DB_DRIVER_FILE))
        qd.QDEnv.set_value_to_prop_file(was_properties,
                                        'jvmcp.file.ORACLE_JDBC_DRIVER_FILE', 'server/oracle')
    if env_props.get_values(qd_props.PN_DB_VENDOR) == 'sqlserver':
        qd.QDEnv.set_value_to_prop_file(was_properties,
                                        'SQLSERVER_JAR_NAME',
                                        env_props.get_values(qd_props.PN_DB_DRIVER_FILE))
        qd.QDEnv.set_value_to_prop_file(was_properties,
                                        'jvmcp.file.SQLSERVER_JDBC_DRIVER_FILE', 'server/sqlserver')

    # Update all appropriate values in was.properties file
    qd.sed_in_place(was_properties, replacement_list)

    qd.echo_info("Setting up JVM properties")
    execute_wsadmin_command(env_props, resource.get_resource('ConfigureWAS/set_jvm_props.jy'))

    qd.echo_info("Setting up Threadpool properties")
    execute_wsadmin_command(env_props,
                            resource.get_resource('ConfigureWAS/set_threadpool_props.jy'))

    qd.echo_info("Setting up session manager properties")
    execute_wsadmin_command(env_props,
                            resource.get_resource('ConfigureWAS/set_session_manager_props.jy'))

    if ('DM' in local_tags or 'RM' in local_tags) and qd.is_linux():
        add_display = [('Bootstrap values ...',
                        'ADDED FOR CONVERTER CLUSTERING SUPPORT\nDISPLAY=:3.0\n'
                        'export DISPLAY\n\n# Bootstrap values ...')]
        qd.sed_in_place(qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                        qd_props.DEFAULT_REL_WAS_ROOT, 'bin', 'startServer.sh'), add_display)
        qd.sed_in_place(qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                        qd_props.DEFAULT_REL_WAS_ROOT, 'bin', 'startNode.sh'), add_display)

    enable_ldap.main()
    if 'jts' in env_props.get_values(qd_props.PN_APPS):
        enable_sso.main(['export'])
    qd.echo_info("Successfully completed {scriptname} execution")


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    if qd_props.check_app_server():
        main()
    exit(0)
