#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Create DB2 database for all Jazz products.  See :func:`main` function.
'''
import sys
import os

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from qdlib import qd, qd_props, install_util


def main():
    '''
    Create db2 database.  Create database option is specified by user or use the default QD
    settings.
    '''
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)
    media_props = qd_props.get_media_props(env_props)
    db2admin_user = user_props.get_values(qd_props.PN_DB2_ADMIN_USER)
    temp_db2admin_pwd = qd.decode_base64(user_props.get_values(qd_props.PN_DB2_ADMIN_PWD))
    db_props = qd_props.get_database_props(env_props)

    def create_db(app):
        '''
        create database per application
        :param app application name
        '''
        db_name = db_props.get_values(''.join([app, 'DBName']))
        db_option = db_props.get_values(''.join(['create', app, 'DBOption']))

        if qd.is_linux():
            create_cmd = ['db2', 'create', 'database', db_name, db_option]
        elif qd.is_windows():
            db2_install_dir = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                                           media_props.get_values('db2MediaRoot'), 'SQLLIB')
            db2_path_env = qd.path_join(db2_install_dir, 'BIN')
            db2_path_env += os.pathsep + qd.path_join(db2_install_dir, 'FUNCTION')
            db2_path_env += os.pathsep + qd.path_join(db2_install_dir, 'SAMPLES', 'REPL')
            os.environ['PATH'] += os.pathsep + db2_path_env

            create_cmd = ['db2cmd', '-w', '-c', '-i', 'db2', 'create', 'database',
                          db_name, db_option]
        else:
            qd.exit_with_err_msg("Unsupport platform")
        install_util.su_call_if_necessary('-', (db2admin_user, temp_db2admin_pwd),
                                          create_cmd, call_method=qd.check_call)
        if app == 'lqe' or app == 'dw':
            special_config_db_cmd_1 = 'db2 update db cfg for ' + db_name + ' using LOGFILSIZ 20000'
            special_config_db_cmd_2 = 'db2 update db cfg for ' + db_name + ' using LOGPRIMARY 25'
            special_config_db_cmd_3 = 'db2 update db cfg for ' + db_name + ' using LOGSECOND 100'
            if qd.is_windows():
                special_config_db_cmd_1 = ['db2cmd', '-w', '-c', '-i', special_config_db_cmd_1]
                special_config_db_cmd_2 = ['db2cmd', '-w', '-c', '-i', special_config_db_cmd_2]
                special_config_db_cmd_3 = ['db2cmd', '-w', '-c', '-i', special_config_db_cmd_3]
            install_util.su_call_if_necessary('-', (db2admin_user, temp_db2admin_pwd),
                                              special_config_db_cmd_1, call_method=qd.check_call)
            install_util.su_call_if_necessary('-', (db2admin_user, temp_db2admin_pwd),
                                              special_config_db_cmd_2, call_method=qd.check_call)
            install_util.su_call_if_necessary('-', (db2admin_user, temp_db2admin_pwd),
                                              special_config_db_cmd_3, call_method=qd.check_call)
    all_apps = env_props.get_values(qd_props.PN_DB_APP_LIST)
    apps = getattr(all_apps, 'split')(',')
    for app_name in apps:
        if app_name in 'rs':
            # rs does not need db2 database
            pass
        else:
            create_db(app_name)
    create_db('dw')


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main()
    exit(0)
