#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
    Create DB2 database for all Jazz products
'''
import sys
import os
from subprocess import Popen, PIPE

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from qdlib import qd, qd_props


def run_sql_query(sql_command, connect_string):
    '''
    Execute the oracle sql command
    :param sql_command: the sql command that will need to be executed
    :param connect_string: the database connection string
    :return the sql comand exit code
    '''
    session = Popen(['sqlplus', '-S', connect_string, 'as', 'sysdba'],
                    stdin=PIPE, stdout=PIPE, stderr=PIPE)
    session.stdin.write(sql_command)
    query_result, _ = session.communicate()
    qd.echo_info(query_result)
    return session.returncode


def main():
    '''
    Create database for all Jazz products
    1. Create the db storage folder if it does not exist
    2. Create the sql script file
    3. Execute the sql script file
    '''
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)
    db_props = qd_props.get_database_props(env_props)
    oracle_instance = db_props.get_values(qd_props.PN_ORACLE_INSTANCE)

    temporacle_sys_password = qd.decode_base64(user_props.get_values('oracleSYSPassword'))
    tempdbuserpassword = qd.decode_base64(user_props.get_values('oracleDBUsersPassword'))
    # Create the table space path folder
    table_space_path = db_props.get_values('oracleDBStorage_Home')
    qd.check_and_create_folder(table_space_path)

    def create_db():
        '''
        Run sql script
        '''
        sql_file = qd.path_join(table_space_path, 'create_db.sql')
        # Create the sql script
        with open(sql_file, 'w') as file_handle:
            file_handle.write('whenever sqlerror exit 1 ;\n')
            file_handle.write('set echo on\n')
            file_handle.write('set heading off\n')
            file_handle.write('set pagesize 0\n')
            file_handle.write('set linesize 210\n')
            file_handle.write('set space 0\n')
            file_handle.write('set newpage 0\n')
            file_handle.write('set verify off\n')

            # Get all applications that need create database
            apps = getattr(env_props.get_values(qd_props.PN_DB_APP_LIST), 'split')(',')
            apps.extend(['dw'])
            for app in apps:
                if app in 'rs':
                    # rs does not need db2 database
                    pass
                else:
                    app_ts_path = qd.path_join(table_space_path, app)
                    qd.check_and_create_folder(app_ts_path)

                    ts_name = db_props.get_values(''.join(['oracle', app, 'TableSpace']))
                    ts_create_option = db_props.get_values(''.join(['oracle', app, 'CreateTS']))
                    temp_ts_name = db_props.get_values(''.join(['oracle', app, 'TempTableSpace']))
                    temp_ts_create_option = db_props.get_values(''.join(['oracle',
                                                                         app, 'CreateTempTS']))
                    db_user = db_props.get_values(''.join(['oracle', app, 'User']))
                    db_user_grant_option = db_props.get_values(''.join(['oracle',
                                                                        app, 'GrantOption']))
                    data_file_path = qd.escape_path(qd.path_join(app_ts_path,
                                                                 ''.join([app, '.dbf'])))
                    temp_data_file_path = qd.escape_path(qd.path_join(app_ts_path,
                                                                      ''.join([app, 'temp.dbf'])))

                    file_handle.write('CREATE BIGFILE TABLESPACE %s  DATAFILE \'%s\' %s;%s' %
                                      (ts_name, data_file_path, ts_create_option, '\n'))
                    file_handle.write('CREATE TEMPORARY TABLESPACE %s TEMPFILE \'%s\' %s;%s' %
                                      (temp_ts_name, temp_data_file_path,
                                       temp_ts_create_option, '\n'))
                    # pylint: disable=line-too-long
                    file_handle.write('CREATE USER %s IDENTIFIED BY %s DEFAULT TABLESPACE %s QUOTA UNLIMITED ON %s TEMPORARY TABLESPACE %s;%s'
                                      % (db_user, tempdbuserpassword, ts_name, ts_name, temp_ts_name, '\n'))
                    file_handle.write('GRANT %s TO %s;%s' % (db_user_grant_option, db_user, '\n'))
                    # pylint: enable=line-too-long
            file_handle.write('exit;\n')
        connect_string = ''.join(['sys/', temporacle_sys_password, '@', oracle_instance])
        sql_command = '@' + sql_file
        if run_sql_query(sql_command, connect_string) == 0:
            qd.echo_info('Successfuly create database for all application')
        else:
            qd.exit_with_err_msg('Failed to create database for application')

    create_db()


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main()
    exit(0)
