#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Create SQLSEVER databases
See :func:`main` function.
'''
import sys
import os

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from qdlib import qd, qd_props, install_util


def create_db(app):
    '''
    Create database for each application

    :param app: Application name
    '''
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)
    db_props = qd_props.get_database_props(env_props)
    db_name = db_props.get_values(''.join([app, 'DBName']))
    sql_instance = db_props.get_values(qd_props.PN_SQLSERVER_INSTANCE)
    sql_cmd = 'CREATE DATABASE ' + db_name
    run_sql_cmd(env_props, sql_instance, sql_cmd)

    def change_owner():
        '''
        Change SQL database owner
        '''
        db_user = db_props.get_values(''.join([app, 'DBuser']))
        db_pwd = qd.decode_base64(user_props.get_values(qd_props.PN_SQL_DB_USER_PASS))
        sql_cmd = 'CREATE LOGIN ' + db_user + ' WITH PASSWORD = ' + "'" + db_pwd + "'"
        run_sql_cmd(env_props, sql_instance, sql_cmd)
        sql_cmd = 'USE ' + db_name + ';' + 'exec sp_changedbowner ' + "'" + db_user + "'"
        run_sql_cmd(env_props, sql_instance, sql_cmd)

    def alter_database():
        '''
        Alter Database to change collation and row version system
        '''
        sql_collation = db_props.get_values(qd_props.PN_SQLSERVER_COLLATION)
        sql_cmd = 'ALTER DATABASE ' + db_name + ' COLLATE ' + sql_collation
        run_sql_cmd(env_props, sql_instance, sql_cmd)
        sql_cmd = 'ALTER DATABASE ' + db_name + ' SET READ_COMMITTED_SNAPSHOT ON'
        run_sql_cmd(env_props, sql_instance, sql_cmd)
    change_owner()
    alter_database()


def run_sql_cmd(env_props, sql_instance, sql_cmd):
    '''
    Run SQL Command

    :param env_props: Environmental properties
    :param sql_instance: SQL server Instance
    :param sql_cmd: SQL command to run
    '''
    user_props = qd_props.get_user_props(env_props)
    os_user_name, os_user_password = qd_props.get_os_users_creds(user_props)
    cmd = ['sqlcmd', '-S', '.\\' + sql_instance, '-Q', sql_cmd]
    install_util.su_call_if_necessary('--', (os_user_name, os_user_password), cmd)


def main():
    '''
    Create SQL Server Database
    1. Create Database for each instance of the Jazz team Server
    2. Create User and Password and change ownership of the database
    3. Change the collation of the databases
    4. Change the row version system
    '''
    env_props = qd_props.get_default_env_props()
    all_apps = env_props.get_values(qd_props.PN_DB_APP_LIST)
    apps = getattr(all_apps, 'split')(',')
    qd.echo_info('==== Creating SQL Server Database ====')
    for app_name in apps:
        if app_name in 'rs':
            # rs does not need sql database
            pass
        else:
            create_db(app_name)
    create_db('dw')


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    if qd.is_windows():
        main()
    exit(0)
