#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Install the various applications.  See :func:`main` function.
'''
import os
import sys

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from qdlib import qd, install_manager, media, qd_props, qd_topo, install_util, resource

TAG_NAME_LIBERTY = 'liberty'  #:
''' Tag name for OOTB embedded liberty app server '''
TAG_NAME_DERBY = 'derby'  #:
''' Tag name for OOTB embedded liberty DB '''

PN_APP_MEDIA_ROOT = 'appMediaRoot'  #:
''' Property name - app media root, in environment, used by UCD '''
PN_APP_MEDIA_CONTENT = 'appMediaContent'  #:
''' Property name - app media content, in media (only when ZIP/HTTP) '''
PN_APP_FILE = 'appFile'  #:
''' Property name - app file, in media '''

PN_ORACLE_MEDIA_ROOT = 'oracleMediaRoot'  #:
''' Property name - oracle media root '''
PN_ORACLE_VERSION = 'oracleVersion'  #:
''' Property name - oracle version '''
PN_SQL_MEDIA_ROOT = 'sqlMediaRoot'
''' Property name - sql media root '''
PN_SQL_VERSION = 'sqlVersion'
''' Property name - sql version '''

PN_SSE_BUILD_ID = 'SSE_BUILD_ID'  #:
''' Property name - SSE build id, in environment, by Rational_SSE UCD component '''
PN_SSE_BUILDROOTURL = 'SSE_BUILDROOTURL'  #:
''' Property name - SSE build root URL, in environment, by Rational_SSE UCD component '''
PN_SSE_PRODUCT_OFFERING_SUFFIX = 'SSEProductOfferingSuffix'  #:
''' Property name - SSE product offering suffix, in environment, by Rational_SSE UCD component '''
PN_CLM_BUILD_ID = 'CLM_BUILD_ID'  #:
''' Property name - CLM build id, in environment, by clm UCD component '''
PN_CLM_BUILDROOTURL = 'CLM_BUILDROOTURL'  #:
''' Property name - CLM build root URL, in environment, by CLM UCD component '''
PN_CLM_PRODUCT_OFFERING_SUFFIX = 'CLMProductOfferingSuffix'  #:
''' Property name - CLM product offering suffix, by CLM UCD component '''

DB_DRIVER = {'oracle': ['ojdbc6.jar', 'ojdbc7.jar', 'ojdbc8.jar'],
             'sqlserver': ['sqljdbc4.jar', 'sqljdbc41.jar']}
''' DB Driver options '''


class AppMedia(media.QDMedia):
    '''App IM Repo Media'''

    def __init__(self, media_props):

        super(AppMedia, self).__init__(media_props)
        self.a_media_root_prop = PN_APP_MEDIA_ROOT
        self.a_file_prop = PN_APP_FILE

        self.app_media_content = None  # calculated. SSE, CLM, CE
        self.app_media_type = None  # stored in _calculate_app_media_content for later use
        self.app_media_root = None  # stored in _calculate_app_media_content for later use
        self.media_access = None  # stored in _calculate_app_media_content for later use

    def _calculate_app_media_type(self):
        '''
        Calculate the app_media_type value (SSE, CE, CLM, FILE)
        '''

        def _get_app_media_type(app_media_root):
            # get app_media_type from the first two letters of app_media_root
            return {'SS': 'SSE', 'CE': 'CE', 'CL': 'CLM'}.get(app_media_root[:2],
                                                              app_media_root[:2])

        media_access = self.get_media_access()
        app_media_root = self.get_app_media_root()
        qd.echo_info('using [%s] for media access' % media_access)
        qd.echo_info("using [%s] for media root" % app_media_root)

        if media_access in ['HTTP', 'HTTPS', 'FTP']:
            app_media_type = _get_app_media_type(app_media_root)
        elif media_access == 'FILE':
            app_media_type = media_access
        else:
            assert False, '[%s] media access not supported' % media_access
        self.app_media_type = app_media_type

        qd.echo_info('using [%s] for media type' % app_media_type)
        return app_media_type

    def _calculate_app_media_content(self):
        '''
        Calculate the app_media_content value (SSE, CE, CLM)
        '''

        def _get_app_media_content(app_media_root):
            # get app media content form the first two letters of app_media_root
            return {'SS': 'SSE', 'CE': 'CE', 'CL': 'CLM'}.get(app_media_root[:2],
                                                              app_media_root[:2])

        media_access = self.get_media_access()
        app_media_type = self.get_app_media_type()
        app_media_root = self.get_app_media_root()
        qd.echo_info('using [%s] for media access' % media_access)
        qd.echo_info("using [%s] for media type" % app_media_type)
        qd.echo_info("using [%s] for media root" % app_media_root)

        if app_media_type in ['SSE', 'CE', 'CLM']:
            app_media_content = app_media_type
        elif app_media_type == 'ZIP':
            if media_access in ['HTTP', 'HTTPS']:
                qd.echo_info('$appMediaContent supplied by media.properties')
                app_media_content = self.pget(PN_APP_MEDIA_CONTENT)
            else:
                app_media_content = _get_app_media_content(app_media_root)

        elif app_media_type == 'FILE':
            app_media_content = _get_app_media_content(app_media_root)
        else:
            assert False, '[%s] media location type not supported'

        qd.echo_info('using [%s] media content' % app_media_content)
        return app_media_content

    def get_app_media_type(self):
        '''
        Override QDMedia.get_app_media_type

        :return: the internal property [app_media_type]
        '''

        if not self.app_media_type:
            self.app_media_type = self._calculate_app_media_type()
            assert self.app_media_type, 'media type not set!'
            assert self.app_media_type in ['SSE', 'CLM', 'CE', 'FILE'], \
                '[%s] media type not supported' % self.app_media_type

        return self.app_media_type

    def get_app_media_content(self):
        '''
        Get the app media content.

        :return: the internal property [app_media_content]
        '''

        if not self.app_media_content:
            self.app_media_content = self._calculate_app_media_content()
            assert self.app_media_content, 'appMediaContent not set!'
            assert self.app_media_content in ['SSE', 'CLM', 'CE'], \
                '[%s] media content not supported' % self.app_media_content

        return self.app_media_content

    def get_app_media_root(self):
        '''
        Get the app media root.

        :return: the internal property [app_media_root]
        '''

        if not self.app_media_root:
            self.app_media_root = self.pget(self.a_media_root_prop)
            assert self.app_media_root, 'app media root not set!'
        return self.app_media_root

    def get_media_access(self):
        '''
        Get the media_access value.

        :return: the internal property [media_access]
        '''

        if not self.media_access:
            self.media_access = self.pget(self.media_access_prop)
            assert self.media_access, 'media access not set!'
            assert self.media_access in ['HTTP', 'HTTPS', 'FILE', 'FTP'], \
                '[%s] media access type not supported'
        return self.media_access

    def _custom_prepare_media(self):
        '''
        Determine Jazz product repository location

        :return: Repository location if determined, None otherwise
        '''

        assert self.get_app_media_content()  # to ensure app_media_content
        # app_media_type = self.pget(self.a_media_type_prop)
        app_media_type = self.app_media_type
        if self.pget(self.media_access_prop) in ['FTP']:
            media_server = self.pget(self.media_server_prop)
            a_media_root = self.pget(self.a_media_root_prop)
            a_file = self.pget(self.a_file_prop)
            return qd.url_join(media_server, a_media_root, a_file)

        elif app_media_type in ['SSE', 'CE']:
            # using SSE repo
            sse_build_id = self.pget(PN_SSE_BUILD_ID)
            sse_build_root_url = self.pget(PN_SSE_BUILDROOTURL)
            sse_product_offering_suffix = self.pget(PN_SSE_PRODUCT_OFFERING_SUFFIX)

            qd.echo_info("using SSE common repository, $!sse_build_root_url [%s], "
                         "$!sse_build_id [%s]" % (sse_build_root_url, sse_build_id))
            return qd.url_join(sse_build_root_url, sse_product_offering_suffix)

        elif app_media_type in ['CLM']:
            # using CLM repo
            clm_build_id = self.pget(PN_CLM_BUILD_ID)
            clm_buildrooturl = self.pget(PN_CLM_BUILDROOTURL)
            clm_product_offering_suffix = self.pget(PN_CLM_PRODUCT_OFFERING_SUFFIX)
            qd.echo_info(
                "using CLM common repository, $!clm_buildrooturl [%s], $!clm_build_id [%s]" %
                (clm_buildrooturl, clm_build_id))
            return qd.url_join(clm_buildrooturl, clm_product_offering_suffix)

        else:
            return None

    def get_remote_file_location(self):
        '''
        Override QDMedia.get_remote_file_location
        Jazz product does not require version in the remote media location

        :return: The URL path for the Jazz product remote location
        '''

        # aVersion is removed from the original
        media_server = self.pget(self.media_root_prop)
        a_media_root = self.pget(self.a_media_root_prop)
        a_file = self.pget(self.a_file_prop)
        return qd.url_join(media_server, a_media_root, a_file)

    def get_unzip_dir_from_repo(self):
        ''' '$downloadDir/repo' instead of 'unzipped' '''
        return media.QDMedia.get_unzip_dir(self, 'repo')

    @classmethod
    def get_result_path_from_unzip_dir(cls, unzip_dir):
        '''
        Get repository.config file path from unzip directory

        :param unzip_dir: The directory name for the unzipped content
        :return: repository.config absolute path
        '''
        return qd.path_join(unzip_dir, 'repository.config')  # specify 'repository.config'


# pylint: disable=too-few-public-methods
class DriverMedia(media.QDMedia):
    '''Driver Media'''
    def __init__(self, media_props, db_vendor):
        super(DriverMedia, self).__init__(media_props)
        if db_vendor == 'oracle':
            self.a_media_root_prop = PN_ORACLE_MEDIA_ROOT
            self.a_version_prop = PN_ORACLE_VERSION
        elif db_vendor == 'sqlserver':
            self.a_media_root_prop = PN_SQL_MEDIA_ROOT
            self.a_version_prop = PN_SQL_VERSION

    def calculate_driver_download_path(self, driver_name=None):
        # Disable pylint invalid-name because function name is > 30 chars and want to keep it
        # pylint: disable=invalid-name
        '''
        Get the db driver location in the media server

        :param driver_name: The oracle ojdbc driver name. e.g. ojdbc8.jar, ojdbc7.jar, ojdbc6.jar
        :return: The absolute oracle driver path
        :param driver_name: The db driver name. e.g. ojdbc7.jar, sqljdbc4.1.jar
        :return: The absolute db driver path
        '''
        # aVersion is removed from the original
        media_server = self.pget(self.media_server_prop)
        media_root = self.pget(self.media_root_prop)
        a_media_root = self.pget(self.a_media_root_prop)
        a_media_version = self.pget(self.a_version_prop)
        media_access = self.pget(self.media_access_prop)
        if media_access in ['HTTP', 'HTTPS', 'FTP']:
            return qd.url_join(media_server, a_media_root, a_media_version, driver_name)
        else:
            return qd.path_join(media_root, a_media_root, a_media_version, driver_name)

    def get_download_dir(self):
        '''
        Get the downloaded db driver folder path

        :return: db driver folder path
        '''
        return qd.path_join(self.pget(self.qd_internal_media_prop), "drivers")


def get_im_components(app_media_content, repo_loc, clm_home, clm_cache=None):
    '''App IM Components'''

    def get_license_compo():
        '''media-dependent components'''

        # pylint: disable=misplaced-bare-raise
        if app_media_content in ['SSE', 'CE']:
            return install_manager.IMRespCompo(name='_license', flags='SSE_LICENSE_OFFER')
        elif app_media_content == 'CLM':
            return install_manager.IMRespCompo(name='_license', flags='CLM_LICENSE_OFFER')
        else:
            raise  # unknown app media content

    # list all components
    result = [
        # base component
        install_manager.get_default_base_compo(repo_loc, clm_home, clm_cache),

        # app server
        install_manager.IMRespCompo(name=qd_topo.TAG_LIBERTY,
                                    repl={'NOAPPSERVER_DISABLE': 'true',
                                          'LIBERTY_DISABLE': 'false'}),
        install_manager.IMRespCompo(name='!' + qd_topo.TAG_LIBERTY,
                                    repl={'NOAPPSERVER_DISABLE': 'false',
                                          'LIBERTY_DISABLE': 'true'}),
        install_manager.IMRespCompo(name=TAG_NAME_DERBY),  # dummy. nothing to do.

        # jts and its options
        install_manager.IMRespCompo(name='jts', depends='_license,_profile',
                                    flags='JTS_OFFER,HIDE_PROFILE_DATA', data=dict(war='jts.war')),
        get_license_compo(),
        install_manager.IMRespCompo(name='_profile', flags='HIDE_PROFILE_DATA'),

        # rs related components
        install_manager.IMRespCompo(name='lqe', depends='_lqejrs'),
        install_manager.IMRespCompo(name='dcc', depends='_dccjrs'),
        install_manager.IMRespCompo(name='rs', depends='_rsjrs'),
        install_manager.IMRespCompo(name='_lqejrs', flags='JRS_OFFER,HIDE_PROFILE_DATA',
                                    repl={'JRS_FEATURES': ',lqe'}, data=dict(war='lqe.war')),
        install_manager.IMRespCompo(name='_dccjrs', flags='JRS_OFFER,HIDE_PROFILE_DATA',
                                    repl={'JRS_FEATURES': ',dcc'}, data=dict(war='dcc.war')),
        install_manager.IMRespCompo(name='_rsjrs', flags='JRS_OFFER,HIDE_PROFILE_DATA',
                                    repl={'JRS_FEATURES': ',jrs'}, data=dict(war='rs.war')),
        # relm
        install_manager.IMRespCompo(name='relm', flags='RELM_OFFER,HIDE_PROFILE_DATA',
                                    repl={'RELM_FEATURES': ',relm'}, data=dict(war='relm.war')),

        # common apps
        install_manager.IMRespCompo(name='ccm', flags='CCM_OFFER,HIDE_PROFILE_DATA',
                                    data=dict(war='ccm.war')),
        install_manager.IMRespCompo(name='qm', flags='QM_OFFER,HIDE_PROFILE_DATA',
                                    data=dict(war='qm.war')),
        install_manager.IMRespCompo(name='rm', flags='RM_OFFER,HIDE_PROFILE_DATA',
                                    data=dict(war='rm.war')),
        install_manager.IMRespCompo(name='ldx', flags='LDX_OFFER,HIDE_PROFILE_DATA',
                                    data=dict(war='ldx.war')),
        install_manager.IMRespCompo(name='gc', flags='GC_OFFER,HIDE_PROFILE_DATA',
                                    data=dict(war='gc.war')),

        # aggr
        install_manager.IMRespCompo(name='clm', depends='jts,ccm,qm,rm,dcc,rs,relm,gc,ldx'),
    ]

    # add SSE components
    if app_media_content in ['SSE', 'CE']:
        result += [
            # common apps
            install_manager.IMRespCompo(name='am', flags='AM_OFFER,HIDE_PROFILE_DATA',
                                        data=dict(war='am.war')),
            install_manager.IMRespCompo(name='dm', flags='DM_OFFER,HIDE_PROFILE_DATA',
                                        data=dict(war='dm.war')),
            install_manager.IMRespCompo(name='vvc', flags='VVC_OFFER,HIDE_PROFILE_DATA',
                                        data=dict(war='vvc.war')),
            # aggr
            install_manager.IMRespCompo(name='sse',
                                        depends='jts,am,ccm,qm,rm,dm,vvc,dcc,rs,relm,gc,ldx'),
        ]

    return result


def check_wars_exist(installed_components, clm_home):
    '''
    Check if the WAR files exist.

    :param installed_components: application name
    :param clm_home: Jazz application installed directory
    '''

    qd.echo_info('echo "check war file exists')
    is_embliberty = any(component.name == TAG_NAME_LIBERTY for component in installed_components)
    for component in installed_components:
        if component.data and component.data.get('war'):
            war = component.data.get('war')
            if is_embliberty:
                war_path = qd.path_join(clm_home, 'server', 'liberty', 'clmServerTemplate',
                                        'apps', war + '.zip')
            else:
                war_path = qd.path_join(clm_home, 'server', 'webapps', war)
            qd.check_file_exists(war_path)


def get_im_template_file(app_media_content, app_media_root):
    '''
    App IM Components

    :param app_media_content: Product type. e.g. SSE, CLM, CE
    :param app_media_root: Product root folder in the media server
    :return: Generic template file name based on product release version
    '''

    # already know app_media_content, now set the IM XML response file name
    if app_media_content in ['SSE', 'CLM', 'CE']:
        # get app_release by removing SSE-, CLM-, or CE- from the beginning of app_media_root
        app_release = app_media_root[len(app_media_content):]
    else:
        # does not matter as $appMediaContent will fail when tested later
        app_release = '601'
    qd.echo_info('app_media_root [%s], app_release [%s]' % (app_media_root, app_release))

    # so now have $appRelease = one of {60,600,601,602,etc}
    if int(app_release) < 601:
        # pre liberty
        im_template_file = 'generic_template600.xml'
    else:
        # liberty supported
        #  im_template_file=generic_template601.xml
        im_template_file = 'generic_template%s.xml' % app_release
    qd.echo_info('using [%s] as the IM response file' % im_template_file)
    im_template_file = resource.get_resource('InstallApp/%s' % im_template_file)
    return im_template_file


# pylint: disable=too-many-locals,too-many-statements,too-many-branches
def main():
    '''
    Main code for install_app - no arguments are passed in.
    '''

    env_props = qd_props.get_default_env_props()
    media_props = qd_props.get_media_props(env_props)

    # ======================================================================
    # ==== IM command params, not supporting use of keyring file
    keyring_file_loc = media.get_im_keyring_file_location()
    im_secure_storage_file = qd.path_join(keyring_file_loc, 'IM.Cred')
    im_password_file = qd.path_join(keyring_file_loc, 'password.txt')

    # ======================================================================
    create_media_only = qd_props.load_create_media_only(media_props)
    debug_do_not_install = qd_props.load_debug_do_not_install(media_props)

    qd_root, clm_cache, im_location = env_props.get_values(qd_props.PN_QD_ROOT,
                                                           qd_props.PN_IM_SHARED_DIR_PREP,
                                                           qd_props.PN_IM_ECLIPSE_DIR_PREP)
    clm_home = qd.path_join(qd_root, qd_props.DEFAULT_REL_CLM_HOME)
    # save for other installers to use
    env_props.set_value(qd_props.PN_CLM_HOME_PREP, clm_home)
    # ==== input parameters come from media.properties or the GUI via env.properties ====
    # -- $APPS
    # -- $appMediaType
    topo = qd_topo.QDTopo(env_props.get_values(qd_props.UCD_PN_GLOBAL_TAG_LISTING))
    thost = topo.get_host(env_props.get_values('localHostname', localHostname=qd.get_host_name()))
    apps = thost.get_local_apps()  # env_props.get_values('APPS')

    # === exit if no apps are selected
    if not apps:
        qd.echo_info("No apps are selected to install. Exiting...")
        qd.exit_(0)

    # --- Media
    med = AppMedia(media_props)
    med.get_app_media_content()  # try to load mandatory props early
    # get template
    im_template_file = get_im_template_file(med.get_app_media_content(), med.get_app_media_root())

    # ==== Set correct ulimit ====
    if qd.is_linux():
        script_dir = os.path.dirname(resource.get_resource('InstallApp/testulimit.sh'))
        qd.chmod_r_777(script_dir, mode='755')
        qd.echo_info("Set correct ulimit")
        qd.check_call('./testulimit.sh', cwd=script_dir)

    # set $RepoLoc location
    install_repo_loc = med.prepare_media()
    # === exit if only creating a media server
    if create_media_only:
        qd.echo_info("media server update done")
        qd.exit_(0)

    # ======================================================================
    # ==== set control flags ====
    # decide when an app can be included and which license to use
    additional_im_compo_candidates = ['liberty']
    required_compos = apps + [t for t in thost.tags if t in additional_im_compo_candidates]

    qd.echo_info("setting control flags")
    im_components = get_im_components(med.get_app_media_content(), med.result, clm_home, clm_cache)
    selected_components, selected_names, unresolved_names = \
        install_manager.select_im_resp_compos(required_compos, im_components)
    if unresolved_names:
        # === only continue if all required apps are in the repo ====
        qd.echo_error('One or more of the required apps [%s] are missing from the repository'
                      % unresolved_names)
        qd.exit_(1)
    else:
        qd.echo_info('all required repo apps are in the repository, continue to install step')

    # generate im_xml
    flags, replacements = install_manager.get_im_flags_and_replacements(selected_components)
    im_xml = qd.path_join(os.path.dirname(im_template_file), 'IMresponses.xml')
    install_manager.generate_imxml(im_template_file, im_xml, replacements, flags)

    # === DEBUG early exit
    if debug_do_not_install:
        qd.echo_debug(
            'XML response file created but exiting prior to installing applications')
        qd.exit_(0)

    # ======================================================================
    # ==== install component via IM ====
    user_props = qd_props.get_user_props(env_props)
    user_name, user_password, primary_group = qd_props.get_os_users_creds_and_group(user_props)
    # need to create {clm_home} folder so IM can install into it
    install_util.ensure_folder_and_set_owner(clm_home, user_name, primary_group)
    qd.chmod_777(clm_home, '755')

    qd.echo_info("Install repository location [%s]" % (install_repo_loc))
    qd.echo_info("Install [%s] to the system as [%s],"
                 " using Installation Manager response file [%s]"
                 % (selected_names, user_name, im_xml))
    if med.app_media_type in ['CLM', 'CE', 'SSE'] and med.media_access not in ['FTP']:
        # only need credentials for a "secure repo"
        install_manager.execute_im(im_xml, secure_storage=im_secure_storage_file,
                                   password_file=im_password_file, im_location=im_location,
                                   im_usercreds=(user_name, user_password))
    else:
        install_manager.execute_im(im_xml,
                                   im_location=im_location,
                                   im_usercreds=(user_name, user_password))

    # ==== check war file installed ====
    qd.echo_info('check war file exists')
    check_wars_exist(selected_components, clm_home)
    qd.echo_info('component [%s] installed' % selected_names)

    # tidy up by deleting anything downloaded or unzipped
    med.cleanup()
    db_vendor = env_props.get_values('DB_VENDOR')
    if db_vendor == "oracle" or db_vendor == "sqlserver":
        qd.echo_info('download %s driver file' % db_vendor)
        driver_med = DriverMedia(media_props, db_vendor)
        if db_vendor == "oracle":
            db_version = driver_med.get_my_version()
            if db_version == '12.2':
                db_driver_file = DB_DRIVER[db_vendor][2]
            elif db_version == '12.1':
                db_driver_file = DB_DRIVER[db_vendor][1]
            else:
                db_driver_file = DB_DRIVER[db_vendor][0]
        else:
            if med.app_media_content in ['SSE', 'CLM', 'CE']:
                # get app_release by removing SSE-, CLM-, CE- from the beginning of app_media_root
                app_release = med.app_media_root[len(med.app_media_content):]
            else:
                # does not matter as $appMediaContent will fail when tested later
                app_release = '601'

            if int(app_release) < 603:
                db_driver_file = DB_DRIVER[db_vendor][0]
            else:
                db_driver_file = DB_DRIVER[db_vendor][1]
        qd_props.get_default_env_props().set_value('dbDriverFile', db_driver_file)

        driver_loc = driver_med.calculate_driver_download_path(db_driver_file)
        download_file = driver_med.download_or_copy(driver_loc)
        if driver_loc == download_file:
            driver_dir = driver_med.get_download_dir()
            qd.echo_info("download dir [%s]" % driver_dir)
            qd.check_and_create_folder_777(driver_dir)
            qd.copy_file(download_file, driver_dir)

    # === finished ====
    qd.echo_info('SUCCESS: installation complete')


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main()
    exit(0)
