#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Install DB2 on the target machine.
'''
import sys
import os

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from qdlib import qd, media, qd_props, install_util, resource

PN_DB2_MEDIA_TYPE = 'db2MediaType'  # in media
PN_DB2_MEDIA_ROOT = 'db2MediaRoot'  # in env, by UCD
PN_DB2_FILE = 'db2File'
PN_DB2_VERSION = 'db2Version'


class DB2Media(media.QDMedia):
    '''
    DB2Media class, inherit from QDMedia

    - extended custom_prepare_media
    '''
    def __init__(self, media_props, env_props):
        super(DB2Media, self).__init__(media_props, env_props)
        self.a_media_type_prop = PN_DB2_MEDIA_TYPE
        self.a_media_root_prop = PN_DB2_MEDIA_ROOT
        self.a_version_prop = PN_DB2_VERSION
        self.a_file_prop = PN_DB2_FILE
        self.a_dict = {"OSdependent": qd.os_switch()}


def update_db2fs_folder(source_file, fs_dir):
    '''
    Update the db2fs folder during run time,
    should only be run if a new db2 server to be setup

    :param source_file: the file to be updated
    :param fs_dir: the db2fs directory, used by create_database.py
    '''
    if qd.is_linux():
        qd.sed_in_place(source_file, [('_DB2FS_DIR_', fs_dir),
                                      ('_DRIVE_LETTER_', fs_dir)])
    else:
        (drive, _) = os.path.splitdrive(fs_dir)
        qd.sed_in_place(source_file, [('_DB2FS_DIR_', qd.escape_path(fs_dir)),
                                      ('_DRIVE_LETTER_', drive)])


# pylint: disable=too-many-locals, too-many-branches
def main():
    '''Main method; see module documentation.'''
    # set up local environment for this script
    working_dir = os.getcwdu()
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)  # get credentials
    media_props = qd_props.get_media_props(env_props)  # get media properties

    check_db2_user = user_props.get_values(qd_props.PN_DB2_ADMIN_USER)
    if qd.is_linux():
        if qd.call(['id', check_db2_user]) == 0:
            qd.echo_info("The DB2 user [%s] already exist, assume db2 is installed"
                         % check_db2_user)
            qd.exit_(0)
    else:
        if qd.call(['net', 'user', check_db2_user]) == 0:
            qd.echo_info("The DB2 user [%s] already exist, assume db2 is installed"
                         % check_db2_user)
            qd.exit_(0)

    # === prepare media
    db2med = DB2Media(media_props, env_props)
    db2med.prepare_media()

    if qd_props.load_create_media_only(media_props):
        qd.echo_info("media server update done")
        qd.exit_(0)

    if qd_props.load_debug_do_not_install(media_props):
        qd.echo_info(
            'XML response file created but exiting prior to installing applications')
        qd.exit_(0)

    # check and update users [DAS User, Instance User and Fence User]
    # This is only needed for Linux, windows installation will create automatically
    if qd.is_linux():
        # pylint: disable=line-too-long
        # ibmQseInst_User property used in multiple files, so the property name set in qd_props.py,
        # All other db2 user properties used only in this file, so we direct used in this file
        das_user, instance_user, fence_user = user_props.get_values('ibmQseDAS_User', qd_props.PN_DB2_ADMIN_USER, 'ibmQseFence_User')
        das_group, instance_group, fence_group = user_props.get_values('ibmQseDAS_Group', 'ibmQseInst_Group', 'ibmQseFence_Group')
        # pylint: enable=line-too-long
        db_user_pwd = qd.decode_base64(user_props.get_values(qd_props.PN_DB2_ADMIN_PWD))
        install_util.ensure_os_user_and_group(das_user, db_user_pwd, das_group)
        install_util.ensure_os_user_and_group(instance_user, db_user_pwd, instance_group)
        install_util.ensure_os_user_and_group(fence_user, db_user_pwd, fence_group)
        # add second group for db2instance user
        install_util.ensure_os_user_and_group(instance_user, db_user_pwd, das_group)
    else:
        das_user, instance_user = user_props.get_values('ibmQseDAS_User',
                                                        qd_props.PN_DB2_ADMIN_USER)
        das_group, instance_group = user_props.get_values('ibmQseDAS_Group', 'ibmQseInst_Group')
        db_user_pwd = qd.decode_base64(user_props.get_values(qd_props.PN_DB2_ADMIN_PWD))
    qd.echo_info("Installing db2")
    replacement_list = []
    if qd.is_linux():
        db2_install_dir = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                                       media_props.get_values(db2med.a_media_root_prop),
                                       db2med.get_my_version())
        db2_exe_name = 'db2setup'
    else:
        db2_exe_name = 'setup.exe'
        db2_install_dir = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                                       media_props.get_values(db2med.a_media_root_prop), 'SQLLIB')
        db2_common_app_dir = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                                          media_props.get_values(db2med.a_media_root_prop),
                                          'ProgramData')
        db2_ssh_dir = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                                   media_props.get_values(db2med.a_media_root_prop),
                                   'IBM_SSH_Server')
        replacement_list += [('_PORT_NUMBER_', env_props.get_values('DB2PORT'))]
        replacement_list += [('_SSH_INSTALL_FOLDER_', qd.escape_path(db2_ssh_dir))]
        replacement_list += [('_DB2_COMMON_APP_DATA_DIR_', qd.escape_path(db2_common_app_dir))]
        replacement_list += [('_DB2_ADMIN_USER_', instance_user)]
        replacement_list += [('_DB2_DAS_USER_', das_user)]
        replacement_list += [('_DB2_ADMIN_PASSWORD_', db_user_pwd)]
        replacement_list += [('_DB2_DAS_PASSWORD_', db_user_pwd)]
        replacement_list += [('_DB2_ADMIN_GROUP_', instance_group)]

    replacement_list += [('_DB2_INSTALL_FOLDER_', qd.escape_path(db2_install_dir))]

    if media_props.get_values('installDB2TSAMP') == 'FALSE':
        replacement_list += [('_DB2_TSAMP_COMP_', r'*COMP')]
    else:
        replacement_list += [('_DB2_TSAMP_COMP_', 'COMP')]
    db_template = resource.get_resource('InstallDB2/db2server_template.rsp')
    db_server_rsp = 'db2server.rsp'
    qd.duplicate_file(db_template, db_server_rsp)
    db_server_rsp = resource.get_resource('InstallDB2/db2server.rsp')
    qd.sed_in_place(db_server_rsp, replacement_list)
    db_exe_path = qd.find_a_file(db2med.result, db2_exe_name)
    if db_exe_path is None:
        qd.exit_with_err_msg("Failed to find the db2 exeutable file [%s]" % db2_exe_name)
    # Install DB2
    install_cmd = [db_exe_path, qd.os_switch('-r', '/u'), db_server_rsp, qd.os_switch('', '/m'),
                   qd.os_switch('-l', '/l'), qd.path_join(working_dir, 'db_log.log')]

    qd.check_call(install_cmd, cwd=qd.dirname(db_exe_path))

    def add_license():
        '''
        Add db2 license
        '''
        license_dir = qd.path_join(db2_install_dir, 'license')
        downloaded_license = qd.find_a_file_ext(qd.dirname(db2med.get_download_dir()), '.lic')
        if downloaded_license:
            license_file_name = qd.basename(downloaded_license)
            qd.check_and_create_folder(license_dir)
            qd.copy_file(downloaded_license, license_dir)
            add_license_cmd = [apply_license_exe, '-a',
                               qd.path_join(license_dir, license_file_name)]
            qd.check_call(add_license_cmd)
        else:
            qd.exit_with_err_msg("Unable to find the license file under [%s]"
                                 % qd.dirname(db2med.result))

    apply_license_exe = qd.os_switch(qd.path_join(db2_install_dir, 'adm', 'db2licm'),
                                     qd.path_join(db2_install_dir, 'BIN', 'db2licm'))
    add_license()
    qd.check_call([apply_license_exe, '-l'])

    def post_install_config_linux():
        '''
        Special configuration needed for Linux platform
        1. Create the instance
        2. Setting environment
        3. Enable connections from remote client
        4. Enable auto restart
        5. Start db2
        6. Verify db2set settings
        '''
        sqllib_dir = 'sqllib'
        db2_adm_path = qd.path_join(sqllib_dir, 'adm')
        db2_set_path = qd.path_join(db2_adm_path, 'db2set')
        qd.echo_info("Creating instance....")
        cmd = [qd.path_join(db2_install_dir, 'instance', 'db2icrt'),
               '-a', 'server', '-p', env_props.get_values('DB2PORT'),
               '-u', fence_user, instance_user]
        qd.check_call(cmd)

        qd.echo_info("Setting environment .....")
        cmd = ['.', qd.path_join('~', sqllib_dir, 'db2profile')]
        install_util.su_call_if_necessary('-', (instance_user, db_user_pwd),
                                          cmd, call_method=qd.check_call)

        qd.echo_info("enable connections from remote clients")
        cmd = [qd.path_join('~', db2_set_path), 'DB2COMM=TCPIP']
        install_util.su_call_if_necessary('-', (instance_user, db_user_pwd),
                                          cmd, call_method=qd.check_call)
        qd.echo_info("enable auto restart (after a server reboot)")
        cmd = [qd.path_join('~', db2_set_path), 'DB2AUTOSTART=YES']
        install_util.su_call_if_necessary('-', (instance_user, db_user_pwd),
                                          cmd, call_method=qd.check_call)

        qd.echo_info("start DB2")
        cmd = [qd.path_join('~', db2_adm_path, 'db2start')]
        install_util.su_call_if_necessary('-', (instance_user, db_user_pwd),
                                          cmd, call_method=qd.check_call)
        qd.echo_info("verify db2set")
        cmd = [qd.path_join('~', db2_set_path), '-all']
        install_util.su_call_if_necessary('-', (instance_user, db_user_pwd),
                                          cmd, call_method=qd.check_call)
    if qd.is_linux():
        post_install_config_linux()
        qd.echo_info("Creating the service for Linux")
        ibm_db2_file = resource.get_resource('InstallDB2/ibm-db2')
        qd.sed_in_place(ibm_db2_file, [('_INSTANCE_OWNER_', instance_user)])
        qd.create_service('ibm-db2', os.path.dirname(ibm_db2_file))
    # Update the db2Parameter.properties.
    # If user choose use QD deploy DB2,
    # then we assume put everything under QD_Root directory
    db2fs_dir = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                             qd_props.DEFAULT_REL_DB2FS_DIR)
    # We need update both files since db2Parameter.properties file is copied to env.properties
    update_db2fs_folder(env_props.filename, db2fs_dir)
    update_db2fs_folder(qd_props.get_database_props(env_props).filename, db2fs_dir)
    # === finished ====
    qd.echo_info("SUCCESS: installation complete")
    db2med.cleanup()


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main()
    exit(0)
