#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
    Install Installation Manager on the target machine
'''
import os
import sys

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from qdlib import qd, install_manager, install_util, media, qd_props, resource

PN_IM_MEDIA_TYPE = 'imMediaType'  #:
''' Property name - installation manager media type '''
PN_IM_MEDIA_ROOT = 'imMediaRoot'  #:
''' Property name - installation manager media root '''
PN_IM_VERSION = 'imVersion'  #:
''' Property name - installation manager version '''
PN_IM_FILE = 'imFile'  #:
''' Property name - installation manager file '''


# pylint: disable=too-few-public-methods
class IMMedia(media.QDMedia):
    '''Represents Installation Manager media.'''
    def __init__(self, media_props):
        super(IMMedia, self).__init__(media_props)
        self.a_media_type_prop = PN_IM_MEDIA_TYPE
        self.a_media_root_prop = PN_IM_MEDIA_ROOT
        self.a_version_prop = PN_IM_VERSION
        self.a_file_prop = PN_IM_FILE  # resolves later to the win32 or linux file
        self.a_dict = {"OSdependent": qd.os_switch()}


# pylint: disable=too-many-locals
# pylint: disable=too-many-statements
def main():
    '''Main method; see module documentation.'''

    # set up local environment for this script
    working_dir = os.getcwdu()
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)  # get credentials
    media_props = qd_props.get_media_props(env_props)  # get media properties

    # Files
    im_template_file = resource.get_resource('InstallIM/IM_response_template.xml')
    im_xml = qd.path_join(working_dir, 'IMresponse.xml')

    # Flags
    create_media_only = qd_props.load_create_media_only(media_props)
    debug_do_not_install = qd_props.load_debug_do_not_install(media_props)

    # === prepare media
    med = IMMedia(media_props)
    med.prepare_media()
    im_install_folder = med.result

    qd.echo_debug("installation repo found at [%s]" % im_install_folder)

    # === exit if only creating a media server
    if create_media_only:
        qd.echo_info("media server update done")
        qd.exit_(0)

    # ==== create IM install XML ====
    qd.echo_info("Create IM response file from template [%s]" % im_template_file)
    if os.path.exists(im_xml):
        qd.echo_info("deleting old version of %s" % im_xml)
        os.remove(im_xml)

    qd_root = env_props.get_values(qd_props.PN_QD_ROOT)
    im_shared_dir = qd.path_join(qd_root, qd_props.DEFAULT_REL_IM_SHARED)
    im_dir = qd.path_join(qd_root, qd_props.DEFAULT_REL_IM_ECLIPSE_DIR)
    im_parent_dir = qd.dirname(im_dir)
    # save for other installers to use
    env_props.set_value(qd_props.PN_IM_SHARED_DIR_PREP, im_shared_dir)
    env_props.set_value(qd_props.PN_IM_INST_DIR_PREP, im_parent_dir)
    env_props.set_value(qd_props.PN_IM_ECLIPSE_DIR_PREP, im_dir)

    replacements = {'IM_INSTALL_DIR': im_dir,
                    'IBMQD_IM_REPO': im_install_folder}
    install_manager.generate_imxml(im_template_file, im_xml, replacements)
    qd.check_file_exists(im_xml)

    # === DEBUG early exit
    if debug_do_not_install:
        qd.echo_debug('XML response file created '
                      'but exiting prior to installing applications')
        qd.exit_(0)

    # Create local user for the installation - now done in core_automation_prep.py

    # Create IM installation directory
    # update the user|group chmod to 755; creating folder first if required
    # need to use the primary group
    user_name, user_password, primary_group = qd_props.get_os_users_creds_and_group(user_props)
    install_util.ensure_folder_and_set_owner(im_parent_dir, user_name, primary_group)
    qd.chmod_777(im_parent_dir, '755')

    # Create shared directory used by IM
    install_util.ensure_folder_and_set_owner(im_shared_dir, user_name, primary_group)
    qd.chmod_777(im_shared_dir, '755')
    # zipfile module does not make the file executable while unzip
    qd.chmod_r_777(im_install_folder, 'o+x')
    # Run setup
    # =========
    # installer is at $imInstallFileLocation=${im_install_folder}"/"${im_install_file}
    qd.echo_info("Installing IBM Installation Manager [%s] as %s"
                 % (med.get_my_version(), user_name))

    if install_util.need_user_inst(user_name):
        im_install_file = qd.os_switch('userinstc', 'userinstc.exe')
    else:
        im_install_file = qd.os_switch('installc', 'installc.exe')
    im_exe_path = qd.find_a_file(med.result, im_install_file)
    if im_exe_path is None:
        qd.exit_with_err_msg("Failed to find IM installation executable file [userinstc] from [%s]"
                             % med.result)

    # install
    install_cmd = [im_exe_path, 'input', im_xml, '-acceptLicense']
    if install_util.su_call_if_necessary('--', (user_name, user_password),
                                         install_cmd, cwd=im_install_folder) != 0:
        qd.exit_with_err_msg("Installation Manager install failed")
    qd.echo_info("Installation Manager Successful")

    # tidy up by deleting anything downloaded or unzipped
    med.cleanup()

    # === finished ====
    qd.echo_info("SUCCESS: installation complete")


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main()
    exit(0)
