#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Install Oracle on the target machine.  See :func:`main` function.
'''
import sys
import os
import time
import glob
import subprocess

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from qdlib import qd, media, qd_props, install_util, resource

PN_ORACLE_MEDIA_TYPE = 'oracleMediaType'  # in media
PN_ORACLE_MEDIA_ROOT = 'oracleMediaRoot'  # in env, by UCD
PN_ORACLE_FILE = 'oracleFile'
PN_ORACLE_VERSION = 'oracleVersion'


class OracleMedia(media.QDMedia):
    '''
    Oracle media class, inherit from QDMedia

    - extended custom_prepare_media
    '''
    def __init__(self, media_props, env_props):
        super(OracleMedia, self).__init__(media_props, env_props)
        self.a_media_type_prop = PN_ORACLE_MEDIA_TYPE
        self.a_media_root_prop = PN_ORACLE_MEDIA_ROOT
        self.a_version_prop = PN_ORACLE_VERSION
        self.a_file_prop = PN_ORACLE_FILE
        '''
        Set the OS folder name
        '''
        self.a_dict = {"OSdependent": qd.os_switch()}


def create_win_key(inventory_path):
    '''
    Create a windows registry key for oracle inventory location

    :param inventory_path: oracle installation files location
    '''
    import _winreg
    sub_key_id = r'SOFTWARE\Oracle'
    sub_key_property = 'inst_loc'

    key_handle = _winreg.CreateKey(_winreg.HKEY_LOCAL_MACHINE, sub_key_id)
    _winreg.SetValueEx(key_handle, sub_key_property, 0, _winreg.REG_SZ, inventory_path)


def alter_system(sys_pwd, instance_name, oracle_home, sql_command):
    '''
    Run oracle query to alter database parameters.

    :param sys_pwd: sys account password
    :param instance_name: oracle instance name
    :param oracle_home: The path to oracle install home
    :param sql_command: The sql command to be run
    '''
    # Add the oracle path to PATH variable
    os.environ['PATH'] += os.pathsep + qd.path_join(oracle_home, 'bin')
    os.environ['ORACLE_HOME'] = oracle_home

    connect_string = ''.join(['sys/', sys_pwd, '@', instance_name])
    session = subprocess.Popen(['sqlplus', '-S', connect_string, 'as', 'sysdba'],
                               stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
    qd.echo_info('running sql: [%s]' % sql_command)
    session.stdin.write(sql_command)
    query_result, stderr = session.communicate()
    if stderr:
        qd.exit_with_err_msg("Failed to run sql [%s]" % sql_command)
    qd.echo_info(query_result)


def check_listener(instance_name):
    '''
    Check if the database already exists and if listener is created.  Exit the installation if
    existing oracle instance is found.

    :param instance_name: the database instance name
    :param oracle_home: the oracle_home path
    :return: true if the database already exists; false otherwise
    '''
    cmd = ['tnsping', instance_name]
    try:
        return_status = qd.call(cmd)
    # We need catch the error and continue the script execution.
    # If the path did not found, that means database was not installed on the target machine
    except OSError as error_msg:
        if error_msg.errno == os.errno.ENOENT:
            return False
    return bool(return_status == 0)


def listener(action, oracle_user, oracle_user_password):
    '''
    Perform a specific listener operation.

    :param action: 'stop', 'start', 'status'
    :param oracle_user: user name to run the listener action
    :param oracle_user_password: password of the oracle_user
    '''
    cmd = ['lsnrctl', action]
    install_util.su_call_if_necessary('-', (oracle_user, oracle_user_password),
                                      cmd, call_method=qd.check_call)


def restart_db(os_user, os_user_pwd, oracle_home, oracle_instance):
    '''
    Restart database after the SQL net listener port is changed.

    :param os_user: oracle user name
    :param os_user_pwd: oracle user password
    :param oracle_home: oracle home path
    :param oracle_instance: oracle instance name
    '''
    if qd.is_linux():
        qd.echo_info("Stopping oracle processes")
        cmd_stop = [qd.path_join(oracle_home, 'bin', 'dbshut'), oracle_home]
        cmd_start = [qd.path_join(oracle_home, 'bin', 'dbstart'), oracle_home]
        install_util.su_call_if_necessary('--', (os_user, os_user_pwd), cmd_stop,
                                          call_method=qd.check_call)
        install_util.su_call_if_necessary('--', (os_user, os_user_pwd), cmd_start,
                                          call_method=qd.check_call)
    else:
        service_name = 'OracleService' + oracle_instance
        cmd_stop = ['net', 'stop', service_name]
        cmd_start = ['net', 'start', service_name]
        # We need start the service as administrator who installed oracle,
        # otherwise the service will stop and then disappeared
        service_cmd_retries(cmd_stop)
        service_cmd_retries(cmd_start)
    # We need wait 2 minutes for listener to start
    time.sleep(120)


def service_cmd_retries(cmd):
    '''
    Determine if the service command should re-run
    :param cmd: The start/stop command to be run
    :return none
    '''
    retry = 5
    number = 0
    for number in range(retry):
        number = number + 1
        try:
            output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            qd.echo_info(output)
            break
        except subprocess.CalledProcessError as exc:
            qd.echo_debug(exc.output)
            fatal_message = 'The service name is invalid'
            retry_message = 'The service is not responding to the control function'
            if exc.output.startswith(fatal_message):
                qd.exit_with_err_msg("Failed to run the command [%s]" % cmd)
            elif exc.output.startswith(retry_message):
                # If we have run out of retry, still get the exception,
                # Exit with error
                if number == retry:
                    qd.exit_with_err_msg("Failed to run the command [%s] after retry [%s] times"
                                         % (cmd, number))
                qd.echo_info('Wait for one minute and retry the [%s] times' % number)
                time.sleep(60)
                continue


# pylint: disable=too-many-branches
def main():
    '''
    Install oracle in the target machine

    1. Check if oracle is installed and instance is created
    2. Download the media
    3. Update response file
    4. Execute installation
    5. Post installation
    '''
    # pylint: disable=too-many-locals, too-many-statements
    # set up local environment for this script
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)  # get credentials
    media_props = qd_props.get_media_props(env_props)  # get media properties
    db_props = qd_props.get_database_props(env_props)

    oracle_user = user_props.get_values(qd_props.PN_ORACLE_OS_USER)
    oracle_user_password = qd.decode_base64(user_props.get_values(qd_props.PN_ORACLE_OS_PWD))
    oracle_instance = db_props.get_values(qd_props.PN_ORACLE_INSTANCE)
    oracle_port = db_props.get_values(qd_props.PN_ORACLE_PORT)

    if check_listener(oracle_instance):
        qd.echo_info("Oracle listener [%s] already started, skip installation" % oracle_instance)
        qd.exit_(0)
    else:
        pass
    # === prepare media
    oraclemed = OracleMedia(media_props, env_props)
    oraclemed.prepare_media()

    # get all passwords
    oracle_all_password = qd.decode_base64(user_props.get_values('oracleALLPassword'))
    oracle_sys_password = qd.decode_base64(user_props.get_values('oracleSYSPassword'))
    oracle_dbsnmp_password = qd.decode_base64(user_props.get_values('oracleDBSNMPPassword'))
    oracle_system_password = qd.decode_base64(user_props.get_values('oracleSYSTEMPassword'))
    oracle_pdbadmin_password = qd.decode_base64(user_props.get_values('oraclePDBADMINPassword'))

    replacement_list = []
    oracle_root = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                               media_props.get_values(oraclemed.a_media_root_prop), 'app')
    oracle_base = qd.path_join(oracle_root, oracle_user)
    oracle_home_dir = qd.path_join(oracle_base, 'product',
                                   media_props.get_values(oraclemed.a_version_prop), 'dbhome_1')
    oracle_data_loc = qd.path_join(oracle_base, 'oradata')
    oracle_inventory_loc = qd.path_join(oracle_root, 'oraInventory')
    replacement_list += [('_ORACLE_TARGET_SERVER_HOST_', env_props.get_values('DB_HOSTNAME'))]
    replacement_list += [('_ORACLE_HOME_', qd.escape_path(oracle_home_dir))]
    replacement_list += [('_ORACLE_BASE_', qd.escape_path(oracle_base))]
    replacement_list += [('_InstanceName_', oracle_instance)]
    replacement_list += [('_ORACLE_DATA_LOC_', qd.escape_path(oracle_data_loc))]
    replacement_list += [('_ORACLE_INVENTORY_LOC_', qd.escape_path(oracle_inventory_loc))]
    replacement_list += [('_ORACLE_USER_', oracle_user)]
    replacement_list += [('_ORACLE_OS_PASSWORD_', oracle_user_password)]

    replacement_list += [('_ALL_PASSWORD_', oracle_all_password)]
    replacement_list += [('_SYS_PASSWORD_', oracle_sys_password)]
    replacement_list += [('_SYSTEM_PASSWORD_', oracle_dbsnmp_password)]
    replacement_list += [('_DBSNMP_PASSWORD_', oracle_system_password)]
    replacement_list += [('_PDBADMIN_PASSWORD_', oracle_pdbadmin_password)]

    if qd.is_linux():
        # pylint: disable=line-too-long
        oracle_primary_group, oracle_additional_group = user_props.get_values('oracleOSUserPrimaryGroup', 'oracleOSUserAdditionalGroup')
        # pylint: enable=line-too-long
        replacement_list += [('_ADDTIONAL_GROUP_', oracle_additional_group)]
        replacement_list += [('_PRIMARY_GROUP_', oracle_primary_group)]
    else:
        replacement_list += [('_ADDTIONAL_GROUP_', '')]
        replacement_list += [('_PRIMARY_GROUP_', '')]
    response_template_name = 'oracle_server_template_' + oraclemed.get_my_version() + '.rsp'
    db_template = resource.get_resource('InstallOracle/%s' % response_template_name)
    db_server_rsp = 'oracle_server.rsp'
    qd.duplicate_file(db_template, db_server_rsp)
    qd.sed_in_place(resource.get_resource('InstallOracle/%s' % db_server_rsp), replacement_list)

    if qd_props.load_create_media_only(media_props):
        qd.echo_info("media server update done")
        qd.exit_(0)

    if qd_props.load_debug_do_not_install(media_props):
        qd.echo_info(
            'XML response file created but exiting prior to installing applications')
        qd.exit_(0)

    # check and update oracle user
    # This is only needed for Linux, windows installation will create automatically
    if qd.is_linux():
        cfg_template_name = 'cfg_template_' + oraclemed.get_my_version() + '.rsp'
        cfg_template = resource.get_resource('InstallOracle/%s' % cfg_template_name)
        cfg_rsp = 'cfg.rsp'
        qd.duplicate_file(cfg_template, cfg_rsp)
        qd.sed_in_place(resource.get_resource('InstallOracle/%s' % cfg_rsp), replacement_list)
        # pylint: disable=line-too-long
        install_util.ensure_os_user_and_group(oracle_user, oracle_user_password, oracle_primary_group)
        install_util.ensure_os_user_and_group(oracle_user, oracle_user_password, oracle_additional_group)
        # pylint: enable=line-too-long
        script_dir = os.path.dirname(resource.get_resource(
            'InstallOracle/prepare_oracle_install.sh'))
        qd.check_call(['sh', 'prepare_oracle_install.sh', oracle_user], cwd=script_dir)
        oracle_exe_name = 'runInstaller'
    else:
        create_win_key(oracle_inventory_loc)
        oracle_exe_name = 'setup.exe'
    db_exe_path = qd.find_a_file(oraclemed.result, oracle_exe_name)
    if db_exe_path is None:
        qd.exit_with_err_msg("Failed to find oracle installation executable file [%s]"
                             % oracle_exe_name)

    def post_linux_install():
        '''
        Configuration after Linux installation completed
        '''
        qd.echo_info('Run Root scripts to clean up install...')
        qd.check_call(['sh', qd.path_join(oracle_inventory_loc, 'orainstRoot.sh')])
        qd.check_call(['sh', qd.path_join(oracle_home_dir, 'root.sh')])
        # Create root .bashrc file
        with open(qd.path_join('/root', '.bashrc'), 'a') as file_h:
            file_h.write('TMP=/tmp;\n')
            file_h.write('export TMP;\n')
            file_h.write(r'TMPDIR=$TMP;')
            file_h.write('\n')
            file_h.write('export TMPDIR;\n')
            file_h.write(''.join(['ORACLE_HOME=', oracle_home_dir, '\n']))
            file_h.write('export ORACLE_HOME\n')
            file_h.write(r'PATH=$PATH:$HOME/bin:$ORACLE_HOME/bin')
            file_h.write('\n')
            file_h.write('export PATH\n')
            file_h.write('unset USERNAME\n')

        bash_profile_file = resource.get_resource('InstallOracle/bash_profile')
        bash_profile_dest = qd.path_join('/home', oracle_user, '.bash_profile')
        bashrc_file = resource.get_resource('InstallOracle/bashrc')
        bashrc_file_dest = qd.path_join('/home', oracle_user, '.bashrc')
        qd.sed_in_place(bash_profile_file, [('_oracleOSUser_', oracle_user)])
        qd.sed_in_place(bashrc_file, [('_oracleOSUser_', oracle_user),
                                      ('_ORACLE_HOME_', oracle_home_dir)])

        qd.move_file(bash_profile_file, bash_profile_dest)
        qd.move_file(bashrc_file, bashrc_file_dest)

        qd.echo_info('Configure listener, create instance based on the installation response file')
        config_cmd = [qd.path_join(oracle_home_dir, 'cfgtoollogs', 'configToolAllCommands'),
                      ''.join(['RESPONSE_FILE=', resource.get_resource(
                          'InstallOracle/%s' % cfg_rsp)])]
        install_util.su_call_if_necessary('-', (oracle_user, oracle_user_password),
                                          config_cmd, call_method=qd.check_call)

        qd.echo_info("Start Oracle Server")
        start_cmd = [qd.path_join(oracle_home_dir, 'bin', 'dbstart'), oracle_home_dir]
        install_util.su_call_if_necessary('-', (oracle_user, oracle_user_password),
                                          start_cmd, call_method=qd.check_call)

        qd.sed_in_place(resource.get_resource('InstallOracle/ora_start.sh'),
                        [('_ORACLE_DATA_LOC_', oracle_data_loc)])
        qd.sed_in_place(resource.get_resource('InstallOracle/ora_stop.sh'),
                        [('_ORACLE_DATA_LOC_', oracle_data_loc)])
        qd.sed_in_place(resource.get_resource('InstallOracle/oracle'),
                        [('_OracleOSUser_', oracle_user)])

        install_util.su_call_if_necessary('-', (oracle_user, oracle_user_password),
                                          ['mkdir', qd.path_join('/home', oracle_user, 'scripts')],
                                          call_method=qd.check_call)

        qd.copy_file(resource.get_resource('InstallOracle/ora_start.sh'),
                     qd.path_join('/home', oracle_user, 'scripts'))
        qd.copy_file(resource.get_resource('InstallOracle/ora_stop.sh'),
                     qd.path_join('/home', oracle_user, 'scripts'))
        qd.chmod_777(qd.path_join('/home', oracle_user, 'scripts', 'ora_start.sh'), '755')
        qd.chmod_777(qd.path_join('/home', oracle_user, 'scripts', 'ora_stop.sh'), '755')

        qd.create_service('oracle', os.path.dirname(resource.get_resource('InstallOracle/oracle')))
        qd.sed_in_place('/etc/oratab', [(''.join([oracle_home_dir, ':N']),
                                         ''.join([oracle_home_dir, ':Y']))])

        # Update the oracleParameter.properties.
        # If user choose use QD deploy oracle,
        # then we assume put everything under QD_Root directory
        db_storage_dir = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                                      qd_props.DEFAULT_REL_ORACLE_STORAGE_DIR)
        # Update both files since oracleParameter.properties file is copied to env.properties
        qd.sed_in_place(env_props.filename, [('_DBSTORAGE_FOLDER_', db_storage_dir)])
        qd.sed_in_place(qd_props.get_database_props(env_props).filename,
                        [('_DBSTORAGE_FOLDER_', db_storage_dir)])

    def post_windows_install():
        '''
        Checking if windows installation completed and the listener is started
        '''
        time.sleep(1200)
        number_tries = 20
        wait_time = 120
        number = 0
        install_log_existed = False
        for number in range(number_tries):
            number = number + 1
            log_loc = qd.path_join(oracle_inventory_loc, 'logs\\installActions*.log')
            install_log = glob.glob(log_loc)
            if len(install_log) == 0:
                qd.exit_with_err_msg("No installation action log file [%s] found, exiting"
                                     % log_loc)
            else:
                # sort the installAction*.log file, only take the latest install action log file
                install_log.sort(key=os.path.getmtime)
                grep_message = 'Exit Status is '
                log_file_name = install_log[len(install_log) - 1]
                qd.echo_debug("Found the log file [%s]" % log_file_name)
                exit_code_line = qd.grep(log_file_name, grep_message, ignore_case=True)
                # check if "Exit status is " string existence,
                # make sure installation and configuration completed
                if exit_code_line:
                    import re
                    # Get the exit status number
                    match_str = re.search(''.join([grep_message, r'(\d+)$']), exit_code_line[0])
                    if match_str:
                        install_exit_code = match_str.group(1)
                        qd.echo_debug("Install exit code is [%s]" % install_exit_code)
                        if install_exit_code != "0":
                            qd.echo_debug("Failed to install and configure oracle database")
                            qd.echo_debug("Please check the log file [%s]" % install_log)
                            qd.exit_with_err_msg("Oracle installation failed with exit code [%s]"
                                                 % install_exit_code)
                        else:
                            install_log_existed = True
                            qd.echo_info("Checking oracle listener status")
                        break
                else:
                    if number == number_tries:
                        total_time = number_tries * (wait_time / 60) + 10
                        qd.exit_with_err_msg("Failed to finish oracle installation in [%s] minutes"
                                             % total_time)
                    qd.echo_debug("Waiting for 2 minute to re-check install status")
                    time.sleep(wait_time)
                    continue

        if install_log_existed:
            wait_time = 120
            number_tries = 20
            number = 0
            listener_start = False
            oracle_bin_dir = qd.path_join(oracle_home_dir, 'bin')
            os.environ['PATH'] += os.pathsep + oracle_bin_dir
            for number in range(number_tries):
                number = number + 1
                if check_listener(oracle_instance):
                    qd.echo_info("instance [%s] is started" % oracle_instance)
                    listener_start = True
                    break
                else:
                    qd.echo_info('[%s] did not start, wait for [%s] seconds and retry number [%s]'
                                 % (oracle_instance, wait_time, number))
                    time.sleep(wait_time)
            if listener_start:
                # delay for 2 minutes to wait all services are up running
                qd.echo_info("Waiting for 2 minutes to wait all services up running")
                time.sleep(120)
            else:
                # if run out of the number of loops, it means oracle listener did not start
                qd.exit_with_err_msg("Oracle listener did not start")
        else:
            # if run out of the number of loops, it means oracle installation failed
            qd.echo_error("Please check the install log file [%s]"
                          % r'C:\Program Files\Oracle\Inventory\logs\*')
            qd.exit_with_err_msg("Oracle installation failed")

    response_file = resource.get_resource('InstallOracle/%s' % db_server_rsp)

    if qd.is_linux():
        qd.chmod_r_777(oraclemed.temp_unzip_dir, '755')
        # Create the oracle folder, make sure it can be writable by oracle OS user
        # pylint: disable=line-too-long
        qd.check_and_create_folder_777(qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                                                    media_props.get_values(oraclemed.a_media_root_prop)), '777')
        # pylint: enable=line-too-long
        install_cmd = [db_exe_path, '-silent', '-showProgress', '-waitforcompletion',
                       '-responseFile', response_file]
        install_util.su_call_if_necessary('-', (oracle_user, oracle_user_password),
                                          install_cmd, call_method=qd.check_call)
        post_linux_install()
    else:
        install_cmd = [db_exe_path, '-silent', '-nowait', '-responseFile', response_file]
        # We can not use su - oracle to run the installation as the oracle user does not exist yet.
        qd.call(install_cmd)
        post_windows_install()

    # change the port number
    if oracle_port == '1521':
        pass
    else:
        qd.echo_info("change port")
        # delay for 2 minutes to allow all oracle services have been completely started
        time.sleep(120)
        alter_system(oracle_sys_password, oracle_instance, oracle_home_dir,
                     ''.join(['alter system set local_listener=', oracle_instance, ' scope=both;']))
        qd.echo_info("Stopping listener")
        listener('stop', oracle_user, oracle_user_password)
        qd.echo_info("Modify oraclePort [%s] in listener and tnsnames file" % oracle_port)
        listner_file = qd.path_join(oracle_home_dir, 'network', 'admin', 'listener.ora')
        tnsname_file = qd.path_join(oracle_home_dir, 'network', 'admin', 'tnsnames.ora')
        qd.sed_in_place(listner_file, [('1521', oracle_port)])
        qd.sed_in_place(tnsname_file, [('1521', oracle_port)])
        listener('start', oracle_user, oracle_user_password)

    # delay for 2 minutes to allow all oracle services have been completely started
    # time.sleep(120)
    # Restart DB before alter the system process and open cursors,
    # so spfile will created before the query is executed
    restart_db(oracle_user, oracle_user_password, oracle_home_dir, oracle_instance)

    alter_system(oracle_sys_password, oracle_instance, oracle_home_dir,
                 'alter system set processes=1000 scope=spfile;')
    alter_system(oracle_sys_password, oracle_instance, oracle_home_dir,
                 'alter system set open_cursors = 10000 scope=both;')
    restart_db(oracle_user, oracle_user_password, oracle_home_dir, oracle_instance)

    qd.echo_info("SUCCESS: installation complete")
    # Update the oracleParameter.properties.
    # If user choose use QD deploy Oracle,
    # then we assume put everything under QD_Root directory
    dbstorage_dir = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                                 qd_props.DEFAULT_REL_ORACLE_STORAGE_DIR)
    qd.sed_in_place(env_props.filename, [('_DBSTORAGE_FOLDER_', qd.escape_path(dbstorage_dir))])
    qd.sed_in_place(qd_props.get_database_props(env_props).filename,
                    [('_DBSTORAGE_FOLDER_', qd.escape_path(dbstorage_dir))])

    # We need update both files since oracleParameter.properties file is copied to env.properties
    oraclemed.cleanup()


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main()
    exit(0)
