#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Install SQL Server
See :func:`main` function.
'''
import sys
import os
import xml.etree.ElementTree as ET
import time

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from qdlib import qd, media, qd_props, install_util, resource

PN_SQL_MEDIA_TYPE = 'sqlMediaType'  # in media
PN_SQL_MEDIA_ROOT = 'sqlMediaRoot'  # in env, by UCD
PN_SQL_FILE = 'sqlFile'
PN_SQL_VERSION = 'sqlVersion'
''' SQL media properties '''

SQL_BASIC_FEATURES = ('SQLENGINE', 'SNAC_SDK')
SQL_ADVANCED_FEATURES = ('SSMS', 'ADV_SSMS')
''' SQL Server features '''

DEFAULT_INSTALL_PATH = "Microsoft SQL Server"
DEFAULT_SHARED_PATH = "Microsoft SQL Server (x86)"
''' SQL SERVER INSTALL PATH '''

DETACHED_PROCESS = 0x00000008
''' Windows process creation Flag '''


class SQLMedia(media.QDMedia):
    '''
    SQLMedia class, inherit from QDMedia
    - extended custom_prepare_media
    '''
    def __init__(self, media_props, env_props):
        super(SQLMedia, self).__init__(media_props, env_props)
        self.a_media_type_prop = PN_SQL_MEDIA_TYPE
        self.a_media_root_prop = PN_SQL_MEDIA_ROOT
        self.a_version_prop = PN_SQL_VERSION
        self.a_file_prop = PN_SQL_FILE
        self.a_dict = {"OSdependent": qd.os_switch()}


def start_stop_sql_server(sql_instance, action, os_user_name, os_user_password):
    '''
    Start stop SQL Server instance

    :param action: 'start' or 'stop' server
    :param sql_instance: SQL server Instance
    :param os_user_name: OS User to run the service
    :param os_user_password: User decoded password
    '''
    if action != 'start' and action != 'stop':
        qd.exit_with_err_msg('Argument not supported, only start and stop are supported')
    server_cmd = ['net', action, 'SQLAgent$' + sql_instance]
    return_code = install_util.su_call_if_necessary('--', (os_user_name,
                                                           os_user_password), server_cmd)
    if return_code == 0:
        server_cmd = ['net', action, 'MSSQL$' + sql_instance]
        qd.echo_info("==== %s SQL Server Instance [%s] ====" % (action.upper(), sql_instance))
        return_code = install_util.su_call_if_necessary('--', (os_user_name,
                                                               os_user_password), server_cmd)
        if return_code == 0 or return_code == 2:
            qd.echo_info("==== Complete SQL Server %s ====" % action.upper())
        else:
            qd.echo_error("==== Fail to %s SQL Server Instance [%s]" % (action.upper(),
                                                                        'MSSQL$' + sql_instance))
            qd.exit_with_err_msg("Service need to be restarted in order to set SQL Port")
    else:
        qd.echo_error("==== Fail to %s SQL Server Instance [%s]" % (action.upper(),
                                                                    'SQLAgent$' + sql_instance))
        qd.exit_with_err_msg("Service need to be restarted in order to set SQL Port")


def restart_sql_server(env_props, sql_instance):
    '''
    Restart SQL Server Instance

    :param env_props: Environmental properties
    :param sql_instance: SQL server Instance
    '''
    user_props = qd_props.get_user_props(env_props)
    os_user_name, os_user_password = qd_props.get_os_users_creds(user_props)
    start_stop_sql_server(sql_instance, 'stop', os_user_name, os_user_password)
    time.sleep(10)
    start_stop_sql_server(sql_instance, 'start', os_user_name, os_user_password)


def change_port(env_props):
    '''
    Change SQL Server Port number

    :param env_props: Environmental properties
    '''
    qd.echo_info("Set SQL Server Port")
    sdk_tools, sql_tools = get_installation_path()
    new_env = os.environ.copy()
    sdk_tools = qd.path_join(sdk_tools, 'Client SDK/ODBC/110/Tools/Binn')
    sql_tools = qd.path_join(sql_tools, '120/Tools/Binn')
    new_env['PATH'] = new_env['PATH'] + ';' + sdk_tools + ';' + sql_tools
    user_props = qd_props.get_user_props(env_props)
    os_user_name, os_user_password = qd_props.get_os_users_creds(user_props)
    db_props = qd_props.get_database_props(env_props)
    sql_instance = db_props.get_values(qd_props.PN_SQLSERVER_INSTANCE)
    cmd = ['sqlps', '-Command',
           "&{$smo = 'Microsoft.SqlServer.Management.Smo.'" +
           "; $wmi = new-object ($smo +'Wmi.ManagedComputer'); $wmi.Name}"]
    computer_name = qd.check_output(cmd, shell=True, env=new_env).replace('\r\n', '')
    sql_port = db_props.get_values(qd_props.PN_SQLSERVER_PORT)
    replacement_list = [('_NAME_', computer_name)]
    replacement_list += [('_INSTANCE_', sql_instance)]
    replacement_list += [('_PORT_', sql_port)]
    change_port_file = resource.get_resource('InstallSQL/change_port.ps1')
    qd.sed_in_place(change_port_file, replacement_list)
    cmd = ['sqlps', '-Command', ' &{' + change_port_file + '} ']
    return_code = install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                                    cmd, env=new_env)
    if return_code == 1:
        qd.exit_with_err_msg("Error while trying to change SQL Server Port")
    else:
        restart_sql_server(env_props, sql_instance)


def find_file(root, file_name):
    '''
    Find file in directory

    :param root: path to start the search
    :param file_name: file name to search
    :return: Return file_name path
    '''
    for path, _, dir_file in os.walk(root):
        for _file in dir_file:
            if file_name.lower() in _file.lower():
                return qd.path_join(path, _file)
    return None


def get_media_layout(media_info):
    '''
    Get SQL media layout (None, Core, Advanced, Full) to determine version,
    to make the proper installation

    :param media_info: MediaInfo.xml path
    :return: Return SQL version to install
    '''
    qd.check_file_exists(media_info)
    tree = ET.parse(media_info)
    root = tree.getroot()
    for child in root:
        for prop in child:
            if prop.get('Id') == 'MediaLayout':
                return prop.get('Value')
    return None


def get_installation_path():
    '''
    Get the installation and shared folders for installation

    :return: Returns a List containing the installation and shared folders for SQL Server
    '''
    env_props = qd_props.get_default_env_props()
    install_path = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT), DEFAULT_INSTALL_PATH)
    shared_path = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT), DEFAULT_SHARED_PATH)
    return (install_path, shared_path)


def create_config_file(env_props, sql_version):
    '''
    Create the configuration file to be use in the installation

    :param env_props: Environmental properties
    :param sql_version: SQL version to configure (Full, Advanced, Core, None)
    :return: config_file: Return ConfigurationFile path
    # TODO Modify may need to handle 3 different config files
    '''
    qd.echo_info("==== Preparing Configuration file for Installation ====")
    user_props = qd_props.get_user_props(env_props)
    db_props = qd_props.get_database_props(env_props)
    os_user_name, _, _ = qd_props.get_os_users(user_props)
    sa_pass = qd.decode_base64(user_props.get_values(qd_props.PN_SQLSERVER_SA_PASS))
    sql_instance = db_props.get_values(qd_props.PN_SQLSERVER_INSTANCE)
    sql_collation = db_props.get_values(qd_props.PN_SQLSERVER_COLLATION)
    install_path, shared_path = get_installation_path()
    qd.echo_info("==== SQL Version to configure [%s] ====" % sql_version)
    if sql_version == 'None':
        qd.exit_with_err_msg("Unknown SQL version please check media")
    elif sql_version == 'Full' or sql_version == 'Advanced':
        sql_features = ','.join(SQL_BASIC_FEATURES) + ',' + ','.join(SQL_ADVANCED_FEATURES)
        replacement_list = [('_FEATURES_', sql_features)]
    elif sql_version == 'Core':
        replacement_list = [('_FEATURES_', ','.join(SQL_BASIC_FEATURES))]
    replacement_list += [('_INSTANCE_', sql_instance)]
    replacement_list += [('_COLLATION_', sql_collation)]
    replacement_list += [('_OSUSER_', os_user_name)]
    replacement_list += [('_OSPASS_', sa_pass)]
    replacement_list += [('_DATADIR_',
                          qd.escape_path(qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                                                      qd_props.DEFAULT_REL_SQLSERVER_STORAGE_DIR)))]
    replacement_list += [('_SQL_DIR_', qd.escape_path(install_path))]
    replacement_list += [('_SQL_SHARED_DIR_', qd.escape_path(shared_path))]
    configuration_file = resource.get_resource('InstallSQL/ConfigurationFile.ini')
    qd.sed_in_place(configuration_file, replacement_list)
    return configuration_file


def check_sql_server_running(sql_instance, cmd='sqlcmd -L'):
    '''
    Verify SQL Server is installed

    :param sql_instance: SQL Server Instance name
    :return: Return 0 if successful otherwise return 1
    '''
    try:
        sql_cmd = qd.check_output(cmd, shell=True)
        if sql_instance in sql_cmd:
            if cmd == 'sqlcmd -L':
                qd.echo_info("SQL Instance [%s]SQL Server already installed " % sql_instance)
            else:
                qd.echo_info("SQL Instance [%s] Running " % sql_instance)
            return 0
    except (AssertionError, WindowsError):
        if cmd == 'sqlcmd -L':
            qd.echo_info("==== No SQL Server Installation Detected ====")
        else:
            qd.echo_info("==== No SQL Instance Running with name [%s] ====" % sql_instance)
        return 1
    return 1


def main():
    '''
     Install SQL Server
    1. Check if SQLSERVER is installed and instance is created
    2. Download the media
    3. Update response file
    4. Execute installation
    5. Change SQL port
    '''
    env_props = qd_props.get_default_env_props()
    media_props = qd_props.get_media_props(env_props)  # get media properties
    db_props = qd_props.get_database_props(env_props)
    sql_instance = db_props.get_values(qd_props.PN_SQLSERVER_INSTANCE)
    if check_sql_server_running(sql_instance) == 0:
        exit(0)
    else:
        qd.echo_info('==== Prepare to Install SQL Server ====')
    # === prepare media
    sqlmed = SQLMedia(media_props, env_props)
    sqlmed.prepare_media()
    media_info = find_file(sqlmed.result, 'MediaInfo.xml')
    sql_version = get_media_layout(media_info)
    sql_config_file = create_config_file(env_props, sql_version)
    cmd = find_file(sqlmed.result, 'SETUP.exe')
    config_file = '/ConfigurationFile=' + sql_config_file
    sql_setup_cmd = [cmd, config_file]
    qd.echo_info(sql_setup_cmd)

    if qd.check_call(sql_setup_cmd, stdin=None, stdout=None,
                     stderr=None, close_fds=True, creationflags=DETACHED_PROCESS) == 0:
        if check_sql_server_running(sql_instance, cmd='net continue MSSQL$' + sql_instance) == 0:
            qd.echo_info("==== SQL Server Setup Complete Successfully ====")
        else:
            qd.exit_with_err_msg("==== SQL Instance not Started as expected ====")
    else:
        qd.exit_with_err_msg("==== Failed to Install SQL Server ====")
    change_port(env_props)
    qd.echo_info('==== Finish SQL Server Configuration ====')

    sqlmed.cleanup()


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    if qd.is_windows():
        main()
    else:
        qd.exit_with_err_msg("==== OS not supported only windows support SQL Server ====")
    exit(0)
