#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
    Install WebSphere application server
    or IBM Http server based on the machine role
'''
import os
import socket
import sys

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from qdlib import qd, install_manager, media, qd_props, install_util, resource

PN_WAS_MEDIA_TYPE = 'wasMediaType'  # in media
PN_WAS_MEDIA_ROOT = 'wasMediaRoot'  # in env, by UCD
PN_WAS_BASE_VERSION = 'wasBaseVersion'  # in media
PN_WAS_FP_VERSION = 'wasFPVersion'  # in media
PN_WAS_FILE = 'wasFile'
PN_WAS_JAVA_SDK_VERSION = 'wasJavaSDKVersion'


PN_IHS_MEDIA_TYPE = 'ihsMediaType'
PN_IHS_MEDIA_ROOT = 'ihsMediaRoot'
PN_IHS_BASE_VERSION = 'ihsBaseVersion'
PN_IHS_FP_VERSION = 'ihsFPVersion'
PN_IHS_FILE = 'ihsFile'

PN_JAVA_FP_VERSION = 'javaFPVersion'
PN_JAVA_MEDIA_TYPE = 'javaMediaType'
PN_JAVA_FILE = 'javaFile'


class ServerMedia(media.QDMedia):
    '''
    ServerMedia class, inherit from QDMedia

    - extended custom_prepare_media
    '''
    def __init__(self, media_props):
        super(ServerMedia, self).__init__(media_props)
        self.a_fp_version_prop = 'shouldOverRide'
        self.a_name_convention = 'server'

    def _custom_prepare_media(self):
        '''
        Override the QDMedia prepare method

        - Determine the versions (Base and Fix Pack)
        - Download all files under the src directory
        - Process each file as downloaded
        '''
        result_path = ''
        a_media_type = self.pget(self.a_media_type_prop)
        if a_media_type == 'ZIP':
            # Get all media versions that need download
            for index, media_version in enumerate(self.calculate_src_zip_content()):
                # construct unzip directory name
                if index == 0:
                    base_or_fp = ''.join([self.get_name_convention(), 'Base'])
                else:
                    base_or_fp = ''.join([self.get_name_convention(), 'FP'])

                self.a_dict = {"version": media_version}
                # Download or copy, and then extract zip file
                src_zip_dir = self.get_directory_location(**self.a_dict)
                qd.echo_info('using zipped up file repository, %s#src_zip_dir [%s]'
                             % (type(self).__name__, src_zip_dir))
                assert 'notUsed' not in src_zip_dir

                # get a list of files for the src directory and process each file
                image_files = self.get_file_list(src_zip_dir)
                if len(image_files) <= 0:
                    assert False, "File list empty [%s]" % (src_zip_dir)
                for image_file in image_files:
                    result_path = self.process_each_zip_file(src_zip_dir, image_file, base_or_fp)
        else:
            assert False, "unknown media type $%s [%s]" % (self.a_media_type_prop, a_media_type)

        # get the result
        self.result = result_path

        return self.result

    def calculate_src_zip_content(self):
        '''
        Calculate the source zip content.

        - if both base and fix pack version are defined in media.property
          we need install both the base and fix pack.
        - if fix pack version is zero, then only need install base version.
        '''
        version_list = []
        if self.pget(self.a_fp_version_prop) != "0":
            version_list = [self.pget(self.a_version_prop), self.pget(self.a_fp_version_prop)]
        else:
            version_list = [self.pget(self.a_version_prop)]
        return version_list

    def get_name_convention(self):
        '''
        return the name convention
        '''
        return self.a_name_convention

    def get_formatted_version(self):
        '''
        Determine the server version format, then return the correct format of the version hash
        Java version example:

        - major_rel_ver -- 7.1 or 8
        - dot_ver -- 7.1 or 8.0
        - no_dot_ver - 71 or 80
        '''
        version = self.pget(self.a_version_prop)
        version_format = {}
        if '.' in version:
            ver_list = version.split('.')
            no_dot_ver = ver_list[0] + ver_list[1]
            if len(ver_list) == 2:
                if ver_list[1] == '0':
                    version_format = {'major_rel_ver': ver_list[0], 'dot_ver': version,
                                      'no_dot_ver': no_dot_ver, 'jdk_ver': ver_list[0]}
                else:
                    version_format = {'major_rel_ver': version, 'dot_ver': version,
                                      'no_dot_ver': no_dot_ver, 'jdk_ver': version}
            elif len(ver_list) > 2:
                dot_ver = ver_list[0] + '.' + ver_list[1]
                version_format = {'major_rel_ver': dot_ver, 'dot_ver': dot_ver,
                                  'no_dot_ver': no_dot_ver, 'jdk_ver': dot_ver}
            else:
                qd.echo_error("Incorrect format for the version format [%s]" % version)
                qd.echo_error("Expected format similar to 7.1, 8.0, 8.5, 8.5.5")
                qd.exit_(1)
        else:
            version_format = {'major_rel_ver': version, 'dot_ver': version + '.0',
                              'no_dot_ver': version + '0', 'jdk_ver': version}

        return version_format

    def get_requested_java_version(self):
        '''
        Return an unformatted wasJavaSDKVersion requested
        '''
        version = self.pget(self.a_version_prop)

        return version


# pylint: disable=too-few-public-methods
class IHSMedia(ServerMedia):
    '''
    def __init__(self, media_props):
        super(IHSMedia, self).__init__(media_props)
    IBM HTTP Server class

    - Inherit from ServerMedia
    - Set IHS media_type, media_root, major version, fix pack version and unzip folder name
    '''
    def __init__(self, media_props):
        super(IHSMedia, self).__init__(media_props)
        self.a_media_type_prop = PN_IHS_MEDIA_TYPE
        self.a_media_root_prop = PN_IHS_MEDIA_ROOT
        self.a_version_prop = PN_IHS_BASE_VERSION
        self.a_fp_version_prop = PN_IHS_FP_VERSION
        self.a_file_prop = PN_IHS_FILE
        self.a_name_convention = 'ihs'


# pylint: disable=too-few-public-methods
class WASMedia(ServerMedia):
    '''
    WebSphere application server class

    - Inherit from ServerMedia
    - Set WAS media_type, media_root, major version, fix pack version and unzip folder name
    '''
    def __init__(self, media_props):
        super(WASMedia, self).__init__(media_props)
        self.a_media_type_prop = PN_WAS_MEDIA_TYPE
        self.a_media_root_prop = PN_WAS_MEDIA_ROOT
        self.a_version_prop = PN_WAS_BASE_VERSION
        self.a_fp_version_prop = PN_WAS_FP_VERSION
        self.a_file_prop = PN_WAS_FILE
        self.a_name_convention = 'was'


# pylint: disable=too-many-instance-attributes
class JAVAMedia(ServerMedia):
    '''
    Java media download class, prepare the source directory location based on the OS

    - inherit from ServerMedia
    - extended custom_prepare_media
    - extended get_remote_directory_location
    - extended get_local_directory_location
    '''
    def __init__(self, media_props):
        super(JAVAMedia, self).__init__(media_props)
        self.a_version_prop = PN_WAS_JAVA_SDK_VERSION
        self.a_media_type_prop = PN_JAVA_MEDIA_TYPE
        self.a_file_prop = PN_JAVA_FILE
        self.a_dict = {"OSdependent": qd.os_switch()}
        self.install_external_java = False
        self.install_java_fp = False
        self.a_name_convention_root = 'JAVA'
        self.set_java_content()

    def _custom_prepare_media(self):
        '''
        Override the QDMedia method
        '''
        result_path = ''
        a_media_type = self.pget(self.a_media_type_prop)
        if a_media_type == 'ZIP':
            if self.get_install_external_java():
                java_types = [self.a_name_convention_root, self.a_name_convention_root + 'FP']
            else:
                java_types = [self.a_name_convention_root + 'FP']
            for java_type in java_types:
                self.set_a_name_convention(java_type)
                # Download or copy, and then extract foreign zip file
                # $repoZipFileLocation in InstallApp.sh # need to calculate $repoZipFileLocation
                src_zip_dir = self.get_directory_location(**self.a_dict)
                if java_type == self.a_name_convention_root + 'FP':
                    # If the JAVAFP folder does not exist or empty,
                    # it means do not need install JAVA Fix pack
                    if not self.verify_file_list_existence(src_zip_dir):
                        self.set_install_java_fp(False)
                        # return a string, so the override _custom_prepare_media()
                        # will return something and do not attempt to execute again
                        result_path = 'Do not need to download'
                        continue
                    else:
                        self.set_install_java_fp(True)
                qd.echo_info('using zipped up file repository, %s#src_zip_dir [%s]'
                             % (type(self).__name__, src_zip_dir))
                assert 'notUsed' not in src_zip_dir

                timer = qd.start_clock("Ready to get ZIP")
                for image_file in self.get_file_list(src_zip_dir):
                    result_path = self.process_each_zip_file(src_zip_dir, image_file,
                                                             self.get_name_convention())
                qd.stop_clock(timer)
        else:
            assert False, "unknown media type $%s [%s]" % (self.a_media_type_prop, a_media_type)

        # get the result
        self.result = result_path  # RepoLoc in InstallApp.sh

        return self.result

    def set_install_external_java(self, value=False):
        '''
        Set the install external java true or false
        '''
        self.install_external_java = value

    def set_install_java_fp(self, value=False):
        '''
        Set the install java fix pack true or false
        '''
        self.install_java_fp = value

    def get_install_external_java(self):
        '''
        get the install external java setting
        '''
        return self.install_external_java

    def get_install_java_fp(self):
        '''
        get the install java fix pack setting
        '''
        return self.install_java_fp

    def set_a_name_convention(self, name_value):
        '''
        Set name convention based on external Java or fix pack
        '''
        self.a_name_convention = name_value

    def set_java_parent_media_root_prop(self, parent_media_root_prop):
        '''
        Set Java parent folder name in the media server (WAS or IHS)
        '''
        self.a_media_root_prop = parent_media_root_prop

    def get_java_parent_media_root_prop(self):
        '''
        Get Java parent folder name in the media server
        '''
        return self.a_media_root_prop

    def set_java_content(self):
        '''
        Based on requested java version to determine if we should install external java
        '''
        java_version = self.get_requested_java_version()
        if 'bundled' in java_version:
            self.set_install_external_java(False)
        else:
            self.set_install_external_java(True)

    def get_remote_directory_location(self, **kwargs):  # from get_zip_location
        '''
        Get the source zip directory URL.

        - Extend method, and add JAVA under WAS in the media server
        - Assume WAS/JAVA media exist in the media server
        - The default is: media_server/$myMediaRoot/JAVA/$myVersion/$myOS

        :param kwargs: construct the src directory path, dictionary
            {"OSdependent": 'linux', "version": '8.5.5'}
        '''
        media_server = self.pget(self.media_server_prop)
        a_media_root = self.pget(self.a_media_root_prop)
        a_version = self.pget(self.a_version_prop)
        if a_media_root == 'IHS':
            a_version = 'bundled8'
            if qd.is_windows():
                a_version = a_version + 'x32'
        if "OSdependent" in kwargs:
            return qd.url_join(media_server, a_media_root, self.a_name_convention,
                               a_version, kwargs["OSdependent"])
        else:
            assert False, "unknown OSdependent did not pass in as argument "

    def get_local_directory_location(self, **kwargs):  # from get_zip_location
        '''
        Get the source zip directory file system path.

        - Extend method, and add JAVA under WAS in the media server,
        - Assume WAS/JAVA media exist in the media server
        - The default is: media_server/$myMediaRoot/JAVA/$myVersion/$myOS

        :param kwargs: construct the src directory path, dictionary
            {"OSdependent": 'linux', "version": '8.5.5'}
        '''
        media_root = self.pget(self.media_root_prop)
        a_media_root = self.pget(self.a_media_root_prop)
        a_version = self.pget(self.a_version_prop)
        if a_media_root == 'IHS':
            a_version = 'bundled8'
            if qd.is_windows():
                a_version = a_version + 'x32'
        if "OSdependent" in kwargs:
            return qd.path_join(media_root, a_media_root, self.a_name_convention,
                                a_version, kwargs["OSdependent"])
        else:
            assert False, "unknown OSdependent did not pass in as argument "


def get_im_components(cred, appmed=None, im_shared_dir=None, install_dir_common=None, javamed=None):
    '''
    Determine the flags that need to enable and replaced key and value.

    :param cred: The user and password that should start IHS service, this is a dictionary list
    :param appmed: media class (WASMedia, IHSMedia, JAVAMedia)
    :param im_shared_dir: IM shared directory path
    :param install_dir_common: media installation directory
    :param javamed: should set this parameter if WAS role is defined for the VM
    '''
    versions = appmed.get_formatted_version()
    if appmed.a_name_convention == 'was':
        java_versions = javamed.get_formatted_version()
        # pylint: disable=line-too-long
        result = [
            install_manager.IMRespCompo('_base', flags='WAS_BASE_REPO, WAS_FP_REPO',
                                        repl={'IM_SHARED': im_shared_dir,
                                              'WAS_INSTALL_DIR': qd.path_join(install_dir_common, qd_props.DEFAULT_REL_WAS_ROOT),
                                              'WAS_BASE_INSTALL_REPO': qd.path_join(qd.dirname(appmed.temp_unzip_dir), appmed.get_name_convention() + 'Base'),
                                              'WAS_FP_INSTALL_REPO': qd.path_join(qd.dirname(appmed.temp_unzip_dir), appmed.get_name_convention() + 'FP'),
                                              'MAJOR_DOT_VER': versions['dot_ver'],
                                              'MAJOR_VER': versions['no_dot_ver'],
                                              'JAVA_NO_DOT_VER': java_versions['no_dot_ver'],
                                              'JAVA_RELEASE': java_versions['major_rel_ver']})
        ]
        # pylint: enable=line-too-long
        # if wasJavaSDKVersion requested is not bundled WAS, then we should install external java
        if javamed.get_install_external_java():
            result += [install_manager.IMRespCompo('_base', flags='WAS_JAVA_PROFILE_DATA')]
            result += [install_manager.IMRespCompo('_base', flags='JAVA_BASE_REPO',
                                                   repl={'JAVA_BASE_INSTALL_REPO':
                                                         qd.path_join(
                                                             qd.dirname(javamed.temp_unzip_dir),
                                                             javamed.a_name_convention_root)})]
        else:
            result += [install_manager.IMRespCompo('_base', flags='WAS_PROFILE_DATA',
                                                   repl={'bundled':
                                                         javamed.get_requested_java_version()[-1:]
                                                         })]
    else:
        # pylint: disable=line-too-long
        result = [
            install_manager.IMRespCompo('_base', flags='IHS_BASE_REPO, IHS_FP_REPO, IHS_PROFILE_DATA',
                                        repl={'IM_SHARED': im_shared_dir,
                                              'HTTP_INSTALL_DIR': qd.path_join(install_dir_common, qd_props.DEFAULT_REL_IHS_HOME),
                                              'PLUGINS_INSTALL_DIR': qd.path_join(install_dir_common, qd_props.DEFAULT_REL_PLUGINS_HOME),
                                              'IHS_BASE_INSTALL_REPO': qd.path_join(qd.dirname(appmed.temp_unzip_dir), appmed.get_name_convention() + 'Base'),
                                              'IHS_FP_INSTALL_REPO': qd.path_join(qd.dirname(appmed.temp_unzip_dir), appmed.get_name_convention() + 'FP'),
                                              'MAJOR_DOT_VER': versions['dot_ver'],
                                              'MAJOR_VER': versions['no_dot_ver']})
        ]
        # pylint: enable=line-too-long
        if qd.is_linux():
            result += [install_manager.IMRespCompo('_base', flags='LINUX_SERVICE',
                                                   repl={'ARCH_BITS': '64'})]
        else:
            result += [install_manager.IMRespCompo('_base', flags='WINDOWS_SERVICE',
                                                   repl={'ARCH_BITS': '32',
                                                         'LOGON_OS_USER': cred['user'],
                                                         'LOGON_OS_PASSWORD': cred['user_pwd']})]
    if javamed.get_install_java_fp():
        result += [install_manager.IMRespCompo('_base', flags='JAVA_FP_REPO',
                                               repl={'JAVA_FP_INSTALL_REPO':
                                                     qd.path_join(
                                                         qd.dirname(javamed.temp_unzip_dir),
                                                         javamed.a_name_convention_root +
                                                         'FP')})]

    return result


def create_was_profile(env_props, os_user, os_user_pwd):
    '''
    Create WebSphere profile with default profile name

    :param env_props: env.properties instance
    :param os_user: the user to run manageprofile script
    :param os_user_pwd: os_user password
    '''
    # Generate profile xml file
    qd_root_local = env_props.get_values(qd_props.PN_QD_ROOT)
    profile_template_file = resource.get_resource('InstallWAS/profile_template.response')

    qd.QDEnv.set_value_to_prop_file(profile_template_file, 'profileName',
                                    qd_props.DEFAULT_REL_WAS_PROFILE_NAME)
    qd.QDEnv.set_value_to_prop_file(profile_template_file, 'templatePath',
                                    qd.path_join(qd_root_local, qd_props.DEFAULT_REL_WAS_ROOT,
                                                 'profileTemplates', 'default'))
    qd.QDEnv.set_value_to_prop_file(profile_template_file, 'hostName', socket.gethostname())

    # Run manageprofiles script to create WAS profile
    manage_profile_exe = qd.os_switch('manageprofiles.sh', 'manageprofiles.bat')
    manage_profile_path = qd.path_join(qd_root_local, qd_props.DEFAULT_REL_WAS_ROOT, 'bin',
                                       manage_profile_exe)
    cmd = [manage_profile_path, '-response', profile_template_file]
    install_util.su_call_if_necessary('--', (os_user, os_user_pwd), cmd, call_method=qd.check_call)
    # Write WAS_PROFILE_ROOT variable into env.properties
    env_props.set_value('WAS_PROFILE_ROOT', qd.path_join(qd_root_local,
                                                         qd_props.DEFAULT_REL_WAS_PROFILE))


def set_default_jdk(javamed, profile_path, os_user, os_user_pwd):
    '''
    Set default JDK in WAS

    :param profile_path: parent directory of managesdk.sh
    :param sdk: sdk name that passed in the managesdk script as argument
    :param os_user: the user name to run the managesdk script
    :param os_user_pwd: the password of the os_user
    '''
    # set Java SDK requested by the user
    jdk_ver = javamed.get_formatted_version()['jdk_ver']
    sdkname = '1.' + jdk_ver + '_64'
    if float(jdk_ver) >= 7.1:
        server_exe = qd.os_switch('managesdk.sh', 'managesdk.bat')
        cmd = [qd.path_join(profile_path, server_exe),
               '-enableProfile', '-profileName', qd_props.DEFAULT_REL_WAS_PROFILE_NAME,
               '-sdkname', sdkname, '-enableServers']
        install_util.su_call_if_necessary('--', (os_user, os_user_pwd), cmd,
                                          call_method=qd.check_call)
        qd.echo_info("Check default java sdkname")
        cmd = [qd.path_join(profile_path, server_exe),
               '-listEnabledProfile', '-profileName', qd_props.DEFAULT_REL_WAS_PROFILE_NAME,
               '-verbose']
        install_util.su_call_if_necessary('--', (os_user, os_user_pwd), cmd,
                                          call_method=qd.check_call)


def post_ihs_install(env_props, appmedia, os_user, os_grp):
    '''
    IHS Post configuration after installation.

    - create IHS service
    - modify admin.conf
    - start HTTP server

    :param env_props: instance of env.properties
    :param appmedia: the media instance
    :param os_user: The user name that should run the command
    :param os_user_pwd: the password of the os_user
    '''
    # create IHS service
    ihs_home = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                            qd_props.DEFAULT_REL_IHS_HOME)
    ihs_conf_home = qd.path_join(ihs_home, 'conf')
    if qd.is_linux():
        ibm_ihs_file = resource.get_resource('InstallWAS/ibm-ihs')
        qd.sed_in_place(ibm_ihs_file, [('_IHS_HOME_', ihs_home)])
        qd.create_service('ibm-ihs', os.path.dirname(ibm_ihs_file))
        # Setting up IBM HTTP Server for starting/stopping from WAS console
        # only need to run "setupadm" if you want to stop the IHS from WAS console
        qd.call([qd.path_join(ihs_home, 'bin', 'setupadm'), '-usr', os_user, '-grp', os_grp,
                 '-cfg', qd.path_join(ihs_conf_home, 'httpd.conf'),
                 '-adm', qd.path_join(ihs_conf_home, 'admin.conf')])
    qd.sed_in_place(qd.path_join(ihs_conf_home, 'admin.conf'), [('@@AdminPort@@', '8008')])

    start_ihs_server(ihs_home, appmedia.get_formatted_version()['dot_ver'])


def start_ihs_server(ihs_home, service_version):
    '''
    Start IHS server process

    :param ihs_home: the IBM HTTP Server home path
    :param service_version: IBM HTTP Server version (8.5 etc)
    '''
    if qd.is_linux():
        # start HTTP server
        qd.echo_info("Starting http server processes")
        qd.check_call([qd.path_join(ihs_home, 'bin', 'adminctl'), 'start'])
        qd.check_call([qd.path_join(ihs_home, 'bin', 'apachectl'), 'start'])
    else:
        # Construct the HTTP server windows service name
        service_name = 'IBMHTTPServerV' + service_version
        qd.check_call(['net', 'start', service_name])


def post_was_install(env_props, appmedia, os_user, os_user_pwd):
    '''
    WAS Post configuration after installation.

    - create WAS profile
    - create WAS service
    - start WAS application server

    :param env_props: instance of env.properties
    :param appmedia: the media instance
    :param os_user: The user name that should run the command
    :param os_user_pwd: the password of the os_user
    '''
    # create profile
    create_was_profile(env_props, os_user, os_user_pwd)
    # Create was service on Linux
    if qd.is_linux():
        ibm_was_file = resource.get_resource('InstallWAS/ibm-was')
        qd.sed_in_place(ibm_was_file,
                        [('_WAS_ROOT_', qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                                                     qd_props.DEFAULT_REL_WAS_ROOT)),
                         ('_WAS_INSTALL_USER_', os_user)])
        qd.create_service('ibm-was', os.path.dirname(ibm_was_file))

        # Start WebSphere application server
        cmd = [qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                            qd_props.DEFAULT_REL_WAS_PROFILE, 'bin',
                            qd.os_switch('startServer.sh', 'startServer.bat')), 'server1']
        install_util.su_call_if_necessary('--', (os_user, os_user_pwd), cmd,
                                          call_method=qd.check_call)
    else:
        # construct the windows service name
        service_name = "IBMWAS" + appmedia.get_formatted_version()['no_dot_ver'] + "Service - "
        service_name += socket.gethostname() + "Node01"
        obj_name = ''.join(['obj=.\\', os_user])
        obj_password = ''.join(['password=', os_user_pwd])
        update_sc_cmd = ['sc', 'config', service_name, obj_name, obj_password]
        qd.echo_info("update_sc_cmd: [%s]" % update_sc_cmd)
        qd.check_call(update_sc_cmd)
        cmd = ['net', 'start', service_name]
        install_util.su_call_if_necessary('--', (os_user, os_user_pwd), cmd,
                                          call_method=qd.check_call)


# pylint: disable=too-many-locals
# pylint: disable=too-many-statements
def main():
    '''Main method; see module documentation.'''
    # set up local environment for this script
    working_dir = os.getcwdu()
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)  # get credentials
    media_props = qd_props.get_media_props(env_props)  # get media properties

    vm_role = env_props.get_values('WASRole')
    # === prepare media
    if vm_role == 'ihs':
        ihsmed = IHSMedia(media_props)
        ihsmed.prepare_media()
        javamed = JAVAMedia(media_props)
        javamed.set_java_parent_media_root_prop(PN_IHS_MEDIA_ROOT)
        # Only use bundled Java for IHS server
        javamed.set_install_external_java(False)
        javamed.prepare_media()
    elif vm_role == 'was':
        wasmed = WASMedia(media_props)
        wasmed.prepare_media()
        javamed = JAVAMedia(media_props)
        javamed.set_java_parent_media_root_prop(PN_WAS_MEDIA_ROOT)
        javamed.prepare_media()
    elif vm_role == 'liberty':
        qd.echo_info("Liberty will be installed on InstallApp step")
        exit(0)
    else:
        assert False, 'Unsupported WASRole is set in env.properties, Neither ihs nor was.'

    if qd_props.load_create_media_only(media_props):
        qd.echo_info("media server update done")
        qd.exit_(0)

    # generate im_xml
    was_template_file = resource.get_resource('InstallWAS/was_response_template.xml')

    qd_root, im_shared_dir, im_location = env_props.get_values(qd_props.PN_QD_ROOT,
                                                               qd_props.PN_IM_SHARED_DIR_PREP,
                                                               qd_props.PN_IM_ECLIPSE_DIR_PREP)
    user_name, user_password, primary_group = qd_props.get_os_users_creds_and_group(user_props)
    cred = {'user': user_name, 'user_pwd': user_password}

    if vm_role == 'was':
        im_components = get_im_components(cred, wasmed, im_shared_dir, qd_root, javamed)

        was_root = qd.path_join(qd_root, qd_props.DEFAULT_REL_WAS_ROOT)
        env_props.set_value('WAS_ROOT', qd.path_join(qd_root, was_root))
        # need to create {was_root_parent_dir} folder so IM can install into it
        was_root_parent_dir = qd.dirname(was_root)
        install_util.ensure_folder_and_set_owner(was_root_parent_dir, user_name, primary_group)
        qd.chmod_777(was_root_parent_dir, '755')
    else:
        im_components = get_im_components(cred, ihsmed, im_shared_dir, qd_root, javamed)
        # need to create {ihs_root} folder so IM can install into it
        ihs_root = qd.path_join(qd_root, qd_props.DEFAULT_REL_IHS_HOME)
        env_props.set_value('IHS_ROOT', qd.path_join(qd_root, ihs_root))
        install_util.ensure_folder_and_set_owner(ihs_root, user_name, primary_group)
        qd.chmod_777(ihs_root, '755')
        # need to create {plugins_root} folder so IM can install into it
        plugins_root = qd.path_join(qd_root, qd_props.DEFAULT_REL_PLUGINS_HOME)
        env_props.set_value('Plugins_ROOT', qd.path_join(qd_root, plugins_root))
        install_util.ensure_folder_and_set_owner(plugins_root, user_name, primary_group)
        qd.chmod_777(plugins_root, '755')

    flags, replacements = install_manager.get_im_flags_and_replacements(im_components)
    install_manager.generate_imxml(was_template_file, qd.path_join(working_dir, 'IMresponses.xml'),
                                   replacements, flags)

    if qd_props.load_debug_do_not_install(media_props):
        qd.echo_debug(
            'XML response file created but exiting prior to installing applications')
        qd.exit_(0)

    # Run the installation via installation manager
    qd.echo_info("Install to the system as [%s], using Installation Manager response file [%s]"
                 % (user_name, qd.path_join(working_dir, 'IMresponses.xml')))
    install_manager.execute_im(qd.path_join(working_dir, 'IMresponses.xml'),
                               im_location=im_location,
                               im_usercreds=(user_name, user_password))

    # ==== Create WAS profile
    if vm_role == 'was':
        post_was_install(env_props, wasmed, user_name, user_password)
        # set default JDK
        if javamed.install_external_java:
            set_default_jdk(javamed, qd.path_join(qd_root, qd_props.DEFAULT_REL_WAS_ROOT, 'bin'),
                            user_name, user_password)
        qd.echo_info("IBM WebSphere Application Server is installed successfully")
        # tidy up by deleting anything downloaded or unzipped
        wasmed.cleanup()
    else:
        post_ihs_install(env_props, ihsmed, user_name, primary_group)
        qd.echo_info("Installation IBM HTTP Server is installed successfully")
        ihsmed.cleanup()
    # === finished ====
    qd.echo_info("SUCCESS: installation complete")


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main()
    exit(0)
