#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Run JTS setup.  See :func:`main` function.

Usage::

    setup.py <RunSetup> <MTM> <ImportUser>
    setup.py Y Y N


'''
import os
import re
import shutil
import string
import sys
import time
import subprocess

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from JTSSetup import setup_response
from ConfigureAppServer import liberty_eval
from ConfigureWAS.WASCommon import single_was_start_stop
from ConfigureLiberty import liberty_start_stop
from qdlib import qd, qd_topo, qd_props, install_util


def split_liberty(jts_dir, build_response_file, run_setup='Y'):
    '''
    Split Liberty into two instances

    :param jts_dir: JazzTeamServer directory path
    :param build_response_file: Setup Response file
    '''
    path = qd.path_join(jts_dir, 'liberty/servers/clm0')
    if not os.path.isdir(path):
        liberty_eval.main()
    else:
        qd.echo_info("Folder [%s] located" % path)
        qd.echo_info("Assuming Liberty is configured")
    if run_setup == 'Y':
        qd.sed_in_place(build_response_file, [('9443/lqe', '9444/lqe')])
        qd.sed_in_place(build_response_file, [('9443/relm', '9444/relm')])
        qd.sed_in_place(build_response_file, [('9443/dcc', '9444/dcc')])
        qd.sed_in_place(build_response_file, [('9443/rs', '9444/rs')])


# pylint: disable=too-many-branches,too-many-locals
def main(argv):
    '''
    Main code for running JTS setup.

    :param RunSetup: Y if setup (using repotools-jts -setup) should be run; N otherwise
    :param MTM: Y if MTM sample should be created; N otherwise.  Sample only created if
                setup requested and successful.
    :param ImportUser: Y if LDAP users should be imported; N otherwise.  LDAP users only
                       imported if setup requested and successful.
    '''

    if len(argv) - 1 != 3:
        qd.echo_error("Script [{scriptname}] has wrong input parameters [%d]"
                      % (len(argv) - 1))
        qd.echo_info("should have 3 parameters :: run_setup=Y/N mtm=Y/N import_user=Y/N")
        qd.exit_(1)

    run_setup = argv[1]
    mtm = argv[2]
    import_user = argv[3]
    qd.echo_info('Run JTS Setup [%s]' % run_setup)
    qd.echo_info('set up MTM [%s]' % mtm)
    qd.echo_info('Import LDAP Users [%s]' % import_user)

    env_props = qd_props.get_default_env_props()
    jts_dir = qd.path_join(env_props.get_values(qd_props.PN_CLM_HOME_PREP), 'server')

    all_tag_listing = env_props.get_values(qd_props.UCD_PN_GLOBAL_TAG_LISTING)
    topo = qd_topo.QDTopo(all_tag_listing)

    # Ensure all other WAS nodes completed copy in previous step,
    # so we can unshare JTS node folder at last
    if qd.is_windows():
        if topo.get_user_registry_type_tmp() != 'DETECT':  # No need to run on evaluation Topology
            try:
                cmd = ['net', 'share', qd_props.DEFAULT_REL_TMP_DIR, '/Y', '/DELETE']
                qd.check_call(cmd)
            except (subprocess.CalledProcessError, WindowsError):
                qd.echo_info("Share already deleted Continue")

    jts_setup_dir = os.getcwdu()
    env_props.set_value('JTSSetup_DIR', jts_setup_dir)  # necessary??
    build_response_file = qd.path_join(jts_setup_dir, 'SetupResponse.properties')

    #  user may not want to run the scripting
    if run_setup != 'Y':  # was == 'N'
        qd.echo_info("User has selected to NOT run the automated scripted JTS Setup.")
        qd.echo_info("User will now be required to MANUALLY complete JTS Setup")
        if topo.get_user_registry_type_tmp() == 'DETECT':
            split_liberty(jts_dir, build_response_file, run_setup)
        return

    env_props.set_value('JTSURLHost', topo.get_jts_url_host())

    local_host = env_props.get_values('localHostname', localHostname=qd.get_host_name())
    local_server_role = topo.get_host(local_host).get_local_server_role()
    assert local_server_role in ['was', 'liberty']  # as this hosts JTS
    assert 'jts' in topo.get_host(local_host).tags  # run in JTS

    ################################
    # -- get the WAS properties -- #
    ################################
    # get credentials
    user_props = qd_props.get_user_props(env_props)
    os_user_name, os_user_password = qd_props.get_os_users_creds(user_props)
    if local_server_role == 'was':
        # Create local WAS ADMIN PASSWORD the installation
        was_user_name = user_props.get_values('WAS_USER_NAME')  # repotools user
        temp_password = qd.decode_base64(user_props.get_values('WAS_PASSWORD'))
    else:
        was_user_name = user_props.get_values('WAS_USER_NAME')  # repotools user
        temp_password = qd.decode_base64(user_props.get_values('WAS_PASSWORD'))

    #################################
    # -- get the LDAP properties -- #
    #################################
    if topo.get_user_registry_type_tmp() == 'LDAP':
        # -- get the LDAP properties -- #
        ldap_props = qd_props.get_ldap_props(env_props)

    #############################
    # ---- nodes to process --- #
    #############################
    qd.echo_info("NUMBER_OF_CUSTOM_NODES to process [%d]" % len(topo.get_app_hosts()))

    #########################
    #  BUILD RESPONSE FILE  #
    #########################
    builder = setup_response.SetupResponseBuilder(env_props, topo)
    builder.was_user_name = was_user_name
    builder.was_password = temp_password
    builder.build(build_response_file)

    # only enable MTM if required
    if mtm == 'Y':
        # skip MTM if AM or DM is present
        if topo.get_hosts_by_tags(qd_topo.TAG_AM):
            qd.echo_warning("AM present so cannot Deploy MTM Sample")
        elif topo.get_hosts_by_tags(qd_topo.TAG_DM):
            qd.echo_warning("DM present so cannot Deploy MTM Sample")
        else:
            qd.echo_info("Deploy MTM Sample")
            qd.sed_in_place(build_response_file, [('MTM=.*$', 'MTM=true')])
    else:
        qd.echo_info("not deploying MTM")

    ########################
    # confirm where we are #
    ########################
    qd.echo_info(os.getcwd())
    qd.chmod_777(build_response_file, mode='755')
    qd.call(qd.os_switch('ls -la', 'dir'), shell=True)

    #########################################################
    # update JTS teamserver.properties file to allow access #
    #########################################################
    # save current teamserver.properties file and update to allow access
    shutil.copy(qd.path_join(jts_dir, 'conf/jts/teamserver.properties'),
                qd.path_join(jts_dir, 'conf/jts/teamserver.properties.bak'))
    # change com.ibm.team.repository.ws.allow.admin.access=false to true
    qd.sed_in_place(qd.path_join(jts_dir, 'conf/jts/teamserver.properties'),
                    [('allow.admin.access=.*$', 'allow.admin.access=true')])

    ###################
    # Restart jts_war #
    ###################
    # pylint: disable=misplaced-bare-raise
    if local_server_role == 'was':
        # dynamic load the module
        single_was_start_stop.main(['restart'])
    elif local_server_role == 'liberty':
        if topo.get_user_registry_type_tmp() == 'DETECT':
            split_liberty(jts_dir, build_response_file)
        else:
            liberty_start_stop.restart_liberty()
    else:
        raise  # never

    def run_repotools_jts(options):
        ''' Run jts repotools.'''

        new_env = os.environ.copy()
        if topo.get_user_registry_type_tmp() == 'LDAP':
            new_env['jazz_primaryid'] = ldap_props.get_values('jazz_ldap_primaryid')
            new_env['jazz_primaryid_Password'] = temp_password
        else:
            new_env['jazz_primaryid'] = 'ADMIN'
            new_env['jazz_primaryid_Password'] = 'ADMIN'
        if env_props.get_values(qd_props.PN_DB_VENDOR) == 'oracle':
            # pylint: disable=line-too-long
            new_env['ORACLE_JDBC_DRIVER_FILE'] = qd.path_join(jts_dir, 'oracle',
                                                              env_props.get_values(qd_props.PN_DB_DRIVER_FILE))
        if env_props.get_values(qd_props.PN_DB_VENDOR) == 'sqlserver':
            # pylint: disable=line-too-long
            new_env['SQLSERVER_JDBC_DRIVER_FILE'] = qd.path_join(jts_dir, 'sqlserver',
                                                                 env_props.get_values(qd_props.PN_DB_DRIVER_FILE))
            # pylint: enable=line-too-long

        user_options = ['adminUserId=$jazz_primaryid',
                        'adminPassword=$jazz_primaryid_Password',
                        'repositoryURL=https://%s:9443/jts' % topo.get_jts_url_host()]

        # if DM present increase timeout
        # pylint: disable=line-too-long
        if topo.get_hosts_by_tags(qd_topo.TAG_DM):
            qd.echo_info("Increasing timeout to 40 minutes because DM is present")

            # timeout specified in milliseconds
            if qd.is_windows():
                qd.echo_debug("Set timeout on DEFINE_INIT when Windows")
                # add the set to the command on windows because win_su does not support env tuple
            else:
                qd.echo_debug("Set timeout on VMARGS when Linux")
                new_env['VMARGS'] = '-Dcom.ibm.team.repository.common.transport.TeamServerConfiguration.socketTimeout=2400000'

            jts_setup_cmd = qd.os_switch(['./repotools-jts.sh'],
                                         ['set', 'DEFINE_INIT=-Dcom.ibm.team.repository.common.transport.TeamServerConfiguration.socketTimeout=2400000', '&', 'cmd', '/c', 'repotools-jts.bat']) + options + user_options

        else:
            jts_setup_cmd = qd.os_switch(['./repotools-jts.sh'],
                                         ['cmd', '/c', 'repotools-jts.bat']) + options + user_options
        # pylint: enable=line-too-long

        if qd.is_windows():
            jts_setup_cmd = [string.Template(e).substitute(new_env) for e in jts_setup_cmd]

        qd.echo_info('jts_setup_cmd: %s' % jts_setup_cmd)

        return install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                                 jts_setup_cmd, env=new_env, cwd=jts_dir)

    #################
    # RUN JTS SETUP #
    #################
    qd.echo_info("run jts setup")

    # Output the response file (minus password lines) for debug info.
    qd.print_file_skip_lines(build_response_file, 'password')

    jts_setup_options = ['-setup', 'includeLifecycleProjectStep=true', 'parametersFile=%s'
                         % build_response_file]
    if run_repotools_jts(jts_setup_options) != 0:
        qd.echo_error("JTS Setup Wizard Automation failed")
        qd.exit_(1)
    else:
        qd.echo_info("------JTS Setup Wizard Completed------")

    # confirm it completed successfully by parsing the repotools-jts_setup.log
    setup_log = qd.path_join(jts_dir, 'repotools-jts_setup.log')

    def has_errors_in_log():
        ''' Check for errors in the log. '''

        #  "The bundle com.ibm.team.jfs.app.error is INSTALLED." may be in the log
        result = False
        error_lines = qd.grep(setup_log, 'error', ignore_case=True)
        if error_lines:
            qd.echo_info('[%d] lines in the log contains "error"' % len(error_lines))
            for error_line in error_lines:
                if re.search(r'\w+\.\w+\.\w+\.error', error_line, re.I):
                    qd.echo_info(' False error (bundle): %s' % error_line.strip())
                else:
                    qd.echo_error('Error in _setup.log : %s' % error_line.strip())
                    result = True
        return result

    if has_errors_in_log():
        # might have captured an error though
        qd.echo_error("JTS Setup Wizard Automation failed: error(s) in log")
        qd.exit_(1)
    elif not qd.grep(setup_log, "Setup completed successfully"):
        qd.echo_error("JTS Setup Wizard Automation failed: "
                      "no success message in the log")
        qd.exit_(1)

    qd.echo_info("------JTS Setup Wizard Successful------")

    #####################
    # IMPORT LDAP USERS #
    #####################
    if import_user == 'Y':
        qd.echo_info("Sending request for immediate LDAP user synchronization")
        sync_users_options = ['-syncUsers', 'logFile=repotools-jts_syncUsers.log']
        if run_repotools_jts(sync_users_options) != 0:
            qd.echo_error("Request failed")
            qd.exit_(1)
        else:
            qd.echo_info("Request successfully sent")
            time.sleep(300)
    else:
        qd.echo_info("No LDAP user synchronization request sent; "
                     "users will be synchronized as per system schedule")

    ###################
    # --- the End --- #
    ###################
    qd.echo_info("-- EOF [{scriptname}] --")


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main(sys.argv)
    exit(0)
