#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Build up JTS setup response file
'''
import os
import string
import sys

# find qdlib from parent paths and add the first one to the Python search path
sys.path.append(next((p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and
                      (p + '/qdlib') not in sys.path), None))

from qdlib import qd, qd_props


class SetupResponseBuilder(object):
    '''
    JTS setup response file builder class
    '''

    def __init__(self, env_props, topo):
        self.env_props = env_props
        self.topo = topo
        self.was_user_name = None
        self.was_password = None
        self.file_pointer = None

    def _echo_from_template(self, template, mapping=None):
        '''
        Loop through each line of the template; substitute the value for any key found in the
        mapping.

        :param template: A list of response entries
        :param mapping: A list of key-value pairs
        '''
        if mapping is None:
            mapping = {}
        lines = [string.Template(t).substitute(**mapping) for t in template]
        for line in lines:
            self.file_pointer.write(line + '\n')

    def build(self, build_response_file=None):
        '''
        Open a file handler to build the setup response entries

        :param build_response_file: response file name
        :return: entire content of the build_response_file
        '''
        def open_target():
            '''
             Open a file handler
            '''
            if not build_response_file:
                import StringIO
                return StringIO.StringIO()
            else:
                return open(build_response_file, 'wt')
        with open_target() as file_pointer:
            try:
                self.file_pointer = file_pointer
                self.create_setup_response()
                if not build_response_file:
                    return file_pointer.getvalue()
            finally:
                self.file_pointer = None

    def create_setup_response(self):
        '''
        Create all sections in the response file

        - Common section
        - License section
        - LDAP user section
        - Database connection section
        '''
        self.create_common_section()
        self.create_license_section()
        self.create_user_section()
        self.create_database_section('jts')

        # add apps
        non_local_app = 0
        for host in self.topo.get_app_hosts():
            host_name = self.topo.get_hostname(host)

            apps = host.get_local_apps()
            if 'jts' not in apps:
                qd.echo_info('This is not a jts node')
                qd.echo_info('apps: [%s]' % apps)
                for app in apps:
                    non_local_app += 1
                    self.create_database_section(app, non_local_app)
                    self.create_register_app_section(host_name, app, non_local_app)
                    self.create_special_section(app, non_local_app)
            else:
                qd.echo_info('apps: [%s]' % apps)
                for app in apps:
                    if app == 'jts':
                        qd.echo_info("Skip jts app, it should automatically added")
                        continue
                    self.create_database_section(app)
                    self.create_register_app_section(host_name, app)
                    self.create_special_section(app)

        # add number of apps
        # pylint: disable=line-too-long
        self._echo_from_template([r'register.applications.page\:jts.nonlocal.application.number=$nonlocalApp'],
                                 {'nonlocalApp': str(non_local_app)})
        # pylint: enable=line-too-long

    def create_common_section(self):
        '''
        Common section for all products
        '''
        qd.echo_info('Build common section...')
        template = [r'admin.access\:ws.allow.admin.access=true',
                    r'com.ibm.team.lpa.deploySampleStep\:lpa.deploy.MTM=false',
                    r'configure.email.page\:notification.mail.enabled=false']

        # configure public URI for Liberty
        if self.topo.get_jts_host().get_local_server_role() == 'liberty':
            template += [r'configure.public.uri.page\:server.webapp.url=https\://%s\:9443/jts'
                         % self.topo.get_jts_url_host(),
                         r'configure.public.uri.page\:server.webapp.url.confirm=true']

        # dump
        self._echo_from_template(template, {})  # no mapping

    def create_register_app_section(self, host_name, app_name, app_number=0):
        '''
        Application registration section

        :param host_name: The host name to access the application
        :param app_name: The application name e.g. jts, ccm
        :param app_number: The number of applications hosted on the server other than JTS server
        '''
        mapping = {'HostName': host_name, 'a': app_name, 'nonlocalApp': app_number}
        # pylint: disable=line-too-long
        if app_number > 0:
            # non local

            template = [r'$nonlocalApp\:register.applications.page\:application.discovery.url=https\://$HostName\:9443/$a/scr',
                        r'$nonlocalApp\:register.applications.page\:application.title=/$a',
                        r'$nonlocalApp\:register.applications.page\:application.user.id=${a}_user']
            if app_name == 'lqe':
                template += [r'lqe.$nonlocalApp\:configure.trs.page\:trs.add=AddTRS']
        else:
            template = [r'register.applications.page\:jts.register.$a=true',
                        r'register.applications.page\:$a.discovery.url=https\://$HostName\:9443/$a/scr',
                        r'register.applications.page\:$a.application.title=/$a']
            if app_name == 'lqe':
                template += [r'lqe\:configure.trs.page\:trs.add=AddTRS']
        # pylint: enable=line-too-long

        # dump
        self._echo_from_template(template, mapping)

    ###
    # USER SECTION
    ###
    def create_user_section(self):
        '''
        LDAP admin user name and password section
        '''
        mapping = {}
        if self.topo.get_user_registry_type_tmp() == 'LDAP':
            template = self._get_ldap_user_section_template(mapping)
        else:
            template = self._get_detect_user_section_template(mapping)
            assert self.was_user_name
            assert self.was_password
            mapping['was_user_name'] = self.was_user_name
            mapping['was_password'] = self.was_password

        # dump
        self._echo_from_template(template, mapping)  # no mapping

    @classmethod
    def _get_detect_user_section_template(cls, mapping):
        # Disable pylint invalid-name because function name is > 30 chars and want to keep it
        # Disable pylint unused-argument  to keep method consistent
        # pylint: disable=invalid-name,unused-argument
        '''
        Detect user information section

        :param mapping: A list of key-value pairs
        :return: A list of user entries in the response file
        '''
        return [r'user.registry.page\:user.registry.type=DETECT',
                r'create.user.page\:emailAddress=ibmqd@example.com',
                r'create.user.page\:name=IBMQD',
                r'create.user.page\:password=$was_password',
                r'create.user.page\:userId=$was_user_name']

    def _get_ldap_user_section_template(self, mapping):
        '''
        LDAP configuration section

        :param mapping: A list of key-value pairs
        :return: A list of LDAP configuration entries in the response file
        '''
        ldap_props = qd_props.get_ldap_props(self.env_props)
        mapping.update(ldap_props.get_dict())

        return [r'compute.LDAP.queries.step\:compute.LDAP.queries=true',
                r'user.registry.page\:ldap.baseGroupDN=$jazz_ldap_baseGroupDN',
                r'user.registry.page\:ldap.baseUserDN=$jazz_ldap_baseUserDN',
                r'user.registry.page\:ldap.groupMapping=$jazz_ldap_groupMapping',
                r'user.registry.page\:ldap.groupNameAttribute=$jazz_ldap_groupNameAttribute',
                r'user.registry.page\:ldap.membersOfGroup=$jazz_ldap_membersOfGroup',
                r'user.registry.page\:ldap.registryLocation=ldap\://$jazz_userrealm',
                r'user.registry.page\:ldap.registryPassword=$jazz_userrole_RegistryPassword',
                r'user.registry.page\:ldap.registryUserName=$jazz_userrole_RegistryUsername',
                r'user.registry.page\:ldap.userAttributesMapping=$jazz_ldap_userAttributesMapping',
                r'user.registry.page\:user.registry.type=LDAP']

    ###
    # LICENSE SECTION
    ###
    def create_license_section(self):
        '''
        License section

        :return: A list of license information entries in the response file
        '''
        all_apps = self.topo.get_all_apps_list()
        activate_trial = True

        def create_single_license(app, licence_key, license_id, active=None):
            '''
            Args:
                app: True to activate license, False not to activate,
                     or app tag to activate only when the app is deployed in the topology.
                     No license is activated when 'activate_trial' flag is False
                licenase_key: LVALUE e.g. CLMServerMonitoring
                license_id: license id to assing e.g. com.ibm.team.csm.user
                active: True or False to activate license regardless of the activate_trial flag.
            '''

            if active is None:
                if isinstance(app, bool):
                    active = app
                else:
                    active = (app in all_apps)

            if active and activate_trial:
                assert license_id
                return [r'license.assignment.page\:%s.activate.trial=true' % license_id,
                        r'license.assignment.page\:%s.agreement=true' % license_id,
                        r'license.assignment.page\:license.%s=%s' % (licence_key, license_id)]
            else:
                return [r'license.assignment.page\:license.%s=%s' % (licence_key, 'None')]

        # add fixed license information
        # pylint: disable=line-too-long
        template = []
        template += create_single_license(True, 'CLMServerMonitoring', 'com.ibm.team.csm.user')
        template += create_single_license(False, 'RationalsolutionforCollaborativeLifecycleManagement', 'com.ibm.team.clm.practitioner')

        # add optional license information
        template += create_single_license('dm', 'RationalRhapsodyDesignManager', 'com.ibm.xtools.dm.rhapsody.manager')
        template += create_single_license('relm', 'RationalEngineeringLifecycleManager', 'com.ibm.team.relm.user')
        template += create_single_license('qm', 'RationalQualityManager', 'com.ibm.rqm.tester')
        # template += create_single_license('ccm', 'RationalTeamConcert', 'com.ibm.team.rtc.developer-iep')  # developer-iep
        template += create_single_license('ccm', 'RationalTeamConcert', 'com.ibm.team.rtc.developer')
        template += create_single_license('rm', 'RationalDOORSNextGeneration', 'com.ibm.team.rrc.author')
        # pylint: enable=line-too-long

        # dump
        self._echo_from_template(template, {})  # no mapping

    ###
    # Special section for specific application
    ###
    def create_special_section(self, app_name, app_number=0):
        '''
        Special settings for a specific application

        :param app_name: Application name, e.g. jts, ccm
        :param app_number: The number of applications hosted on the server other than JTS server
        '''
        mapping = self.env_props.get_dict()
        mapping['EntryHead'] = (app_name if not app_number
                                else ('%s.%d' % (app_name, app_number))) + '\\:'

        if app_name == 'ldx' or app_name == 'lqe':
            template = [r'${EntryHead}configure.backup.page\:backup.enabled=false']
            self._echo_from_template(template, mapping)

    ###
    # DATABASE SECTION
    ###
    def create_database_section(self, app_name, app_number=0):
        '''
        Database configuration section

        :param app_name: Application name. e.g. jts, ccm
        :param app_number: The number of applications hosted on the server other than JTS server
        '''
        mapping = self.env_props.get_dict()
        mapping['appName'] = app_name
        mapping['DBHost'] = self.topo.get_db_hostname()

        db_vendor = self.env_props.get_values('DB_VENDOR')
        # pylint: disable=misplaced-bare-raise
        if db_vendor == 'derby':
            template = self._get_derby_section_template(app_name, mapping)
        elif db_vendor == 'db2':
            template = self._get_db2_section_template(app_name, mapping)
        elif db_vendor == 'oracle':
            template = self._get_oracle_section_template(app_name, mapping)
        elif db_vendor == 'sqlserver':
            template = self._get_sql_section_template(app_name, mapping)
        else:
            raise

        if app_name == 'jts':
            assert not app_number  # JTS => not app_number
            mapping['EntryHead'] = ''
        else:
            mapping['EntryHead'] = (app_name if not app_number
                                    else ('%s.%d' % (app_name, app_number))) + '\\:'

        # dump
        self._echo_from_template(template, mapping)

    @classmethod
    def _get_derby_section_template(cls, app_name, mapping):
        '''
        Derby configuration section

        :param app_name: Application name
        :param mapping: A list of key-value pairs
        :return: A list of derby database configuration
        '''

        mapping['derby_net_dw_port'] = '1527'

        # [rs] does not have a database. all other apps do
        # pylint: disable=line-too-long
        if app_name == 'rs':
            template = [r'${EntryHead}com.ibm.team.jrs.setupStep\:db.jdbc.password=']
        elif app_name in ['lqe', 'ldx']:
            template = [r'${EntryHead}configure.database.page\:db.location=conf/${appName}/derbyDB',
                        r'${EntryHead}configure.database.page\:db.type=JDBC',
                        r'${EntryHead}configure.database.page\:db.vendor=DERBY']
        else:
            template = [r'${EntryHead}configure.database.page\:db.jdbc.location=conf/${appName}/derby/repositoryDB',
                        r'${EntryHead}configure.database.page\:db.type=JDBC',
                        r'${EntryHead}configure.database.page\:db.vendor=DERBY']

        if app_name in ['jts']:
            template += [r'data.warehouse.page\:db.jdbc.location=conf/jts/derby/repositoryDB',
                         r'data.warehouse.page\:db.net.port=${derby_net_dw_port}',
                         r'data.warehouse.page\:db.type=JDBC',
                         r'data.warehouse.page\:db.vendor=DERBY']
        elif app_name in ['am', 'ccm', 'qm', 'rm', 'dcc']:
            template += [r'${EntryHead}data.warehouse.page\:db.jdbc.location=//localhost\:${derby_net_dw_port}/conf/jts/derby/warehouseDB',
                         r'${EntryHead}data.warehouse.page\:db.jdbc.password=',
                         r'${EntryHead}data.warehouse.page\:db.type=JDBC',
                         r'${EntryHead}data.warehouse.page\:db.vendor=derby_net']
        # pylint: enable=line-too-long
        return template

    def _get_db2_section_template(self, app_name, mapping):
        '''
        DB2 specfic configuration section

        :param app_name: application name
        :param mapping: A list of key-value pairs
        :return: A list of db2 database configuration
        '''

        user_props = qd_props.get_user_props(self.env_props)
        mapping['ibmQseInst_User'] = user_props.get_values('ibmQseInst_User')
        mapping['tempibmQseInst_PW'] = qd.decode_base64(user_props.get_values('ibmQseInst_PW'))
        mapping['DWFolder'] = qd.escape_path(qd.path_join(self.env_props.
                                                          get_values(qd_props.PN_QD_ROOT),
                                             qd_props.DEFAULT_DW_FOLDER))

        # [rs] does not have a database. all other apps do
        # pylint: disable=line-too-long
        if app_name == 'rs':
            template = [r'${EntryHead}com.ibm.team.jrs.setupStep\:db.jdbc.password=$tempibmQseInst_PW']
        else:
            mapping['DBTableName'] = self.env_props.get_values('%sDBName' % app_name)
            qd.echo_info('DBTableName: [%s]' % mapping['DBTableName'])
            template = [r'${EntryHead}configure.database.page\:db.jdbc.location=//$DBHost\:$DB2PORT/$DBTableName\:user\=$ibmQseInst_User;password\={password};',
                        r'${EntryHead}configure.database.page\:db.jdbc.password=$tempibmQseInst_PW',
                        r'${EntryHead}configure.database.page\:db.type=JDBC',
                        r'${EntryHead}configure.database.page\:db.vendor=DB2']

        if app_name in ['jts', 'am', 'ccm', 'qm', 'rm', 'dcc']:
            template += [r'${EntryHead}data.warehouse.page\:db.base.folder=$DWFolder',
                         r'${EntryHead}data.warehouse.page\:db.jdbc.location=//$DBHost\:$DB2PORT/$dwDBName\:user\=$ibmQseInst_User;password\={password};',
                         r'${EntryHead}data.warehouse.page\:db.jdbc.password=$tempibmQseInst_PW',
                         r'${EntryHead}data.warehouse.page\:db.type=JDBC',
                         r'${EntryHead}data.warehouse.page\:db.vendor=DB2',
                         r'${EntryHead}data.warehouse.page\:report.user=RPTUSER']
        # pylint: enable=line-too-long
        return template

    def _get_oracle_section_template(self, app_name, mapping):
        # Disable pylint unused-argument to keep method consistent
        # pylint: unused-argument
        '''
        oracle specfic configuration section

        :param app_name: Application name
        :param mapping: A list of key-value pairs
        :return: A list of oracle database configuration
        '''
        user_props = qd_props.get_user_props(self.env_props)

        # pylint: disable=line-too-long
        mapping['temporacleDBUsersPassword'] = qd.decode_base64(user_props.get_values('oracleDBUsersPassword'))
        mapping['DWFolder'] = qd.escape_path(qd.path_join(self.env_props.get_values(qd_props.PN_QD_ROOT),
                                             qd_props.DEFAULT_DW_FOLDER))

        # [rs] does not have a database. all other apps do
        if app_name == 'rs':
            template = [r'${EntryHead}com.ibm.team.jrs.setupStep\:db.jdbc.password=$temporacleDBUsersPassword']
        else:
            mapping['DBUser'] = self.env_props.get_values('oracle%sUser' % app_name)
            template = [r'${EntryHead}configure.database.page\:db.jdbc.location=thin\:$DBUser/{password}@//$DBHost\:$oraclePort/$oracleInstanceName',
                        r'${EntryHead}configure.database.page\:db.jdbc.password=$temporacleDBUsersPassword',
                        r'${EntryHead}configure.database.page\:db.type=JDBC',
                        r'${EntryHead}configure.database.page\:db.vendor=ORACLE']

        if app_name in ['jts', 'am', 'ccm', 'qm', 'rm', 'dcc']:
            mapping['dwUser'] = self.env_props.get_values('oracledwUser')
            template += [r'${EntryHead}data.warehouse.page\:db.base.folder=$DWFolder',
                         r'${EntryHead}data.warehouse.page\:db.jdbc.location=thin\:$dwUser/{password}@//$DBHost\:$oraclePort/$oracleInstanceName',
                         r'${EntryHead}data.warehouse.page\:db.jdbc.password=$temporacleDBUsersPassword',
                         r'${EntryHead}data.warehouse.page\:db.type=JDBC',
                         r'${EntryHead}data.warehouse.page\:db.vendor=ORACLE',
                         r'${EntryHead}data.warehouse.page\:report.user=RPTUSER']
        # pylint: enable=line-too-long
        return template

    def _get_sql_section_template(self, app_name, mapping):
        # Disable pylint unused-argument to keep method consistent
        # pylint: unused-argument
        '''
        sqlserver specfic configuration section

        :param app_name: Application name
        :param mapping: A list of key-value pairs
        :return: A list of oracle database configuration
        '''
        user_props = qd_props.get_user_props(self.env_props)

        # pylint: disable=line-too-long
        mapping['tempsqlDBUsersPassword'] = qd.decode_base64(user_props.get_values('sqlDBUsersPassword'))

        # [rs] does not have a database. all other apps do
        if app_name == 'rs':
            template = [r'${EntryHead}com.ibm.team.jrs.setupStep\:db.jdbc.password=$tempsqlDBUsersPassword']
        else:
            mapping['DBuser'] = self.env_props.get_values('%sDBuser' % app_name)
            mapping['DBTableName'] = self.env_props.get_values('%sDBName' % app_name)
            template = [r'${EntryHead}configure.database.page\:db.jdbc.location=//$DBHost\:$sqlserverPort;databaseName\=$DBTableName;user\=$DBuser;password\={password}',
                        r'${EntryHead}configure.database.page\:db.jdbc.password=$tempsqlDBUsersPassword',
                        r'${EntryHead}configure.database.page\:db.type=JDBC',
                        r'${EntryHead}configure.database.page\:db.vendor=SQLSERVER']

        if app_name in ['jts', 'am', 'ccm', 'qm', 'rm', 'dcc']:
            if qd.is_linux():
                mapping['DWFolder'] = 'C:\\' + self.env_props.get_values(qd_props.PN_QD_ROOT).replace('/', '') + '/' + qd_props.DEFAULT_DW_FOLDER
            else:
                mapping['DWFolder'] = qd.escape_path(qd.path_join(self.env_props.get_values(qd_props.PN_QD_ROOT), qd_props.DEFAULT_DW_FOLDER))
            template += [r'${EntryHead}data.warehouse.page\:db.base.folder=$DWFolder',
                         r'${EntryHead}data.warehouse.page\:db.jdbc.location=//$DBHost\:$sqlserverPort;databaseName\=$dwDBName;user\=$dwDBuser;password\={password}',
                         r'${EntryHead}data.warehouse.page\:db.jdbc.password=$tempsqlDBUsersPassword',
                         r'${EntryHead}data.warehouse.page\:db.type=JDBC',
                         r'${EntryHead}data.warehouse.page\:db.vendor=SQLSERVER',
                         r'${EntryHead}data.warehouse.page\:report.user=RPTUSER']
        # pylint: enable=line-too-long
        return template
