#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Provides common functions for dealing with 'media' in Quick Deployer
'''
import os

import shutil
import tempfile
import time

from qdlib import qd, qd_props
from qdlib.url_utils import FtpClient
from qdlib.url_utils import HttpClient

##
# Property names used in this module
##
PN_MEDIA_QD_INTERNAL_MEDIA = 'QD_InternalMedia'  #:
''' Property name - internal media, in environment, used by UCD '''
PN_MEDIA_MEDIA_ACCESS = 'mediaAccess'  #:
''' Property name - media access, used in media '''
PN_MEDIA_ROOT = 'mediaRoot'  #:
''' Property name - media root, in environment, used by UCD '''
PN_MEDIA_SERVER = 'mediaServer'  #:
''' Property name - media server, in environment, used by UCD '''
PN_MEDIA_COPY_LOCAL = 'copyBeforeUnzip'  #:
''' Property name - copy before unzip, in environment, used by UCD '''

PN_WGET_CHECK_WGET_CERT = 'checkWgetCertificate'  #:
''' Property name - check wget certificate, in media '''
PN_WGET_PROTOCOL = 'wgetProtocol'  #:
''' Property name - wget protocol, in media '''
PN_WGET_MEDIA_USER = 'mediaUser'  #:
''' Property name - media user, in userCreds, used by UCD '''
PN_WGET_MEDIA_PASSWORD = 'mediaPassword'  #:
''' Property name - media password, in userCreds, used by UCD '''

CHUNK_SIZE = 1024 * 1024  #:
''' The buffer size to use when downloading large files in pieces.  Current 1MB. '''


# pylint: disable=too-few-public-methods
class _BaseQDMedia(object):
    '''
    (Internal) Provide the base implementation for property accesses (via pget method) for QDMedia.
    '''

    def __init__(self, media_props, env_props):
        assert env_props or media_props, 'either media_props or env_props  must be provided.'
        self.env_props = env_props
        self.media_props = media_props or qd_props.get_media_props(env_props)

    def pget(self, *var, **defaults):
        '''
        Get media property (base media class)
        '''
        return self.media_props.get_values(*var, **defaults)


##
# QDMedia
##
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-public-methods
class QDMedia(_BaseQDMedia):
    '''
    Provides typical implementation of a QD media.

    Usage:

    - PREP
        - (Opt) In extended constructor, set property names for a particular media:
            - a_media_type_prop, a_media_root_prop, a_version_prop, a_file_prop
        - Override methods which are for calculating media-related locations if necessary:
            - get_local_file_location, get_remote_file_location,
            - get_download_dir, get_unzip_dir, get_result_path_from_unzip_dir,
            - get_result_from_result_path
        - Override _custom_prepare_media if you need more customize
    - RUNTIME
        - Create an extend class instance, or create an instance and set property names
        - Call media.prepare_media()
        - Use the media at media.result
        - Call media.cleanup() to remove all temporary files
    '''

    def __init__(self, media_props=None, env_props=None):
        _BaseQDMedia.__init__(self, media_props, env_props)
        # common properties
        self.qd_internal_media_prop = PN_MEDIA_QD_INTERNAL_MEDIA
        self.media_access_prop = PN_MEDIA_MEDIA_ACCESS  # FILE|HTTP|HTTPS
        self.media_root_prop = PN_MEDIA_ROOT
        self.media_server_prop = PN_MEDIA_SERVER

        # a media specific properties
        self.a_media_type_prop = None  # FILE|ZIP|(CLM|SSE|CE)
        self.a_media_root_prop = None
        self.a_version_prop = None
        self.a_file_prop = None
        self.a_dict = {}  # "OSdependent", "version"

        # calculated during "prepare"
        self.result = None  # The 'prepared' (local) media path

        # This property specifies the most recently downloaded (or copied) file.  It is set by the
        # download_or_copy() method and cleared by the cleanup() method.
        self.downloaded_file = None

        self.temp_unzip_dir = None  # folder containing unzipped files, which can also be cleaned-up

    def get_app_media_type(self):
        '''
        Get the app media type.

        :return: the current app_media_type.
        '''

        app_media_type = self.pget(self.a_media_type_prop)
        return app_media_type

    def prepare_media(self):
        '''
        Gets media ready for installation.  Typically this includes fetching and unzipping the
        necessary files.

        :return: Repository or file or folder location
        '''
        result = self._custom_prepare_media()
        if result:
            # done by _custom_prepare_media
            self.result = result
        else:
            # use the default implementation
            a_media_type = self.get_app_media_type()
            if a_media_type == 'FILE':
                result_path = self.get_local_file_location()
                qd.echo_info('using file repository, %s#result_path [%s]'
                             % (type(self).__name__, result_path))

            elif a_media_type == 'ZIP':

                # Install media directory
                src_zip_dir = self.get_directory_location(**self.a_dict)
                qd.echo_info('using zipped up file repository, %s#src_zip_dir [%s]'
                             % (type(self).__name__, src_zip_dir))

                # get a list of files for the src directory and process each file
                image_files = self.get_file_list(src_zip_dir)
                if len(image_files) <= 0:
                    assert False, "File list empty [%s]" % (src_zip_dir)
                for image_file in image_files:
                    result_path = self.process_each_zip_file(src_zip_dir, image_file)
            else:
                assert False, "unknown media type $%s [%s]" % (self.a_media_type_prop, a_media_type)
            # get the result
            self.result = result_path

        # return the result
        return self.result

    @classmethod
    def _custom_prepare_media(cls):
        '''
        This can be overridden.

        Put customized way to prepare media here.

        :return: result when handled, None when not handled.
        '''
        return None

    def get_file_location(self):  # from prepare_media
        '''
        Returns the location of the file.  This may be a local file path or a URL.  This method
        covers four basic scenarios:

        1.  If media access is set to file, then the local file path is returned
        2.  If media access is a URL and a_file is a URL, it returns that URL
        3.  If media access is a URL and a_file is an absolute file path, then that path is
            returned
        4.  If media access is a URL and the above doesn't apply, then the full path is
            determined and returned

        :return: file path or URL to the file
        '''
        media_access = self.pget(self.media_access_prop)
        if media_access == 'FILE':
            # when mediaAccesss is FILE, get the local file location
            return self.get_local_file_location()
        elif media_access in ['HTTP', 'HTTPS', 'FTP']:
            a_file = self.pget(self.a_file_prop)
            if qd.is_url(a_file):
                return a_file  # when a_file_prop is URL, use the a_file_prop
            elif os.path.isabs(a_file):
                return a_file  # when a_file is absolute, use as-is
            else:
                return self.get_remote_file_location()  # build the url
        else:
            raise Exception('unknown $%s [%s]' % (self.media_access_prop, media_access))

    def get_directory_location(self, **kwargs):
        '''
        Returns the location of the directory.  This may be a local file path or a URL.  This method
        covers four basic scenarios:

        1.  If media access is set to file, then the local file path is determined and returned
        2.  If media access is a URL and a_file is a URL, it returns that URL
        3.  If media access is a URL and a_file is an absolute file path, then that path is
            returned
        4.  If media access is a URL and the above doesn't apply, then the full URL is
            determined and returned

        When determining the file path or URL, the version or OS information is taken into account.

        :param kwargs: construct the src directory path, dictionary {"OSdependent": '<value>',
               "version": '<value>'}
        :return: file path or URL to the directory
        '''
        # Get source zip location
        media_access = self.pget(self.media_access_prop)
        if media_access == 'FILE':
            # when mediaAccesss is FILE, get the zip_file location
            return self.get_local_directory_location(**kwargs)
        elif media_access in ['HTTP', 'HTTPS', 'FTP']:
            a_file = self.pget(self.a_file_prop)
            if qd.is_url(a_file):
                return a_file  # when a_file_prop is URL, use the a_file_prop
            elif os.path.isabs(a_file):
                return a_file  # when a_file is absolute, use as-is
            else:
                # build http zip directory location
                return self.get_remote_directory_location(**kwargs)
        else:
            raise Exception('unknown $%s [%s]' % (self.media_access_prop, media_access))

    def get_local_file_location(self):
        '''
        Returns the path to the local file.  This essentially joins $mediaRoot/$myMediaRoot/$myFile
        into a single path.  This may be overridden by subclasses.

        :return: path to local file
        '''
        media_root = self.pget(self.media_root_prop)
        a_media_root = self.pget(self.a_media_root_prop)
        a_file = self.pget(self.a_file_prop)
        return qd.path_join(media_root, a_media_root, a_file)

    def get_local_directory_location(self, **kwargs):
        '''
        Returns the path to the local directory.  This essentially joins
        $mediaRoot/$myMediaRoot/$myVersion/$myOS/$myfile although the OS portion is not included if
        the OSdependent value is not set in the kwargs dictionary.

        :param kwargs: construct the src directory path, dictionary {"OSdependent": '<value>',
               "version": '<value>'}
        :return: path to local directory
        '''
        media_root = self.pget(self.media_root_prop)
        a_media_root = self.pget(self.a_media_root_prop)
        if "version" in kwargs:
            a_version = kwargs["version"]
        else:
            a_version = self.pget(self.a_version_prop)
        if "OSdependent" in kwargs:
            return qd.path_join(media_root, a_media_root, a_version, kwargs["OSdependent"])
        else:
            return qd.path_join(media_root, a_media_root, a_version)

    def get_remote_file_location(self):
        '''
        Builds and returns a URL to the file.  This essentially joins
        media_server/$myMediaRoot/$myVersion/$myFile.  This method may be overridden by subclasses.

        :return: file URL
        '''
        media_server = self.pget(self.media_server_prop)
        a_media_root = self.pget(self.a_media_root_prop)
        a_version = self.pget(self.a_version_prop)
        a_file = self.pget(self.a_file_prop)
        return qd.url_join(media_server, a_media_root, a_version, a_file)

    def get_remote_directory_location(self, **kwargs):
        '''
        Builds and returns a URL to the directory.  This essentially joins
        media_server/$myMediaRoot/$myVersion/$myFile/$myOS although the OS portion is not set if
        the OSdependent value is not set in the kwargs dictionary.  May be subclassed.

        :param kwargs: construct the src directory path, dictionary
        {"OSdependent": '<value>', "version": '<value>'}
        return directory URL
        '''
        media_server = self.pget(self.media_server_prop)
        a_media_root = self.pget(self.a_media_root_prop)
        if "version" in kwargs:
            a_version = kwargs["version"]
        else:
            a_version = self.pget(self.a_version_prop)
        if "OSdependent" in kwargs:
            return qd.url_join(media_server, a_media_root, a_version, kwargs["OSdependent"])
        else:
            return qd.url_join(media_server, a_media_root, a_version)

    def download_or_copy(self, src_location):
        '''
        Returns a local file path to the artifact specified as the src_location.  If the file is
        already local, the same path is returned.  If the file is on a mounted drive, the file is
        copied to a local directory.  If the src_location is an HTTP, HTTPS, or FTP URL, then the
        file is downloaded to a local directory.

        :param src_location: the location of the desired artifact (file path or URL)
        :return: local path to the desired artifact
        '''

        if qd.is_url(src_location):
            # download the zip from the url
            download_dir = self.get_download_dir()
            self.downloaded_file = download_file(src_location, download_dir,
                                                 self.env_props, self.media_props)
        else:
            copy_local = self.pget(PN_MEDIA_COPY_LOCAL)
            copy_local = qd.parse_boolean(copy_local, PN_MEDIA_COPY_LOCAL, none_value=False)
            if copy_local:
                # only download a copy if NFS has been setup
                download_dir = self.get_download_dir()
                if not os.path.exists(download_dir):
                    qd.echo_debug("Creating folder [%s]" % download_dir)
                    os.makedirs(download_dir)

                qd.echo_debug("Copying file [%s] to directory [%s]" % (src_location, download_dir))
                self.downloaded_file = qd.copy_file(src_location, download_dir)
            else:
                return src_location  # use the src_location itself and skip the check

        # check downloaded file
        qd.check_file_exists(self.downloaded_file)
        return self.downloaded_file

    def get_download_dir(self):
        '''
        Get the download dir. By default, $QD_InternalMedia/$myMediaRoot.

        :return: download directory
        '''
        qd_internal_media = self.pget(self.qd_internal_media_prop)
        a_media_root = self.pget(self.a_media_root_prop)
        return qd.path_join(qd_internal_media, a_media_root)

    def get_file_list(self, src_zip_dir):
        '''
        Get the list of file names of the src_zip_directory.  Used to download all files under that
        directory

        :param src_zip_dir: Absolute path of the media source directory
        :return: list of file names in specified directory
        '''
        media_access = self.pget(self.media_access_prop)
        if media_access == 'FTP':
            qd.echo_debug("Ready to get folder contents [%s]" % (src_zip_dir))
            return self.get_file_list_ftp(src_zip_dir)
        elif self.pget(self.media_access_prop) in ['HTTP', 'HTTPS']:
            timer = qd.start_clock("Ready to get ZIP")
            directory = self.get_file_list_http(src_zip_dir)
            qd.stop_clock(timer)
            return directory
        elif media_access == 'FILE':
            return self.get_file_list_local(src_zip_dir)
        else:
            raise Exception('unknown media access type [%s]' % self.pget(self.media_access_prop))

    def verify_file_list_existence(self, src_zip_dir):
        '''
        Determine if there is any file under src_zip_dir

        :param: src_zip_dir
        :return True if there is a file under the specific directory,
                False if no file found
        '''
        import urllib2
        import ftplib
        try:
            image_list = self.get_file_list(src_zip_dir)
        except (OSError, urllib2.HTTPError, ftplib.error_perm):
            qd.echo_info('%s does not exist' % src_zip_dir)
            return False
        if len(image_list) <= 0:
            return False
        else:
            return True

    # pylint: disable=no-self-use
    def get_file_list_ftp(self, directory):
        '''
        Using ftp protocol to get list of file names from a remote media server

        :param directory: Absolute path of the directory on the media server
        :return: list of file names in specified ftp location
        '''
        client = FtpClient()
        return client.get_file_list(directory)

    def get_file_list_http(self, directory):
        '''
        Using http or https protocol to get list of file names from the remote media repository

        :param directory: Absolute path of the media source directory
        :return: list of file names in specified http location
        '''
        check_wget_certificate = self.media_props.get_values(PN_WGET_CHECK_WGET_CERT)
        check_wget_certificate = qd.parse_boolean(check_wget_certificate, PN_WGET_CHECK_WGET_CERT)
        user_props = qd_props.get_user_props(self.env_props or self.media_props)  # for credentials
        media_user, media_password = user_props.get_values(PN_WGET_MEDIA_USER,
                                                           PN_WGET_MEDIA_PASSWORD)
        client = HttpClient(directory, media_user, qd.decode_base64(media_password),
                            check_ssl_cert=check_wget_certificate)
        return client.get_files(directory)

    def get_file_list_local(self, directory):
        '''
        Returns a list of files in the specified local path.

        :param directory: Absolute path to the filesystem directory
        :return: list of files in local directory
        '''
        return os.listdir(directory)

    def process_each_zip_file(self, src_zip_dir, image_file, unzip_folder='unzipped'):
        # from _custom_prepare_media
        '''
        Process each individual file from the src_zip_dir

        :param src_zip_dir: the parent directory name of the file to be downloaded
        :param image_file: to be downloaded file name
        :param unzip_folder: folder name where zip file should unzip to
        '''
        qd.echo_info("processing file [%s]" % image_file)

        media_access = self.pget(self.media_access_prop)
        if media_access == 'FILE':
            filename = qd.path_join(src_zip_dir, image_file)
        elif media_access in ['HTTP', 'HTTPS', 'FTP']:
            filename = qd.url_join(src_zip_dir, image_file)
        else:
            raise Exception('unknown $%s [%s]' % (self.media_access_prop, media_access))
        local_src_zip = self.download_or_copy(filename)
        timer = qd.start_clock("Ready to unzip ZIP")
        if qd.is_zipfile(local_src_zip):
            unzip_dir = self.unzip_and_get_unzip_dir(local_src_zip, unzip_folder)
        elif qd.is_tarfile(local_src_zip):
            unzip_dir = self.unzip_and_get_unzip_dir(local_src_zip, unzip_folder)
        else:
            qd.echo_info("Do not need unzip the file")
            unzip_dir = self.get_unzip_dir(unzip_folder)
            if media_access == 'FILE':
                download_dir = self.get_download_dir()
                if not os.path.exists(qd.path_join(download_dir, image_file)):
                    if not os.path.exists(download_dir):
                        qd.echo_debug("Creating folder [%s]" % download_dir)
                        os.makedirs(download_dir)
                    self.downloaded_file = qd.copy_file(local_src_zip, download_dir)

        qd.stop_clock(timer)

        result_path = self.get_result_path_from_unzip_dir(unzip_dir)
        qd.echo_info('using unzipped file repository, %s#result_path [%s]'
                     % (type(self).__name__, result_path))
        # check existence of the result_path
        # do not need check the file existence since result_path return the unzipped folder
        # qd.check_file_exists(result_path)
        return result_path

    def unzip_and_get_unzip_dir(self, unzip_zip, folder='unzipped'):
        '''
        Unzips the specified zip and returns the folder in which the zip was extracted.

        :param unzip_zip: file to be unzipped
        :param folder: location where zip will be extracted
        :return: location of extracted files
        '''
        self.temp_unzip_dir = self.get_unzip_dir(folder)
        # unzip
        qd.unzip_zip(unzip_zip, self.temp_unzip_dir)
        return self.temp_unzip_dir

    def get_unzip_dir(self, folder='unzipped'):  # from unzip_and_get_unzip_dir
        '''
        Get the unzip target dir. By default, $downloadDir/unzipped.

        :param folder: optional name of unzip directory
        :return: full path to directory into which files may be unzipped
        '''
        return qd.path_join(self.get_download_dir(), folder)

    @classmethod
    def get_result_path_from_unzip_dir(cls, unzip_dir):  # from prepare_media
        '''
        For media where aMediaType is ZIP, get result_path from the unzip dir.
        By default and typically, same to the unzip_dir.

        # decoration to the unzip_dir

        :param unzip_dir: unzip directory
        :return: unzip directory
        '''
        return unzip_dir

    @classmethod
    def get_result_from_result_path(cls, result_path):  # from prepare_media
        '''
        For all media, get the result string (path or URL depending on media).
        By default, same to result_path.

        # decoration to the result_path

        :param result_path: result path
        :return: result path
        '''
        return result_path

    def cleanup(self):
        '''Delete all "prepared" stuff.'''
        self.result = None
        # tidy up by deleting anything downloaded or unzipped
        download_dir = self.get_download_dir()
        if download_dir and os.path.exists(download_dir):
            qd.echo_info('Removing the directory [%s]' % download_dir)
            number_tries = 10
            # wait_time is in seconds
            wait_time = 120
            number = 0
            for number in range(number_tries):
                number = number + 1
                try:
                    shutil.rmtree(download_dir)
                    qd.echo_info("Deleted directory [%s]" % download_dir)
                    # break out the loop if rmtree call succeeded
                    break
                except OSError as except_msg:
                    qd.echo_info("Caught exception, re-try to delete directory")
                    if number == number_tries:
                        qd.exit_with_err_msg("Failed to delete directory with error message: %s"
                                             % except_msg)
                    time.sleep(wait_time)
            self.temp_unzip_dir = None
            self.downloaded_file = None

    # getter for basic props
    def get_my_version(self):
        '''
         Get media property (QDMedia class).

        :return: media version
        '''
        return self.pget(self.a_version_prop)


def download_file(file_url, download_dir, env_props, media_props=None):
    '''
    Downalods the specified file to the specified location.

    :param file_url: location of file to be downloaded
    :param download_dir: location into which file will be downloaded
    :param env_props: environment properties
    :param media_props: media properties
    :return: the downloaded file path if successful, otherwise None
    '''
    filename = qd.basename(file_url)
    downloaded_file_loc = qd.path_join(download_dir, filename)

    downloaded_file = do_wget(file_url, download_dir, env_props, media_props)
    if downloaded_file:
        assert downloaded_file == downloaded_file_loc
        qd.echo_info("download complete")
        qd.check_file_exists(downloaded_file_loc)
    else:
        qd.exit_with_err_msg("failed to download file [%s]" % filename)

    return downloaded_file


#
# Wget
#
def do_wget(file_url, download_dir, env_props, media_props=None):
    '''
    Download content of the file_url to the download_dir using options in Media/User props.

    :param file_url: location of file to be downloaded
    :param download_dir: location into which file will be downloaded
    :param env_props: environment properties
    :param media_props: media properties
    :return: the downloaded file path if successful, otherwise None
    '''
    if not media_props:
        media_props = qd_props.get_media_props(env_props)  # for media properties

    # check some properties are valid
    check_wget_certificate, wget_protocol = media_props.get_values(PN_WGET_CHECK_WGET_CERT,
                                                                   PN_WGET_PROTOCOL)
    check_wget_certificate = qd.parse_boolean(check_wget_certificate, PN_WGET_CHECK_WGET_CERT)

    def check_protocol(protocol):
        '''
        Checks if the specified protocol is valid.  Allowed values are http, https, and ftp.  If the
        specified protocol is not valid, the code will exit with an error message.

        :param protocol: protocol to check
        '''
        if protocol in ['http', 'https', 'ftp']:
            qd.echo_info('[%s] is valid.' % protocol)
        else:
            qd.exit_with_err_msg('[%s] is not a valid protocol' % protocol)
    check_protocol(wget_protocol)

    # create http client options
    if wget_protocol == 'https':
        user_props = qd_props.get_user_props(env_props or media_props)  # for credentials
        media_user, media_password = user_props.get_values(PN_WGET_MEDIA_USER,
                                                           PN_WGET_MEDIA_PASSWORD)
    else:
        media_user, media_password = None, None

    target_file = qd.path_join(download_dir, qd.basename(file_url))

    download(file_url, target_file, media_user, media_password, True, check_wget_certificate)

    # === finished ====
    qd.echo_info("-- EOF [do_wget] SUCCESS --")
    return target_file


# pylint: disable=too-many-arguments
def download(source_url, target_file, media_user=None, media_password=None,
             password_is_encoded=True, check_ssl_certificate=False):
    '''
    Downloads a file from source_url to the target_file.  If the media_user and media_password are
    specified, those are used to authenticate.  By default the password is expected to be in Base64
    encoded, but a plain text password may be used by setting password_is_encoded=False.  By
    default SSL certificates are not checked, but a check may be forced by setting
    check_ssl_certificate=True.

    :param source_url: a string representation of the URL to the file to be downloaded.  It is
                       should begin with http:// or https://.
    :param target_file: path to the file to be written.  the parent directory does not need to exist
                       but the file must not exist or an exception will be thrown.
    :param media_user: (optional) the user name to be used for authenticating.  If no authentication
                       is required, it should be set to None (the default value).
    :param media_password: (optional) the user password to be used for authenticating.  If no
                       authentication is required, it should be set to None (the default value).
                       The value is assumed to be Base64 encoded unless the password_is_encoded
                       param is set to False in which case the value should be plain text.
    :param password_is_encoded: (optional) set to true to uses a plain text password otherwise
                       use the default value of True
    :param check_ssl_certificate: (optional) set to True check the ssl certificate and fail if
                       anything is amiss.  This is a strict check so the default is False.

    Does not return anything.  Throws an exception if anything goes wrong.
    '''
    qd.echo_info("download [%s] to [%s]" % (source_url, target_file))

    # Create the download directory if it doesn't already exist
    download_dir = os.path.dirname(os.path.realpath(target_file))
    if not os.path.exists(download_dir):
        os.makedirs(download_dir)

    if os.path.exists(target_file):
        qd.echo_error("File [%s] already exists" % target_file)
        raise IOError("File [%s] already exists" % target_file)

    if media_password is None:
        decoded_password = None
    elif password_is_encoded:
        decoded_password = qd.decode_base64(media_password)
    else:
        decoded_password = media_password

    client = HttpClient(source_url, media_user, decoded_password,
                        check_ssl_cert=check_ssl_certificate)

    # Don't keep decoded_password value around
    decoded_password = None

    headers = {'Connection': 'Keep-Alive'}

    response = client.execute(source_url, headers=headers)

    try:
        # Get expected size of file
        headers = response.headers.dict
        expected_size = headers.get('content-length')
        if expected_size:
            expected_size = int(expected_size)

        with open(target_file, 'wb') as file_handle:
            while True:
                chunk = response.read(CHUNK_SIZE)
                if not chunk:
                    break
                file_handle.write(chunk)
    finally:
        response.close()

    actual_size = os.stat(target_file).st_size
    qd.echo_info("expected size: [%s] actual size: [%s]" % (expected_size, actual_size))

    if expected_size and expected_size != actual_size:
        qd.echo_error('The expected size does not match the actual size.')
        raise IOError("File %s does not have expected size %s" % (target_file, expected_size))


# -----------------------------------------------------------------
# Codes below are SSE/CE/CLM specific
# -----------------------------------------------------------------

##
# Special for media version
##
def get_media_content_and_app_release(app_media_root):
    '''
    Calculate the application media content and the application release
    e.g., (CE, 601) from CE601. Return (None,None) for unknown media content

    :param app_media_root: in the form of CLM60x or CE60x
    :return tuple (appMediaContent, appRelease):
        * **appMediaContent**: [CE or CLM] (eg CE from CE601)
        * **appRelease**: [60x] (eg 601 from CE601)
        (None,None) for unknown media root
    '''
    # Disable pylint invalid-name because function name is > 30 chars and want to keep it
    # pylint: disable=invalid-name

    # extract the prefix and then the release
    appMediaContent = {'SS': 'SSE', 'CE': 'CE', 'CL': 'CLM'}.get(app_media_root[:2])
    if appMediaContent:
        appRelease = app_media_root[len(appMediaContent):]
        if appRelease == '60':
            appRelease = '600'
        return appMediaContent, appRelease
    return None, None


def get_ifix_release_and_ifix_patch(app_name, ifix_version, app_release):
    '''
    Calculate the ifix release and ifix_patch
    e.g., (602, 007) from CLM_602_iFix007. Return (None,None) for unknown application prefix

    Usage::

    dmms_ifix_release, dmms_ifix_patch =
         qd_props.media.get_ifix_release_and_ifix_patch('DMMS', dmms_ifix_version, app_release)

    :param app_name: name of application
    :param ifix_version: version of ifix
    :param app_release: application release version
    :return: iFix release and iFix version
    '''
    # Disable pylint invalid-name because function name is > 30 chars and want to keep it
    # pylint: disable=invalid-name

    if (ifix_version == 'CLM' or ifix_version == 'DM' or ifix_version == 'DMMS' or
            ifix_version == 'RMM'):
        qd.echo_info("[%s] iFix version [%s] being installed via the pipeline"
                     % (app_name, ifix_version))
        return app_release, 'pipeline'

    qd.echo_info("extract the iFix release & patch strings")
    if app_name == 'CLM':
        ifix_release = ifix_version.split('_')[1]
        ifix_patch = ifix_version.rsplit('x', 1)[1]
    elif app_name == 'DM':
        ifix_release = ifix_version.split('_')[1]
        ifix_patch = ifix_version.rsplit('_', 1)[1]
    elif app_name == 'DMMS':
        ifix_release = ifix_version.split('_')[3]
        ifix_patch = ifix_version.rsplit('_', 1)[1]
    elif app_name == 'RMM':
        ifix_release = ifix_version.split('_')[1]
        ifix_patch = ifix_version.rsplit('_', 1)[1]

    else:
        qd.echo_error("Does not support iFix product [%s]" % app_name)
        return None, None
    # allow for 60 and 600 being the same stream
    if ifix_release == '60':
        qd.echo_info("modifying iFix Release [%s] to become [600]" % ifix_release)
        ifix_release = '600'

    if ifix_patch != 'pipeline':
        # remove chars from rhs if they are [a-z][A-Z]
        qd.echo_info("modifying iFix-Patch [%s] to remove any [a-z][A-Z] from the rhs" % ifix_patch)
        stripStr = 'abcdefghijklmnopqrstuvwxyz'
        stripStr = stripStr + stripStr.upper()
        ifix_patch = ifix_patch.rstrip(stripStr)

        # pack the [ifix_patch] so that it is always 3 or more characters
        qd.echo_info("modifying iFix-Patch [%s] to ensure it has at least 3 characters"
                     % ifix_patch)
        # `echo "000"${tempReturnPatch} | rev | cut -c 1-3 | rev`
        ifix_patch = ifix_patch.rjust(3, '0')

# last lines
    qd.echo_info("iFix-Version [%s], iFix-Release [%s] and iFix-Patch [%s]"
                 % (ifix_version, ifix_release, ifix_patch))
    return ifix_release, ifix_patch


def get_ifix_stream(app_media_root):
    '''
    calculate the iFix stream property based on the {app_media_root}

    :param app_media_root: in the form of CLM60x or CE60x

    :return:
        (6.0.x.0) from CE60x or CLM60x
        special case (6.0.0.0) from CE60
        None for an unknown media root
    '''

    temp_content = get_media_content_and_app_release(app_media_root)
    temp_release = temp_content[1]
    # from the release produce a [a.b.c] stream
    team_stream = {'60': '6.0.0.0', '600': '6.0.0.0', '601': '6.0.1.0', '602': '6.0.2.0',
                   '603': '6.0.3.0', '604': '6.0.4.0', '605': '6.0.5.0'}.get(temp_release)
    if team_stream:
        qd.echo_info("iFix based on [%s] expands to a stream of [%s]"
                     % (app_media_root, team_stream))
        return team_stream
    return None


##
# Keyring file location
##
def get_im_keyring_file_location():
    # Disable pylint unused-argument for future proofing; other methods call here with value
    # pylint: disable=unused-argument
    '''
    return the OS path to the keyring file
    '''
    keyringfileloc = qd.path_join(tempfile.gettempdir(), 'IM_Keyring')
    return keyringfileloc
