#!/usr/bin/python
# pylint: disable=too-many-lines
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Implement common behaviors and often-used procedures among QD processes.

- Console output and logging
- OS-dependent procedures that are worth sharing among scripts
- Basic HTTP client w/ Basic authentication and controlling SSL certificate check
- Load properties file
'''
from string import Template
import StringIO
import base64
import codecs
import collections
import datetime
import glob
import inspect
import logging
import math
import os
import platform
import re
import shutil
from subprocess import CalledProcessError
import subprocess
import sys
import tempfile
import time
import zipfile
import tarfile
import uuid
from xml.sax.saxutils import escape

from qdlib.url_utils import is_url


##
# Logging
##
def get_logger(name):
    '''
    Create a new logger instance that can interpret {timestamp} and {scriptname}.

    :param name: name of the new logger
    :return: new logger'''
    base_logger = logging.getLogger(name)
    return _QDLogerAdapter(base_logger)


class _QDLogerAdapter(logging.LoggerAdapter):
    '''Log adapter used among QD'''

    def __init__(self, logger):
        super(_QDLogerAdapter, self).__init__(logger, {})

    def process(self, msg, kwargs):
        '''
        Process message to be logged.

        :param msg: message to be logged
        :param kwargs: arguments related to message
        '''
        msg, kwargs = logging.LoggerAdapter.process(self, msg, kwargs)
        return msg, kwargs


SCRIPT_NAME = os.path.basename(sys.argv[0])  #:
''' Name of module at the top of the stack,  i.e., overall caller at current time;
    used to indicate current module being run.  Note: when reading documentation, this will
    actually be the process that creates the documentation.'''

LOG_DEFAULT_LEVEL = logging.INFO  #:
''' All modules log at this level, by default.

    +-------------+-------+
    | Description | Level |
    +=============+=======+
    | CRITICAL    | 50    |
    +-------------+-------+
    | ERROR       | 40    |
    +-------------+-------+
    | WARNING     | 30    |
    +-------------+-------+
    | INFO        | 20    |
    +-------------+-------+
    | DEBUG       | 10    |
    +-------------+-------+
    | NOTSET      |  0    |
    +-------------+-------+
'''
LOG_DEFAULT_FORMAT = '%(asctime)s %(message)s'  #:
''' All modules use this format for logging, by default. '''
LOG_DEFAULT_DATEFMT = '%b-%d %H:%M:%S'  #:
''' All modules use this date format for logging, by default. '''
LOG_INDENT = 58  #:
''' Number of characters in a log line until the actual message is displayed; used
    for indenting long messages.  Make sure to update if format, datefmt or processing
    on message in _get_echo_msg() changes.  Length calculated based on longest module name.'''


def _get_echo_msg(msg, log_level):
    '''
    Get message with keywords replaced; automatically prefaces message with calling module
    and line number.

    Supported keywords:
        * {timestamp}: replaced with the seconds from the EPOC (deprecated, use :func:`start_clock`
                       and :func:`stop_clock` instead)
        * {scriptname}: replaced with the script name

    :param msg: message to be output to log
    :param log_level: log level (used for our custom log formatting)
    :return: message to be logged
    '''
    if isinstance(msg, basestring):
        if r'{timestamp}' in msg.lower():
            timestamp = math.floor(time.time())
            msg = msg.replace('{timestamp}', '%d' % timestamp)
        if r'{scriptname}' in msg.lower():
            msg = msg.replace('{scriptname}', SCRIPT_NAME)

    # Regardless of substitution above, preface with information about caller
    caller_name, caller_line = _get_echo_caller()
    msg = '%-27s %-5s %-7s %s' % (caller_name, caller_line, log_level, msg)
    return msg


def _get_echo_caller():
    '''
    Uses the stack to determine frame immediately before a qd.echo_* call.  When called during
    the output of an echo statement, provides the calling module/line in order to make resulting
    log more useful.  This is instead of a) only being able to show qd.py (which is where
    the echo statements are or b) the SCRIPT_NAME (which is the overall caller, but may not be
    the caller of the echo statement).  Output is cleaned to hide path and file extension; result
    is the name of the python module and line number.

    :return tuple (module_name, module_line):
        * **module_name**: str representing the module name calling the echo statement
        * **module_line**: str representing the line number calling the echo statement

    If an error happens (e.g., problem navigating the stack), the overall SCRIPT_NAME and '---' are
    returned.
    '''

    # Wrap everything in a try; if anything at all goes wrong, make sure to return valid result
    try:
        # Look through the stack; find the level that is in qd.py and echo_* method
        # Use a max, just to make sure we don't run out of stack.
        # Start with first level.
        stack_level = 1
        frame = inspect.stack()[stack_level]
        frame_module = frame[1]
        frame_method = frame[3]
        # Start with the first level; if it doesn't have what we're looking for, move up the stack
        while not (str(frame_module).endswith('qd.py') and
                   str(frame_method).startswith('echo_')) and stack_level < 10:

            try:
                # Get the next level
                stack_level += 1
                frame = inspect.stack()[stack_level]
                frame_module = frame[1]
                frame_method = frame[3]
            except IndexError:
                # Ran out of stack; raise same exception and next catch will handle result
                raise IndexError

        # At this point, the current frame is an echo_ call.  We want the very next frame.
        frame = inspect.stack()[stack_level + 1]
        frame_module = frame[1]
        frame_line = frame[2]

    # pylint: disable=bare-except
    except:
        # Something went wrong; fall back on script name
        frame_module = SCRIPT_NAME
        frame_line = '---'

    # Clean up the module name; don't need the path or the file extension
    index_start = str(frame_module).rfind(os.path.sep)
    index_end = str(frame_module).rfind('.')

    # Return the relevant information
    return frame_module[index_start + 1:index_end], frame_line


def start_clock(msg):
    '''
    Imitate starting a clock timer to see how long something takes; this method returns an
    epoch time that can then be passed into the related stop_clock method, which will output the
    total time taken for some operation.

    :param msg: Message to be output with clock information
    :return: tuple (random_id, msg, timestamp):
        * **random_id**: a random id to identify the clock
        * **msg**: the incoming message, so that both message related to clock are the same
        * **timestamp**: epoch time, will be used by stop_clock method to determine elapsed time
    '''

    timestamp = math.floor(time.time())
    random_id = uuid.uuid4().hex[0:4]  # four hex chars id
    echo_debug('START CLOCK [id:%s]: %s' % (random_id, msg))

    return (random_id, msg, timestamp)


def stop_clock(activity):
    '''
    Second method used to imitate a clock timer to see how long something takes.  Used in
    conjunction with start_clock to output the total time taken for some operation.  Incoming
    'activity' argument is simply the result of a previous call to :func:`start_clock`.

    :param tuple (random_id, msg, start_time):
        * **random_id**: a random id to identify the clock
        * **msg**: the incoming message, so that both message related to clock are the same
        * **timestamp**: epoch time, will be used by stop_clock method to determine elapsed time
    '''

    # Break up the tuple
    random_id = activity[0]
    msg = activity[1]
    start_time = activity[2]

    # Calculate elapsed time
    timestamp = math.floor(time.time())
    elapsed_time = timestamp - start_time

    # Cannot just format the elapsed_time, since it will look like a date from 1969.
    # Instead, do the math.  Assuming we'll only be counting max hours, not days.
    hours = elapsed_time // 3600 % 24
    minutes = elapsed_time // 60 % 60
    seconds = elapsed_time % 60

    echo_debug('STOP CLOCK [id:%s]: %s --> %d h %d m %s s '
               % (random_id, msg, hours, minutes, seconds))


LOGGER = get_logger(__name__)
''' Logger

In other modules, could use::

   logger = qd.get_logger(__name__)

However, using qd.echo_* methods instead of logger.* commands, in order to get standardized
and customized logging.  The only module that should use get_logger() is this one.
'''


#############
# STDOUT utility methods (echo)
# Used instead of LOGGER calls in order to standardize/customizer logging output.
#############
def echo_debug(msg):
    '''
    Echo DEBUG message using LOGGER

    :param msg: message to be logged
    '''

    LOGGER.debug(_get_echo_msg(wrap_long_lines(msg), 'DEBUG'))


def echo_info(msg):
    '''
    Echo INFO message using LOGGER

    :param msg: message to be logged
    '''
    LOGGER.info(_get_echo_msg(wrap_long_lines(msg), 'INFO'))


def echo_error(msg):
    '''
    Echo ERROR message using LOGGER

    :param msg: message to be logged
    '''
    LOGGER.error(_get_echo_msg(wrap_long_lines(msg), 'ERROR'))


def echo_warning(msg):
    '''
    Echo WARNING message using LOGGER

    :param msg: message to be logged
    '''
    LOGGER.warning(_get_echo_msg(wrap_long_lines(msg), 'WARNING'))


def wrap_long_lines(msg):
    '''
    Get a string representing the message, wrapped according to current indent.

    :param msg: message to be wrapped
    :return: String same message but with additional carriage returns and spaces to account for
             wrapping
    '''

    return_msg = msg
    if '\n' in msg or '\r' in msg or '\r\n' in msg:
        # Replace each newline/carriage return with a temp value
        if '\r\n' in return_msg:
            return_msg = return_msg.replace('\r\n', '~!@#$%^&*' + ' ' * LOG_INDENT)
        if '\n' in return_msg:
            return_msg = return_msg.replace('\n', '~!@#$%^&*' + ' ' * LOG_INDENT)
        if '\r' in return_msg:
            return_msg = return_msg.replace('\r', '~!@#$%^&*' + ' ' * LOG_INDENT)
        # Replace the temp value with newline
        return return_msg.replace('~!@#$%^&*', '\n')
    else:
        return msg


def parse_boolean(value, var_name, none_value=None):
    '''
    Check incoming string and return appropriate boolean value; throws exception if string cannot
    be parsed into boolean.

    :param value: string value, e.g., [true, TRUE, false, FALSE]
    :param var_name: variable name (used for output only)
    :param none_value: by default, None

    :return: Returns boolean value based on incoming arguments
    '''

    if not value and none_value is not None:
        value = none_value
    if isinstance(value, basestring):
        value = value.upper()
    if value == 'TRUE':
        value = True
    elif value == 'FALSE':
        value = False
    assert isinstance(value, bool), '[%s] %s not supported' % (value, var_name)
    return value


#############
# Utilities for Windows and Linux interchangeability
#############
def is_windows():
    '''
    Return True when operating system is Windows; False otherwise.

    :return: true when Windows, otherwise false
    '''
    return os.name == 'nt'


def is_linux():
    '''
    Return True when operating system is Linux; False otherwise. Logic depends on
    :func:`is_windows`.

    :return: true when Linux, otherwise false
    '''
    return not is_windows()


def os_switch(lin='linux', win='windows'):
    '''
    Return string representing operating system. Useful when commands differ between operating
    systems, e.g., 'gskcmd' vs. 'gskcmd.bat'.

    :param lin: (optional) linux value
    :param win: (optional) windows value
    :return: lin value when OS is Linux; win value when OS is Windows; error if neither.

    Equivalent to 'win if os.name=='nt' else lin'
    '''
    if is_windows():
        return win
    elif is_linux():
        return lin
    else:
        assert False, 'Unsupported operating system. Neither Linux nor Windows.'


def csv_split(string, separator=','):
    '''
    Split a string on a specific separator.

    :param string: string to be split
    :param separator: separator to be used to separate the string
    :return: list of non-empty values created by splitting the string
    '''
    assert separator
    return [e for e in re.split(r'\s*%s\s*' % separator, string) if e]


def csv_concat(first, second, separator=','):
    '''
    Concatenate two separated strings, removing empty values from incoming strings first.

    :param first: first string to be concatenated
    :param second: second string
    :param separator: separator to be used when pre-processing the first and second strings
    :return: string concatenated from the first and second strings; pre-processing includes
        splitting the incoming strings on the given separator
    '''
    assert separator
    return separator.join(csv_split(first, separator) + csv_split(second, separator))


#############
# Utilities for os.path.* like things
#############
def basename(path):
    '''
    Get basename of path.  Same as::

        os.path.basename(path)

    :param path: path
    :return: basename of path
    '''
    return os.path.basename(path)


def dirname(path):
    '''
    Get directory name of path.  Same as::

        os.path.dirname(path)

    :param path: path
    :return: directory name of path
    '''
    return os.path.dirname(path)


def mkdirs(name, mode=0777):
    '''
    Make directory.  Same as::

        os.makedirs(name, mode)

    :param name: new directory name
    :param mode: file permissions, e.g., '0777'
    '''
    os.makedirs(name, mode)


def path_join(fragment, *path):
    '''
    Join fragment and path arguments to make an OS-specific path.

    :param fragment: first fragment of the path
    :param path: zero or more parts of the path to be added to the fragment
    :return: string path, normalized for current operating system

    Examples:

    .. code-block:: python

            path_join("c:\\bin", "foo", "bar") = "c:\\bin\\foo\\bar"  # on a Windows machine
            path_join("/opt", "foo", "bar") = "/opt/foo/bar"  # on a Linux machine

    .. warning::
        * If the path contains something considered an absolute path, e.g., c:\\\\bin on Windows,
        the path will start with the latest fragment containing an absolute path and everything
        prior will be ignored.

        * Do NOT use for URLs; the normalize functionality changes // to / and switches the slashes
        on windows machines.

    '''
    result = os.path.join(fragment, *path)
    return os.path.normpath(result)


def url_join(fragment, *path):
    '''
    Join fragment and path arguments to make a URL.

    :param fragment: first fragment of the URL
    :param path: zero or more parts of the path to be added to the URL
    :return: string URL

    Example:

    .. code-block:: python

        url_join("https://", "v-12345678-os-1.work.com:9443", "jts", "admin") =
            "https://v-12345678-os-1.work.com:9443/jts/admin"
    '''

    # Join will work on a list; create list from incoming info
    # Strip of starting/trailing / if necessary
    if fragment.endswith('://'):
        fragments = [fragment]  # fragment might end with ://, let it stay
    else:
        fragments = [fragment.rstrip('/')]  # e.g., http://www.foo/, strip the trailing /
    for item in path:
        # rest of items in path should have starting and trailing / removed
        fragments.append(item.strip('/'))

    # Join all the fragments with url-friendly /
    result = '/'.join(fragments)

    # Then make sure we don't have /// anywhere (e.g., we'll have a :// then joined with /)
    clean_result = result.replace("///", "//")

    if not is_url(clean_result):
        echo_error("Using qd.url_join but result URL is not valid: %s" % clean_result)
        raise ValueError("The URL '%s' is not valid." % clean_result)

    return clean_result


def fileurl(path):
    '''
    Convert a Linux path to Windows format.

    :param path: Linux path
    :return: Windows format of that path, e.g,. file:path
    '''
    if is_windows():
        path = path.replace('\\', '/')
    return 'file:' + path


def chmod_777(path, mode='777'):
    '''
    (linux) chmod {mode}
     - current path only, if [mode] not specified then make it wide open [777]
    (windows) cacls - current path only - only supports 777 mode

    :param path: the file/folder name including the path
    :param mode: (optional) (default is 777) the new file/folder's permissions
    '''
    _chomod_777(path, mode, False)


def chmod_r_777(path, mode='777'):
    '''
    (linux) chmod {mode} recursively
     - current path only, if [mode] not specified then make it wide open [777]
    (windows) cacls recursively - current path only - only supports 777 mode

    :param path: the folder name including the path
    :param mode: (optional) (default is 777) the new folder's permissions
    '''
    _chomod_777(path, mode, True)


def _chomod_777(path, mode='777', recursive_flag=False):
    '''
    (linux) chmod {mode}
     - current path only, if [mode] not specified then make it wide open [777]
    (windows) cacls - current path only - only supports 777 mode

    :param path: the file/folder name including the path
    :param mode: (optional) (default is 777) the new file/folder's permissions
    :param recursive_flag: (boolean)
        **True** - will travel the full folder tree
        **False** - will just do the single file / folder
    '''

    assert isinstance(mode, basestring)
    if is_linux():
        if recursive_flag:
            echo_info('recursively make [%s] mode [%s]' % (path, mode))
            check_call(['chmod', '-R', mode, path])
        else:
            echo_info('make [%s] mode [%s]' % (path, mode))
            check_call(['chmod', mode, path])

        # display the changes took place
        check_call(['ls', '-la', path])

    else:
        if mode == '777':
            if recursive_flag:
                echo_info("make %s -R fully accessible to Everyone" % path)
                check_call(['cacls', path, '/T', '/E', '/R', '%s' % 'Everyone'])
                check_call(['cacls', path, '/T', '/E', '/G', '%s:F' % 'Everyone'])
            else:
                echo_info("make %s fully accessible to Everyone" % path)
                check_call(['cacls', path, '/E', '/R', '%s' % 'Everyone'])
                check_call(['cacls', path, '/E', '/G', '%s:F' % 'Everyone'])

            # display the changes took place
            check_call(['cacls', path])

        else:
            echo_warning('chmod [%s] [%s] is not implemented in Windows.' % (mode, path))


def chown(path, user, group=None):
    '''
    For the file/folder defined by {path}

    (linux) chown {user:group}

    (windows) uses cacls - which is (a) depricated

    :param path: the file/folder name including the path
    :param user: user name or id
    :param group: group name or id

    '''
    _chown(path, user, group, False)


def chown_r(path, user, group=None):
    '''
    Recursively for file/folder defined by {path}

    (linux) chown {user:group}

    (windows) uses cacls - which is (a) depricated

    :param path: the file/folder name including the path
    :param user: user name or id
    :param group: group name or id

    '''
    _chown(path, user, group, True)


def _chown(path, user, group=None, recursive_flag=False):
    '''
    For the file/folder defined by {path}

    (linux) chown {user:group}

    (windows) uses cacls - which is (a) depricated

    :param path: the file/folder name including the path
    :param user: user name or id
    :param group: group name or id
    :param recursive_flag: (boolean)
        **True** - will travel the full folder tree
        **False** - will just do the single file / folder

    '''

    if os.path.isdir(path) or os.path.isfile(path):
        # ensure user|group are strings
        if not isinstance(user, basestring):
            user = str(user)
        assert user, "user [%s] not defined!" % user
        if group:
            if not isinstance(group, basestring):
                group = str(group)
            assert group, "group [%s] not defined!" % group

        if is_linux():
            if group:
                chown_string = user + ":" + group
            else:
                chown_string = user

            if recursive_flag:
                echo_info('recursively update [%s] chown [%s]' % (path, chown_string))
                check_call(['chown', '-R', chown_string, path])
            else:
                echo_info('update [%s] chown [%s]' % (path, chown_string))
                check_call(['chown', chown_string, path])

            # display the changes took place
            check_call(['ls', '-la', path])

        else:
            def do_acls(path, owner, recursive_flag):
                '''
                    For the file/folder defined by {path}

                    (windows only) uses cacls - which is (a) depricated

                    :param path: the file/folder name including the path
                    :param owner: user/group name or id
                    :param recursive_flag: (boolean)
                        **True** - will travel the full folder tree
                        **False** - will just do the single file / folder
                '''

                if recursive_flag:
                    echo_info('recursively make [%s] fully accessible to [%s]' % (path, owner))
                    check_call(['cacls', path, '/T', '/E', '/R', owner])
                    check_call(['cacls', path, '/T', '/E', '/G', '%s:F' % owner])
                else:
                    echo_info('make [%s] fully accessible to [%s]' % (path, owner))
                    check_call(['cacls', path, '/E', '/R', owner])
                    check_call(['cacls', path, '/E', '/G', '%s:F' % owner])

            # always do the (owner) user
            echo_info('update [%s] ownership to [%s]' % (path, user))
            check_call(['icacls', path, '/setowner', user])
            do_acls(path, user, recursive_flag)

            if group:
                do_acls(path, group, recursive_flag)

            # display the changes took place
            check_call(['dir', '/q', path], shell=True)
            check_call(['cacls', path])

    else:
        exit_with_err_msg("path [%s] does not refer to an existing file or folder" % path)


def exit_(status):
    '''
    Exit with a specific status; same as sys.exit(status).

    :param status: status to be passed to sys.exit()
    '''
    sys.exit(status)


#############
# Utilities for subprocess
#############
def _call_fixup_stdin_arg(kwargs):
    '''
    Cleanup keyword arguments.

    :param kwargs: keyword arguments to be fixed
    :return: kwargs cleaned up, e.g., minus 'dump', replacing 'stdin'
    '''
    if 'dump' in kwargs:
        del kwargs['dump']
    if 'stdin' in kwargs and isinstance(kwargs.get('stdin'), (basestring, StringIO.StringIO)):
        input_ = kwargs.get('stdin')
        input_ = input_.getvalue() if isinstance(input_, StringIO.StringIO) else input_
        del kwargs['stdin']
        return kwargs, input_
    return kwargs, None


def call(cmd, **kwargs):
    '''
    Run a command.

    :param cmd: command to be run
    :param kwargs: keyword arguments for command
    :return: return code of the command
    '''

    _dump_commands(cmd, **kwargs)
    kwargs, input_ = _call_fixup_stdin_arg(kwargs)
    if input_:
        echo_debug('Running command using STDIN pipe...')
        process = subprocess.Popen(cmd, stdin=subprocess.PIPE, **kwargs)
        _, _ = process.communicate(input_)
        echo_debug('Return code [%d]' % process.returncode)
        return process.returncode
    else:
        return_code = subprocess.call(cmd, **kwargs)
        echo_debug('Return code [%d]' % return_code)
        return return_code


def check_call(cmd, **kwargs):
    '''
    Run a command with check.

    :param cmd: command to be run
    :param kwargs: keyword arguments for command
    :return: return code of the command
    '''

    _dump_commands(cmd, **kwargs)
    if 'stdin' in kwargs:
        return_code = call(cmd, **kwargs)
        assert return_code == 0
        return return_code
    else:
        return_code = subprocess.check_call(cmd, **kwargs)
        echo_debug('Return code [%d]' % return_code)
        return return_code


def check_output(cmd, **kwargs):
    '''
    Run a command and check output.  Note: result may include the last '\\n'.  Same as::

        subprocess.check_output

    :param cmd: command to be run
    :param kwargs: keyword arguments for command
    :return: output of command
    '''
    _dump_commands(cmd, **kwargs)
    kwargs, input_ = _call_fixup_stdin_arg(kwargs)
    if input_:
        process = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE, **kwargs)
        out, err = process.communicate(input_)
    else:
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs)
        out, err = process.communicate()
    echo_debug('tempOut [%s]' % out)
    echo_debug('tempErr [%s]' % err)
    echo_debug('return code [%d]' % process.returncode)
    assert process.returncode == 0
    return out


def check_output_strip(cmd, **kwargs):
    '''
    Run a command and return the output without leading and trailing spaces.

    :param cmd: command to be run
    :param kwargs: keyword arguments for command
    :return: trimmed output of command
    '''
    return check_output(cmd, **kwargs).strip()


def _dump_commands(cmd, **kwargs):
    '''
    Output command and arguments.

    :param cmd: command to output, if 'dump' one of the arguments
    :param kwargs: keyword arguments for command
    '''
    if 'dump' in kwargs:
        cmd = kwargs['dump']
    echo_debug('Calling command-line: <<<%s>>>' % str(cmd))
    echo_debug('kwargs: <<<%s>>>' % str(kwargs))


def su_call(minus_opt, usercreds, cmd, call_method=None, **kwargs):
    '''
    Run a command as super user (Linux) or Run As Administrator (Windows).

    su -c (a.k.a. RunAs in Windows)

    :param minus_opt: '-' for new login shell, '--', or None otherwise
    :param usercreds: new su user name in string (Linux) or tuple (user, password) (Windows)
    :param call_method: way for command to be called
    :param kwargs: keyword arguments
    :return: output of the call
    '''
    # verify the call method
    call_method = call_method or call
    assert call_method in [call, check_call, check_output, check_output_strip,
                           subprocess.call, subprocess.check_call, subprocess.check_output]

    username = usercreds[0] if isinstance(usercreds, (list, tuple)) else usercreds
    if is_linux():
        if username:
            actual_cmd = (['su'] + ([minus_opt] if minus_opt else []) +
                          [username, '-c', cmd if isinstance(cmd, basestring)
                           else subprocess.list2cmdline(cmd)])
        else:
            actual_cmd = cmd
    else:
        if usercreds:
            cwd = kwargs.get('cwd')
            kwargs, cmd_input = _call_fixup_stdin_arg(kwargs)
            command_return_type = CommandReturnType.Return_Code
            if 'check_output' in call_method.__name__:
                command_return_type = CommandReturnType.Output
            if call_method is check_output_strip:
                command_return_type = CommandReturnType.Stripped_Output

            env = kwargs.get('env')

            return_code = _win_su_call(usercreds, cmd, cwd=cwd, cmd_input=cmd_input,
                                       timeout_in_minutes=75,
                                       command_return_type=command_return_type, env=env)

            if return_code != 0 and 'check_call' in call_method.__name__:
                raise CalledProcessError(return_code, cmd)

            return return_code
        else:
            actual_cmd = cmd
    # invoke
    return call_method(actual_cmd, **kwargs)


##
# Base64 aliases
##
def encode_base64(string, altchars=None):
    '''
    Encode a string in base-64.  Shorter name for method call; instead of using alias.

    :param string: string to be encoded
    :param altchar: altchars to pass to underlying encode method
    :return: base-64 encoded string
    '''
    return base64.b64encode(string, altchars)    # encode_base64(s)


def decode_base64(string):
    '''
    Decode a base-64 string.

    :param string: string to be decoded
    :return: decoded string
    '''
    result = base64.b64decode(string)
    return result[:-1] if result[-1:] == '\n' else result  # remove the last \n gracefully


#############
# Env
#############
class QDEnv(object):
    '''
    This class provides access to env.properties and other properties files.

    This uses ConfigParser to parse Java properties file.
    See: http://stackoverflow.com/questions/2819696/parsing-properties-file-in-python/2819788
    '''
    @staticmethod
    def get_values_from_prop_file(filename, *props, **defaults):
        '''
        Get values from file.

            ex1: prop1 = env.get_values('file', 'prop1')
                Get 'prop1' property value from the env. Assertion error if missing.
            ex2: prop1, prop2 = env.get_values('file', 'prop1', 'prop2')
                Get tuple of property values simultaneously. Error if any one is missing.
            ex3: prop1 = env.get_values('file', 'prop1', prop1=<default_value>)
                Get 'prop1' value. If missing, the <default_value> is returned.

        :param filename: file to read value(s) from
        :param props: list of properties to read
        :param defaults: default values for properties
        :return: value of properties from file; if missing, default values
        '''
        env = QDEnv().read(filename)
        return env.get_values(*props, **defaults)

    @staticmethod
    def set_value_to_prop_file(filename, prop, value):
        '''
        Set <prop>=<value> pair into file.

        :param filename: value will be set into this file
        :param prop: property to be set
        :param value: value for property
        '''
        # escape \,=,: --> \\, \=, \:
        prop_e = re.sub(r'([\\=:])', r'\\\1', prop)
        # escape \,=,: --> \\, \=, \:
        value_e = re.sub(r'([\\=:])', r'\\\1', value) if value else value
        if os.path.exists(filename) and grep(filename, r'^%s=' % re.escape(prop_e)):
            # when the file contains the prop_e
            if value_e is None:
                # remove the line
                sed_in_place(filename, [(r'^%s=.*$' % re.escape(prop_e), '')])
            else:
                sed_in_place(filename, [(r'^%s=.*$' % re.escape(prop_e),
                                         ('%s=%s' % (prop_e, value_e)).replace('\\', '\\\\'))])
        else:
            # append new line
            with open(filename, 'at') as file_handle:
                file_handle.write('%s=%s%s' % (prop_e, value_e, '\n'))

    def __init__(self, filename=None, parent=None, error_if_not_exist=True):
        assert not parent or isinstance(parent, QDEnv)
        self.parent = parent
        self.data = collections.OrderedDict()
        if filename:
            if not os.path.isfile(filename) and not error_if_not_exist:
                self.filename = filename  # initialize even file not exists
            else:
                self.read(filename)

    def read(self, filename):
        '''
        Read a file, setting property/value contents into environment's collection.

        :param filename: name of file to be read
        '''
        # Use ConfigParser to load .properties file
        import ConfigParser
        parser = ConfigParser.ConfigParser()
        parser.optionxform = lambda x: x  # override x.lower() to make case-sensitive
        default_section = 'QD_SECTION'

        # pylint: disable=too-few-public-methods
        class FakeSectFh(object):
            '''
            Fake file handler.
            '''

            def __init__(self, file_handle, default_sect):
                self.file_handle = file_handle
                self.sec_head = '[%s]%s' % (default_sect, os.linesep)

            def readline(self):
                '''
                Read a line.

                :return: line read from file
                '''

                if self.sec_head:
                    try:
                        return self.sec_head
                    finally:
                        self.sec_head = None
                else:
                    return self.file_handle.readline()

        parser.readfp(FakeSectFh(open(filename), default_section))
        echo_debug('loaded [%s] and %d properties are defined.'
                   % (filename, len(parser.options(default_section))))
        self.filename = filename

        # fixup value escapes considering the followings
        # - remove double-quote escape  # for compatibility to shell-style VAR="VALUE"
        # - escape \\,\:,\= to \,:,=    # for .properties file
        for opt, value_e in parser.items(default_section):
            if value_e.startswith('"') and value_e.endswith('"'):
                value = re.sub(r'\\([\\:="])', r'\1', value_e[1:-1])  # remove double-quotes
            else:
                value = re.sub(r'\\([\\:=])', r'\1', value_e)
            self.data[opt] = value

    def has_value(self, prop):
        '''
        Check if environment has specific property.

        :param prop: property to check.
        :return: true if property exists in environment data; false otherwise
        '''
        return prop in self.data

    def set_value(self, prop, value):
        '''
        Set value of specific property in environment.

        :param prop: property to set
        :param value: value of property
        '''

        if value is None:
            if prop in self.data:
                del self.data[prop]
        else:
            if isinstance(value, list):
                value = ','.join(value)
            self.data[prop] = value

        if self.filename:
            echo_debug('%s (in %s) [%s]' % (prop, self.filename, value))
            self.set_value_to_prop_file(self.filename, prop, value)

    def _update_values(self, _prop_items):
        '''
        Update values of properties.

        :param _prop_items: list of properties to be updated
        '''

        # _prop_item: [prop_name, existence, value]
        def update_prop(_prop_item):
            '''
            Update specific property.

            :param: _prop_item: property to update
            '''
            if _prop_item[1]:
                return
            prop_name = _prop_item[0]

            # try to find the property with OS suffix
            prop_name_with_os_suffix = prop_name + os_switch('.linux', '.win32')
            if self.has_value(prop_name_with_os_suffix):
                _prop_item[1] = True
                _prop_item[2] = self.data[prop_name_with_os_suffix]

            # try to find using the base name
            elif self.has_value(prop_name):
                _prop_item[1] = True
                _prop_item[2] = self.data[prop_name]

        map(update_prop, _prop_items)  # apply get_prop() for each element of props
        # consider parent if this has parent and any one is not retrieved
        if self.parent and any(not item[1] for item in _prop_items):
            # Disable pylint protected-access so we can do what we need right now.
            # pylint: disable=protected-access
            self.parent._update_values(_prop_items)

    def get_values(self, *props, **defaults):
        '''
        Get values from this QDEnv instance.

            ex1: prop1 = env.get_values('prop1')
                Get 'prop1' property value from the env. Assertion error if missing.
            ex2: prop1, prop2 = env.get_values('prop1', 'prop2')
                Get tuple of property values simultaneously. Error if any one is missing.
            ex3: prop1 = env.get_values('prop1', prop1=<default_value>)
                Get 'prop1' value. If missing, the <default_value> is returned.

        :param props: properties to get
        :param defaults: defalut values of properties
        :return: values of properties; if missing, use defaults
        '''
        # prop_item: [prop_name, existence, value]
        items = [[prop, False, None] for prop in props]
        self._update_values(items)

        # assign default values
        for item in [i for i in items if not i[1] and i[0] in defaults]:
            # prop_item that have not assigned value and has defaults
            item[1] = True
            item[2] = defaults[item[0]]

        # for all v in retrieved_values, test if v[1] is positive.
        def get_assert_msg():
            '''
            Get assert message.

            :return: assert message
            '''
            # pylint: disable=deprecated-lambda
            msg = ' or '.join(map(lambda v: '$%s [%s]' % (v[0], v[2]), items))
            return '%s not set!' % msg

        assert all(v[1] for v in items), get_assert_msg()

        if len(items) == 1:
            return items[0][2]
        else:
            return tuple((v[2]) for v in items)

    def get_value_as_path(self, prop):
        ''' Get a single value from this QDEnv instance and normalize the result for the current OS.
            Assertion error if the property does not exist.  If the resulting value has no slashes,
            the value is returned unchanged.

        :param prop: Name of the property to read from this QDEnv.
        :return: The value of the property, normalized for this OS.  Assertion error if the property
                 does not exist.
        '''

        # Get the value from the properties file
        raw_value = self.get_values(prop)
        # Get rid of any backslashes
        flipped_raw_value = str(raw_value).replace("\\", "/")

        # Split the resulting string into fragments
        result = flipped_raw_value.split('/')

        # Call on existing code to join these fragments into an os-specific path
        return path_join(*result)

    def get_dict(self):
        '''
        Get environment's dictionary containing property/value information.

        :return: copy of dictionary
        '''
        return self.data.copy()


#############
# Test utilities
#############
def get_jazz_release_version(jazz_media_content):
    '''
    Pass in the $appMediaRoot (in env.properties) as argument and
    Return the jazz product release version

    :param jazz_media_content: the directory name of jazz product media location
    :return: release version
    '''
    app_release = jazz_media_content[-3:]
    if app_release.isdigit():
        return app_release
    else:
        exit_with_err_msg("appMediaRoot [%s] does not have last three digits"
                          % jazz_media_content)


def grep(source_file, pattern, ignore_case=False):
    '''
    Perform a unix-like grep on a source file, returning lines that match a pattern.

    :param source_file: file to be grepped
    :param pattern: pattern to look for
    :param ignore_case: true if case should be ignored; false otherwise
    :return: lines from the source_file that match the pattern
    '''
    source_file = os.path.normpath(source_file)
    pattern_compiled = re.compile(pattern, flags=(re.IGNORECASE if ignore_case else 0))
    with open(source_file) as file_handle:
        return [line for line in file_handle if pattern_compiled.findall(line)]


def escape_path(full_path):
    '''
    Converts a normal Python path to an escaped path.  This essentially means on Windows that the
    backslash is escaped.

    :param full_path: an absolute Python path
    :return: escaped path
    '''
    if is_windows():
        full_path = full_path.replace("\\", "\\\\")

    return full_path


def sed_in_place(source_file, replacement_list):
    '''
    Perform a unix-like sed -i on a source file, modifying that file and returning the number of
    lines modified.

    :param source_file: file to be modified
    :param replacement_list: list of replacements in tuple (pattern, replacement)
    :return: number of replaced lines
    '''
    source_file = os.path.normpath(source_file)
    replaced = 0
    with tempfile.NamedTemporaryFile(mode='w+t', delete=False) as tmp_sources:
        with open(source_file) as source_file:
            for line in source_file:
                line_replaced = False
                for replacement in replacement_list:
                    if re.findall(replacement[0], line):
                        line_r = re.sub(replacement[0], replacement[1], line)
                        replaced += 1
                        line_replaced = True
                        line = line_r
                if not line_replaced or line:
                    tmp_sources.write(line)
    if not replaced:
        pass
    else:
        if is_linux():
            shutil.copymode(getattr(source_file, 'name'), getattr(tmp_sources, 'name'))
            file_st = os.stat(getattr(source_file, 'name'))

            user = getattr(file_st, 'st_uid')
            group = getattr(file_st, 'st_gid')
            chown(tmp_sources.name, user, group)

        move_file(getattr(tmp_sources, 'name'), getattr(source_file, 'name'))
    return replaced


def find_a_file(searching_dir, file_name):
    '''
    Search a file path from top down.

    :param searching_dir: the start search point of the directory
    :param file_name: The file name to be searched
    :return The file path name if find, otherwise None
    '''
    for root, _, files in os.walk(searching_dir):
        for name in files:
            if name == file_name:
                return path_join(root, name)
    return None


def find_a_file_ext(searching_dir, ext):
    '''
    Find a file with specific extension

    :param searching_dir: the start search point of the directory
    :param ext: The file name extension to be searched, such as '.lic'
    :return: return the absolute file path if found, otherwise none
    '''
    for root, _, files in os.walk(searching_dir):
        for name in files:
            if name.endswith(ext):
                return path_join(root, name)
    return None


#############
# FUNCTIONS
#############
def append_content_to_file(file_to_append, additional_lines):
    '''
    Append contents to end of the file

    :param file_to_append: file to be appended
    :param additional_lines: list of all lines
    '''
    echo_info('Appending lines to file [%s]' % file_to_append)
    with open(file_to_append, 'a') as file_:
        # add a new line to ensure the cursor is on a new line
        file_.write('\n')
        for line in additional_lines:
            file_.write(line + '\n')
    echo_info('Completed appending lines to file [%s]' % file_to_append)


def print_file_skip_lines(file_to_print, skip=""):
    '''
    Log the contents of a file, skipping any lines with the 'skip' string.

    :param file_to_print: file to output to standard output
    :param skip: (optional) lines containing this string will not be printed. If skip is empty,
                 no lines will be skipped.
    '''
    echo_debug("Contents of [%s] file:" % file_to_print)
    echo_debug("----------------------------------------------------------")
    file_ = open(file_to_print, 'r')
    for line in file_:
        if not skip or skip not in line:
            echo_debug(line.rstrip('\r\n'))  # trim the line endings
        else:
            echo_debug("--- skipping line containing '%s' ---" % skip)
    file_.close()
    echo_debug("----------------------------------------------------------")


# parse arguments
def process_inputs(argv, input_list_src, num_optional_args=0, output=True):
    '''
    Process inputs sent into python call.

    :param logger: Logger to be used for output
    :param argv: List of arguments passed into python call
    :param input_list_src: List of descriptive argument names; no relation to variables in code
    :param integer num_optional_args: Number of arguments that can be optional (e.g.,
                                      if all args are optional, none need to be passed in).
                                      Default is 0.
    :param boolean output: ``True`` if the name:value pairs should be output from here; ``False``
                             otherwise.  Recommend calling this function with ``False`` when there
                             are optional arguments and outputting those values in the calling
                             function.  Default is ``True``.
    '''
    input_list = csv_split(input_list_src)
    # NUMBER_OF_ARGS
    if len(argv) - 1 < len(input_list) - num_optional_args or len(input_list) < len(argv) - 1:
        echo_error('Inputs should be %s' % ''.join('[%s]' % i for i in input_list))
        echo_error('NOT [%d] in [%d:%d]' % (len(argv), len(input_list) - num_optional_args,
                                            len(input_list)))
        exit_(1)

    # get arg values
    result = []
    for i, input_name in enumerate(input_list):
        value = argv[i + 1] if i + 1 < len(argv) else None
        result.append(value)
        # Only output the name:value pairs if output turned on.  With optional agruments,
        # name:value pairs can be useless.
        if output:
            echo_info('%s: %s' % (input_name, value))

    return tuple(result)


def get_host_name():
    '''
    Get hostname; throw error if it contains uppercase characters.

    :return: hostname
    '''
    import socket
    hostname = socket.getfqdn().rstrip()
    echo_debug('tempHostname [%s]' % hostname)
    if any(c.isupper() for c in hostname):
        echo_error("hostname [" + hostname + "] contains uppercase characters")
        exit_(1)
    else:
        echo_info("hostname [" + hostname + "] DOES NOT contains uppercase characters")
    return hostname


# usage check_boolean bFlag
# -- accepts true, false, TRUE, FALSE (and a mismatch of case!)
def check_boolean(flag, var_name=None):
    '''
    Check incoming string and return appropriate boolean value; throws exception if string cannot
    be parsed into boolean.

    :param flag: string value, e.g., [true, TRUE, false, FALSE]
    :param var_name: variable name

    :return: Returns boolean based on incoming arguments.
    '''
    return parse_boolean(flag, var_name or 'boolean value')


# usage copy_file fileSource targetDir
def copy_file(file_, target_dir):
    '''
    Copy file to target directory.

    :param file_: file to be copied
    :param target_dir: directory in which to put copied file
    :return: copied file in target directory; None if copy failed
    '''
    dest_file = path_join(target_dir, basename(file_))
    echo_info("copy [%s] to [%s]" % (file_, dest_file))
    shutil.copy(file_, dest_file)
    echo_info("copy complete")
    return dest_file


def duplicate_file(source_file_path, destination_file_name):
    '''
    Make a copy of a file within the same directory
    :param source_file_path: The absoluate source file path
    :param destination_file_name: The new name of the copied file
    :return:
        * if successful - the full path to the copied file
        * if not - the script will exit with error
    '''
    dest_file_path = path_join(dirname(source_file_path), destination_file_name)
    echo_info("copy [%s] to [%s]" % (source_file_path, dest_file_path))
    shutil.copy(source_file_path, dest_file_path)
    check_file_exists(dest_file_path)
    echo_info("copy complete")
    return dest_file_path


def copy_file_with_stat(filename, target_dir):
    '''
    Copies the file {filename} to folder {target_dir}

    :param filename: name of file to copy
        (if no path part will assume {filename} in the current folder)
    :param target_dir: folder to copy the file into
    :return:
        * if successful - the full path to the copied file
        * if not - the script will exit with error
    '''
    dest_file = path_join(target_dir, basename(filename))
    _copy_file_with_stat(filename, dest_file)
    return dest_file


def copy_file_new_name(filename, target_dir, new_filename):
    '''
    Copies the file {filename} to folder {target_dir}

    :param filename: name of file to copy
        (if no path part will assume {filename} in the current folder)
    :param target_dir: folder to copy the file into
    :param new_filename: the new name of the copied file (no path content)
    :return:
        * if successful - the full path to the copied file
        * if not - the script will exit with error
    '''
    dest_file = path_join(target_dir, new_filename)
    _copy_file_with_stat(filename, dest_file)
    return dest_file


def _copy_file_with_stat(source_filepath, target_filepath):
    '''
    Copies the file {source_filepath} to the {target_filepath}

    Similar to the Unix command cp -p, will attempt to copy owner, group and permission info.
    Script will exit with error if the copy is unsuccessful.

    :param source_filepath: name of file to copy
        (if no path part will assume {target_filepath} in the current folder)
    :param target_filepath: name of the copied file
        (if no path part will assume {target_filepath} in the current folder)
    '''
    check_file_exists(source_filepath)
    echo_info("copy [%s] to [%s]" % (source_filepath, target_filepath))
    # attempt to copy owner, group and permission info
    shutil.copy2(source_filepath, target_filepath)
    # just copy
    # shutil.copy(source_filepath, target_filepath)
    check_file_exists(target_filepath)
    echo_info("copy complete")


def copy_folder_with_stat(source_dir, target_dir):
    '''
    Copies the contents of folder {source_dir} to folder {target_dir}.
     {target_dir} must not already exist; it will be created as well as missing parent directories

    Similar to the Unix command cp -rp, will attempt to copy owner, group and permission info.

    :param source_dir: name of folder whose contents will be copied
        (if no path part will assume {source_dir} is in the current folder)
    :param target_dir: folder to copy the file into (cannot already exist)
    :return:
        * if successful - the full path to the copied file
        * if not - the script will exit with error
    '''
    check_folder_exists(source_dir)
    echo_info("copy contents of folder [%s] to [%s]" % (source_dir, target_dir))
    shutil.copytree(source_dir, target_dir)
    check_folder_exists(target_dir)
    echo_info("copy complete")
    return target_dir


def create_file(file_to_create):
    '''
    Create file with r,w,x permissions; exit with error if file creation fails.

    :param file_to_create: file to be created
    '''
    echo_info("creating [" + file_to_create + "]")
    open(file_to_create, 'w').close()
    if os.path.exists(file_to_create):
        echo_info("file exists")
        chmod_777(file_to_create, mode='u+x')
    else:
        echo_error("File does not exist")
        exit_(1)


# usage delete_file fileSource
def delete_file(file_to_delete):
    '''
    Delete file by calling on OS functionality.  Exit with error if the file does not exist;
    exit with error if the file not deleted.

    :param file_to_delete: file to be deleted.
    '''
    echo_info("delete [%s]" % file_to_delete)
    check_file_exists(file_to_delete)

    os.remove(file_to_delete)
    echo_info('deleting...')
    if os.path.exists(file_to_delete):
        exit_with_err_msg('File still exists')
    echo_info('File deleted')


# Function delete_file_if_exists(fileToDelete)
def delete_file_if_exists(file_to_delete):
    '''
    Delete file if it exists; otherwise, do nothing.

    :param file_to_delete: file to be deleted
    '''
    echo_info("deleting [" + file_to_delete + "] if it exists")
    if os.path.exists(file_to_delete):
        delete_file(file_to_delete)
    else:
        echo_info("file does not exist, so no need to delete it")


def delete_folder(foldername):
    '''
    Delete folder; exit with error if deletion fails.  Exit with error if folder does not initially
    exist.

    :param foldername: folder to delete (if no path, assume {foldername} is in the current folder)
    '''
    echo_info("delete folder [%s]" % foldername)
    check_folder_exists(foldername)

    shutil.rmtree(foldername)
    echo_info('deleting...')
    if os.path.exists(foldername):
        exit_with_err_msg('Folder still exists')
    echo_info('Folder deleted')


def delete_folder_if_exists(foldername):
    '''
    Delete folder if it exists; otherwise do nothing.

    :param foldername: folder to delete (if no path, assume {foldername} is in the current folder)
    '''
    echo_info("deleting folder [%s] if it exists" % foldername)
    if os.path.exists(foldername):
        delete_folder(foldername)
    else:
        echo_info("folder does not exist, so no need to delete it")


# usage move_file fileSource fileTarget
def move_file(source, target):
    '''
    Move file by calling on OS functionality.
    If source and target are in the same folder this is the equivalent of renaming the file.

    Exit with error if the file does not exist prior to move.
    Exit with error if the file is not moved to the target.

    :param source: full path to thefile to be moved
    :param target: full path to the moved file
    '''
    echo_info('move [%s] to [%s]' % (source, target))
    check_file_exists(source)

    shutil.move(source, target)
    echo_info('moving...')
    check_file_exists(target)


# usage move_folder fileSource fileTarget
def move_folder(source, target):
    '''
    Move folder by calling on OS functionality.
    If source and target are in the same folder this is the equivalent of renaming the folder.

    Exit with error if the folder does not exist prior to move.
    Exit with error if the folder is not moved to the target.

    :param source: full path to the folder to be moved
    :param target: full path to the moved folder
    '''
    echo_info('move [%s] to [%s]' % (source, target))
    check_folder_exists(source)

    shutil.move(source, target)
    echo_info('moving...')
    check_folder_exists(target)


# usage check_and_create_folder tempfolder
def check_and_create_folder(folder):
    '''
    - If [folder] exists do nothing
    - If [folder] does not exist, create the [folder]

    :param folder: the folder name including the path
    '''
    echo_info('check [%s] exists' % folder)
    if os.path.isdir(folder):
        echo_info('folder exists')
    else:
        echo_info("folder does not exist, create it...")
        os.makedirs(folder)
        if os.path.isdir(folder):
            echo_info('folder created')
        else:
            exit_with_err_msg("failed to create folder")


def check_and_create_folder_777(folder, mode='777'):
    '''
    Expands on qd.check_and_create_folder() in qd.py

    - If [folder] exists do nothing
    - If [folder] does not exist, create the [folder]
        - (Linux) attempt to set the [folder] permissions to [mode]
            - if [mode] not specified then make it wide open [777]
        - (Windows) only supports mode = '777'

    :param folder: the folder name including the path
    :param mode: (optional) (default is 777) the new folder's permissions
    '''
    echo_info("check [%s] exists" % folder)
    if os.path.isdir(folder):
        echo_info("folder exists")
    else:
        echo_info("folder does not exist, create it...")
        os.makedirs(folder)
        if os.path.isdir(folder):
            echo_info("folder created")
            chmod_777(folder, mode=mode)
        else:
            exit_with_err_msg("failed to create folder")


# usage check_file_exists tempFile
def check_file_exists(path):
    '''
    Check if a file exists.  Logs success but does not return value.

    :param str path: path of the file to check
    :return: nothing returned, but success is logged and failure results in hard exit
    '''
    assert path, 'path must not be empty'
    if os.path.isfile(path):
        echo_info("File [%s] located" % path)
    else:
        echo_error("File [%s] does not exist" % path)
        exit_(1)


# usage check_folder_exists tempFile
def check_folder_exists(path):
    '''
    Check if a folder exists.  Logs success but does not return value.

    :param str path: path of the folder to check
    :return: nothing returned, but success is logged and failure results in hard exit
    '''

    assert path, 'path must not be empty'
    if os.path.isdir(path):
        echo_info("Folder [%s] located" % path)
    else:
        echo_error("Folder [%s] does not exist" % path)
        exit_(1)


def zipup_dir(source_dir, zip_file):
    '''
    Zip the contents of the {source_dir} into the {zip_file},
    the zip structure will be relative to the {source_dir} and will include any child folders

    :param source_dir: folder to zip up
    :param zip_file: full path to the zip file created
    '''

    echo_info("Construct zip [%s] from the contents of folder [%s]" % (zip_file, source_dir))
    ziph = zipfile.ZipFile(zip_file, 'w', zipfile.ZIP_DEFLATED)
    f_source_path = path_join(source_dir)

    # for root, dirs, files in os.walk(source_dir):
    for content in os.walk(source_dir):
        root = content[0]
        files = content[2]
        for f_temp in files:
            f_temp_path = path_join(root, f_temp)  # exact path to file being added
            f_rel_path = os.path.relpath(f_temp_path, f_source_path)  # relative path to the file
            ziph.write(f_temp_path, f_rel_path)

    ziph.close()
    list_zip(zip_file)


def list_zip(zip_file):
    '''
    Returns a list of all the files in the zip file.

    :param zip_file: name of file to list
        (if no path part will assume it is in the current folder)

    :return: list of files in the {zip_file}
    '''

    echo_info("listing contents of zip file [%s]" % zip_file)
    if not is_zipfile(zip_file):
        exit_with_err_msg("file [%s] is not a valid zip file" % zip_file)
    zipped_file = zipfile.ZipFile(zip_file)
    file_contents = zipped_file.namelist()
    list_length = len(file_contents)
    echo_info("file contains [%s] entries" % list_length)

    # For debugging, output the list of all files in the zipped file
    if LOG_DEFAULT_LEVEL == logging.DEBUG:
        for enum_count, f_temp in enumerate(file_contents):
            echo_debug("%s of %s. [%s]" % (enum_count + 1, list_length, f_temp))

    return file_contents


def list_tar(tar_file):
    '''
    Returns a list of all the files in the tar file.

    :param tar_file: name of file to list
        (if no path part will assume it is in the current folder)

    :return: list of files in the {tar_file}
    '''

    echo_info("listing contents of tar file [%s]" % tar_file)
    if not is_tarfile(tar_file):
        exit_with_err_msg("file [%s] is not a valid tar file" % tar_file)
    tared_file = tarfile.open(tar_file)
    file_contents = tared_file.getnames()
    list_length = len(file_contents)
    echo_info("file contains [%s] entries" % list_length)

    # For debugging, output the list of all files in the zipped file
    if LOG_DEFAULT_LEVEL == logging.DEBUG:
        for enum_count, f_temp in enumerate(file_contents):
            echo_debug("%s of %s. [%s]" % (enum_count + 1, list_length, f_temp))

    return file_contents


def is_zipfile(zip_file):
    '''
    Wraps zipfile.is_zipfile(zip_file) so that it exits if the file does not exist
    Note: if {zip_file} is actually a jar file it will also return {True}

    :param zip_file: name of file to check
        (if no path part will assume it is in the current folder)

    :return:
        * True - {zip_file} exists and is a valid zip file
        * False - {zip_file} exists but is NOT a valid zip file
    '''

    check_file_exists(zip_file)
    return zipfile.is_zipfile(zip_file)


def is_tarfile(tar_file):
    '''
    Wraps tarfile.is_tarfile(tar_file) so that it exits if the file does not exist

    :param tar_file: name of file to check
        (if no path part will assume it is in the current folder)

    :return:
        * True - {tar_file} exists and is a valid zip file
        * False - {tar_file} exists but is NOT a valid zip file
    '''

    check_file_exists(tar_file)
    return tarfile.is_tarfile(tar_file)


def unzip_zip(zip_file, target_dir, section_reference=None):
    '''
    Extracts a file to a specified directory.  Works for .zip, .tar, and .gz files.

    :param zip_file: compressed/archive file to be expanded
    :param target_dir: location to write contents of zip_file
    :param section_reference: pattern for files to unzip (does not work for .tar or .gz files)
    '''

    def extract_zip():
        '''
        Extracts content from a zip archive, preserving the original date & time information.
        If required partial extraction can be configured using the {section_reference} parameter
        '''

        dest = os.path.abspath(target_dir)
        echo_info("extracting zip file [%s] to [%s]..." % (zip_file, dest))
        list_length = len(list_zip(zip_file))

        dirs = {}
        with zipfile.ZipFile(zip_file) as zipped_file:
            for enum_count, zip_info in enumerate(zipped_file.infolist()):
                filename = zip_info.filename
                # get original date and time for later
                orig_date_time = time.mktime(zip_info.date_time + (0, 0, -1))
                if section_reference:
                    if not glob.fnmatch.filter([filename], section_reference):
                        continue
                # validate the file is in the dest
                norm_path = os.path.abspath(os.path.join(dest, filename))
                assert norm_path.startswith(dest)
                echo_debug("extracting %s of %s. [%s]..." % (enum_count + 1, list_length, filename))
                zipped_file.extract(zip_info, dest)

                # now update with the original time from zip
                if os.path.isdir(norm_path):
                    # as files are being updated inside any folders its best to update folders later
                    # otherwise changes here will be overwritten
                    dirs[norm_path] = orig_date_time
                else:
                    os.utime(norm_path, (orig_date_time, orig_date_time))

        # all files extracted and updated, can now do the folders
        for foldername in dirs:
            orig_date_time = dirs[foldername]
            os.utime(foldername, (orig_date_time, orig_date_time))

        echo_info("unzip successful")

    def extract_tar():
        '''
        Extracts content from a tar archive (.gz or .tar) file,
        including those using gzip or bz2 compression

        .. warning::
             does not yet do any partial extraction using the {section_reference} parameter

        '''
        dest = os.path.abspath(target_dir)
        echo_info("extracting tar file [%s] to [%s]..." % (zip_file, dest))
        list_length = len(list_tar(zip_file))

        with tarfile.open(zip_file) as tared_file:
            for enum_count, tar_info in enumerate(tared_file.getmembers()):
                norm_path = os.path.abspath(os.path.join(dest, tar_info.name))
                assert norm_path.startswith(dest)
                echo_debug("extracting %s of %s. [%s]..."
                           % (enum_count + 1, list_length, tar_info.name))
                tared_file.extract(tar_info, dest)
        echo_info("untar successful")

    if is_zipfile(zip_file):
        extract_zip()
    elif is_tarfile(zip_file):
        extract_tar()
    else:
        exit_with_err_msg("Unknown archive file type [%s]" % zip_file)


def list_files_in_folder(foldername):
    '''
    Lists all the files contained in the folder {foldername}

    :param foldername: name of folder to list
        (if no path part will assume {foldername} is in the current folder)

    :return: list of files in folder {foldername}
    '''

    file_list = [f_name for f_name in os.listdir(foldername)
                 if os.path.isfile(path_join(foldername, f_name))]
    echo_info("Contents of folder [%s] [%s]" % (foldername, file_list))

    return file_list


def exit_with_warning_msg(msg):
    '''
    Exit with a warning.

    :param msg: the warning message to display
    '''
    echo_warning("-- message: %s" % msg)
    exit_(0)


def exit_with_err_msg(msg):
    '''
    Exit with an error.

    :param msg: the error message to display
    '''
    echo_error("-- Error in [{scriptname}] --")
    echo_error("-- message: %s" % msg)
    exit_(1)


def create_service(service_file_name, template_file_dir=None):
    '''
    Create service on the target machine

    :param service_file_name: the service file that include start and stop script
    :param template_file_dir: the template service file directory path
    '''
    echo_info("The script of the service is called [%s]" % service_file_name)
    if is_linux():
        if template_file_dir:
            os_id = platform.linux_distribution()[2]
            if os_id == 'Santiago':
                target_dir = '/etc/init.d'
                echo_info("Setup service [%s] on redhat 6" % service_file_name)
                copy_file(path_join(template_file_dir, service_file_name), target_dir)
                check_call(['chkconfig', '--add', service_file_name])
                check_file_exists(path_join(target_dir, service_file_name))
                chmod_777(path_join(target_dir, service_file_name), 'uo+x')
                check_call(['chkconfig', '--list', service_file_name])
            elif os_id == 'Maipo':
                target_dir = '/usr/lib/systemd'
                service_name = service_file_name + '.service'
                echo_info("Setup service [%s] on redhat 7" % service_name)
                check_and_create_folder(path_join(target_dir, 'scripts'))
                copy_file(path_join(template_file_dir, service_file_name),
                          path_join(target_dir, 'scripts'))
                copy_file(path_join(template_file_dir, service_name),
                          path_join(target_dir, 'system'))
                chmod_777(path_join(target_dir, 'scripts', service_file_name), 'uo+x')
                check_call(['systemctl', 'enable', service_name])
                check_file_exists(path_join(target_dir, 'scripts', service_file_name))
                check_file_exists(path_join('/etc', 'systemd', 'system',
                                            'multi-user.target.wants', service_name))
            else:
                exit_with_err_msg("ERROR: unsupported platform [%s]" % os_id)
        else:
            exit_with_err_msg("ERROR: service template file directory should be set")


####################################################
# Windows version of SU implementation
####################################################

class CommandReturnType(object):
    # pylint: disable=too-few-public-methods,no-init
    ''' This class acts as an enum for valid return types from the su_call_window function. '''
    Return_Code, Output, Stripped_Output = range(3)


# pylint: disable=too-many-branches,too-many-locals,too-many-statements,too-many-arguments
def _win_su_call(usercreds, cmd, cwd=None, cmd_input=None, timeout_in_minutes=1,
                 command_return_type=CommandReturnType.Return_Code, env=None):
    # pylint: disable=line-too-long,redefined-variable-type
    '''
    Run the cmd as the usercreds using Task Scheduler (TS).  By default the function returns the
    exit code for the command, but if the command_return_type is set to a different value it may
    return the output or stripped output.

    :param usercreds: tuple containing (user, password)
    :param cmd: command string or list of command args
    :param cwd: (optional) directory to use as current working directory
    :param cmd_input: (optional) input to be directed to the command
    :param timeout_in_minutes: (optional) time to wait for task to complete--default is one minute
    :param command_return_type: see the CommandReturnType class for choices
    :param env: (optional) dictionary of environment vars to be set

    .. note:: The user must has permission to the working directory (from 'cwd' in kwargs, \
        otherwise os.getpwd()) otherwise, this invocation may fail due to no permission to "CD" \
        to the dir.  Also, everything referred form the cmd must be accessible from the su user.

    Overall, we:
        1. Create a batch file (cmd_file) containing the command-line built from the cmd
           - The batch-file does: CD to the workdir, run the cmd, then write the %ERRORLEVEL% to \
               a file (rc_file).
        2. Register a Task to TS to invoke the cmd_file as the given user
        3. Run the Task and wait for it to complete
        4. Read the return code and output from the files
        5. Delete the task and files created
        6. Return the exit code or output from the command
    '''
    assert is_windows(), 'The _win_su_call() function only works on Windows operating systems'

    assert isinstance(usercreds, tuple) and len(usercreds) == 2, 'su_win requires tuple(user, password) as the usercreds.'
    os_user_name, os_user_password = usercreds

    if cwd:
        assert os.path.exists(cwd), 'The directory "%s" does not exist' % cwd
    else:
        cwd = os.getcwd()

    # Set default values for variables read in finally block
    user_added_to_admin_group = False
    command_line = None
    return_code = 'NOT_SET'

    # Create unique name for task
    taskname = 'IBMQD_su_' + str(uuid.uuid4())

    # prepare tmp file names
    cmd_file = tempfile.mktemp('.cmd', 'qd_su_call_cmd_')
    stdin_file = tempfile.mktemp('.tmp', 'qd_su_call_in_')
    stdout_file = tempfile.mktemp('.tmp', 'qd_su_call_out_')
    rc_file = tempfile.mktemp('.tmp', 'qd_su_call_rc_')
    xml_file = os.path.join(tempfile.gettempdir(), taskname + '.xml')

    try:
        # if necessary, add the user to the Administrators group
        if subprocess.call('net localgroup "%s" | findstr "%s"'
                           % ('Administrators', usercreds[0]), shell=True) != 0:
            echo_debug('Adding user [%s] to admin group for scheduled tasks...'
                       % os_user_name)
            subprocess.check_call(['net', 'localgroup', 'Administrators', os_user_name, '/add'])
            user_added_to_admin_group = True
        else:
            echo_debug('User [%s] is in the Administrators group' % os_user_name)

        # create files
        with open(rc_file, 'wt'):
            pass

        with open(stdin_file, 'wt') as file_handle:
            file_handle.write(cmd_input or "")

        # Create command batch file
        with open(cmd_file, 'wt') as command:
            command.write('rem === scheduled batch file ===' + os.linesep)

            # Set env vars if necessary
            if env:
                for env_key, env_val in env.items():
                    command.write('set %s=%s%s' % (env_key, env_val, os.linesep))

            command.write(('cd /d "%s"' % cwd) + os.linesep)
            command_line = (cmd if isinstance(cmd, basestring) else subprocess.list2cmdline(cmd))
            # if kwargs.get('shell'):
            if ('.bat' in command_line or '.cmd' in command_line) and 'cmd /c' not in command_line:
                command_line = subprocess.list2cmdline(['cmd', '/c', command_line])
            command.write('%s < "%s" > "%s" 2>&1%s' % (command_line, stdin_file, stdout_file, os.linesep))
            command.write('echo %%ERRORLEVEL%% > "%s"%s' % (rc_file, os.linesep))

        echo_info("Preparing to run command:\n" + command_line)

        # create and start the task (the task is expected to be running in background)
        echo_info('Creating TS task [%s]...' % taskname)
        create_task_xml(cmd_file, xml_file, os_user_name, command_line)
        subprocess.check_call(['schtasks', '/create', '/tn', taskname, '/xml', xml_file,
                               '/ru', os_user_name, '/rp', os_user_password, '/f'])

        echo_info('Starting TS task [%s]...' % taskname)
        subprocess.check_call(['schtasks', '/run', '/tn', taskname])

        start_time = datetime.datetime.utcnow()
        timeout_time = start_time + datetime.timedelta(minutes=timeout_in_minutes)
        idle_time = 1

        # wait until the return code has been published in the file
        while os.path.exists(rc_file) and os.path.getsize(rc_file) == 0:
            time.sleep(idle_time)
            idle_time += 1
            if idle_time > 5:
                idle_time = 5

            # Check for timeout and break from loop if timeout is exceeded
            if datetime.datetime.utcnow() > timeout_time:
                # timeout reached (can be happen in PyDev)
                echo_error('su_call_win timeout reached. Terminating the process...')
                if subprocess.call(['schtasks', '/end', '/tn', taskname]) != 0:
                    echo_error('Error reported on terminating the process.')
                break

        # get command return code
        with open(rc_file) as file_handle:
            return_code = file_handle.read()
        return_code = return_code.strip()
        assert return_code, 'The return code must be set. The program may be aborted. Check the <Systemroot>\\SchedLgU.txt.'
        return_code = int(return_code)

        if command_return_type in [CommandReturnType.Output, CommandReturnType.Stripped_Output]:
            with open(stdout_file) as file_handle:
                content = file_handle.read()
                if command_return_type == CommandReturnType.Stripped_Output:
                    return content.strip()
                else:
                    return content
        else:
            return return_code

    finally:
        if command_line:
            echo_info("Command:\n" + command_line)

        # Print command output
        if os.path.exists(stdout_file):
            with open(stdout_file) as file_handle:
                content = file_handle.read()
                echo_info('Command output:\n' + content)

        if return_code != 'NOT_SET':
            echo_info('Command exit code:\t' + str(return_code))

        # Remove user from Administrators group if necessary
        if user_added_to_admin_group:
            echo_debug('Removing user [%s] from Administrators group...' % os_user_name)
            try:
                subprocess.check_call(['net', 'localgroup', 'Administrators', os_user_name, '/delete'])
            except CalledProcessError as cpe:
                echo_error('Error removing %s from Administrators group.  %s' % (os_user_name, cpe))

        # Delete Task Scheduler task
        if taskname:
            echo_debug('Deleting the TS task %s' % taskname)
            try:
                subprocess.check_call(['schtasks', '/delete', '/tn', taskname, '/F'])
            except CalledProcessError as cpe:
                echo_error('Error deleting task %s.  %s' % (taskname, cpe))

        # Remove temporary files
        for tmp_file in (cmd_file, stdin_file, stdout_file, rc_file, xml_file):
            if os.path.exists(tmp_file):
                echo_debug('Removing file %s' % tmp_file)
                os.unlink(tmp_file)


def create_task_xml(cmd_file, xml_file, os_user_name, description=''):
    '''
    Creates a Windows Task Scheduler task xml file.
    :param cmd_file: file to be executed (typically a .cmd file)
    :param xml_file: file to be created--the file name should be the same as the task name
    :param os_user_name: the user that should run the task
    :param description: human readable description of what the task does
    '''
    escaped_description = escape(description)
    template = Template(TASK_TEMPLATE)
    with open(xml_file, 'wb') as tmp_file:
        tmp_file.write(codecs.BOM_UTF16_LE)
        tmp_file.write(template.substitute(os_user_name=os_user_name, cmd_file=cmd_file,
                                           description=escaped_description).encode('utf-16-le'))


TASK_TEMPLATE = """<?xml version="1.0" encoding="UTF-16"?>
<Task version="1.2" xmlns="http://schemas.microsoft.com/windows/2004/02/mit/task">
  <RegistrationInfo>
    <Date>2017-03-13T17:05:08.4493662</Date>
    <Author>Quick Deployer</Author>
    <Description>$description</Description>
  </RegistrationInfo>
  <Triggers>
    <TimeTrigger>
      <StartBoundary>2017-03-17T13:37:34</StartBoundary>
      <Enabled>true</Enabled>
    </TimeTrigger>
  </Triggers>
  <Principals>
    <Principal id="Author">
      <UserId>$os_user_name</UserId>
      <LogonType>Password</LogonType>
      <RunLevel>HighestAvailable</RunLevel>
    </Principal>
  </Principals>
  <Settings>
    <MultipleInstancesPolicy>IgnoreNew</MultipleInstancesPolicy>
    <DisallowStartIfOnBatteries>true</DisallowStartIfOnBatteries>
    <StopIfGoingOnBatteries>true</StopIfGoingOnBatteries>
    <AllowHardTerminate>true</AllowHardTerminate>
    <StartWhenAvailable>false</StartWhenAvailable>
    <RunOnlyIfNetworkAvailable>false</RunOnlyIfNetworkAvailable>
    <AllowStartOnDemand>true</AllowStartOnDemand>
    <Enabled>true</Enabled>
    <Hidden>false</Hidden>
    <RunOnlyIfIdle>false</RunOnlyIfIdle>
    <WakeToRun>false</WakeToRun>
    <ExecutionTimeLimit>P3D</ExecutionTimeLimit>
    <Priority>4</Priority>
  </Settings>
  <Actions Context="Author">
    <Exec>
      <Command>$cmd_file</Command>
    </Exec>
  </Actions>
</Task>
"""


def get_memory_size_on_linux():
    '''
    Get the total memory size on Linux platform.

    :return: memory size in MB
    '''
    meminfo = open('/proc/meminfo').read()
    matched = re.search(r'^MemTotal:\s+(\d+)', meminfo)
    mem_total_kb = 0
    if matched:
        mem_total_kb = int(matched.groups()[0])
        # return the memory size in MB
        return mem_total_kb / 1000
    else:
        exit_with_err_msg("Having trouble to get the memory size")


def get_memory_size_on_windows():
    '''
    Get the total memory size on Windows platform.

    :return: memory size in MB
    '''
    process = os.popen('wmic memorychip get capacity')
    result = process.read()
    process.close()
    mem_total_b = 0
    for mem in result.split("  \r\n")[1:-1]:
        mem_total_b += int(mem)
    if mem_total_b == 0:
        exit_with_err_msg("Failed to get the total memory size on windows,"
                          " please make sure wmic utility is enabled")
    # return the memory size in MB
    return mem_total_b / (1000**2)


def get_heapsize_replacement():
    '''
    Determine appropriate heap size arguments, based on total physical memory size.

    :return: list of heapsize arguments (Heapsize, -Xmx, -Xms, etc.), based on memory size
    '''
    if is_linux():
        mem_total = get_memory_size_on_linux()
    elif is_windows():
        mem_total = get_memory_size_on_windows()
    else:
        exit_with_err_msg("Unsupported platform")
    if mem_total < 4000:
        echo_info("The system RAM size is less than 4GB")
        replacement_list = [('HeapSize=4096', 'HeapSize=' + str(mem_total)),
                            ('-Xmx4g -Xms4g', '-Xmx2g -Xms2g')]
    elif mem_total < 6000:
        echo_info("The system RAM size is 4gb, do nothing")
        replacement_list = []
    elif mem_total < 7500:
        echo_info("The system RAM size is less than 7GB")
        replacement_list = [('HeapSize=4096', 'HeapSize=' + str(mem_total)),
                            ('-Xmx4g -Xms4g', '-Xmx6g -Xms6g'), ('-Xmn512m', '-Xmn1g')]
    else:
        echo_info("The system RAM size more than 8gb, allocate 75% of total memory $RAMSize")
        allocate_mem = int(mem_total * 0.75)
        heapsize = allocate_mem / 1000
        echo_info("Allocate memory is [%s]" % allocate_mem)
        replacement_list = [('HeapSize=4096', 'HeapSize=' + str(allocate_mem)),
                            ('-Xmx4g -Xms4g',
                             '-Xmx' + str(heapsize) + 'g' + ' -Xms' + str(heapsize) + 'g'),
                            ('-Xmn512m', '-Xmn1g')]
    return replacement_list
