#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Provides access to properties files used in the Quick Deployer process.

- Define commonly-used property names
- Get env.properties, userCreds.properties, media.properties, Custom_LDAP.properties
- Get os (linux?) user names
'''
import os

from qdlib import qd, resource

UCD_PN_QD_VERSION = 'QD_VERSION'  #:
''' Property name for Quick Deployer scripts version'''
QD_VERSION = '2.1.0'  #:
''' Quick Deployer scripts version'''


UCD_PN_GLOBAL_TAG_LISTING = "globalTagListing"  #:
''' UCD environment global tag listing '''


PN_AUTO_PREP = 'AUTO_PREP'  #:
''' Property name for env.properties.  In env, by CoreAutomationPrep.sh. '''
PN_WAS_COMMON_DIR = 'WAS_COMMON_DIR'  #:
''' Property name for env.properties. '''
PN_QD_ROOT = 'QD_Root'  #:
''' Property name for the top QD parent directory '''
PN_DB_APP_LIST = 'databaseList'  #:
''' property name for all applications that need database '''
PN_APPS = 'APPS'  #:
''' Property name for APPlications to be installed '''
PN_CUSTOM_NODES = 'NUMBER_OF_CUSTOM_NODES'  #:
''' Property name for the total number of custom nodes'''
PN_LDAP_PROPS = 'LDAPPROP'  #:
''' Property name for prefix of LDAP properties file '''
PN_INTERNAL_MEDIA_PROPS = 'QD_InternalMedia'  #:
''' Internal media property name '''
PN_IM_SHARED_DIR_PREP = 'IM_SHARED_DIR'  #:
''' Property name for env.properties, contains location of the IM shared directory '''
PN_IM_INST_DIR_PREP = 'IM_INST_DIR'  #:
''' Property name for env.properties, contains location of the IM InstallationManager directory '''
PN_IM_ECLIPSE_DIR_PREP = 'IM_ECLIPSE_DIR'  #:
''' Property name for env.properties, contains location of the IM eclipse directory '''
PN_CLM_HOME_PREP = 'CLM_HOME'  #:
''' Property name for env.properties, contains location of "JazzTeamserver". '''
PN_WAS_HOME_PREP = 'WAS_HOME'  #:
''' Property name for env.properties, contains location of "WebSphere". '''
PN_IHS_HOME_PREP = 'IHS_HOME'  #:
''' Property name for env.properties, contains location of "IHS". '''
PN_PLUGINS_HOME_PREP = 'PLUGINS_HOME'  #:
''' Property name for env.properties, contains location of "IHS Plugins", '''
PN_DB2_ADMIN_USER = 'ibmQseInst_User'  #:
''' Property name of the DB2 admin user, stored in userCreds.properties'''
PN_DB2_ADMIN_PWD = 'ibmQseInst_PW'  #:
''' Property name for the DB2 admin user password, stored in userCreds.properties'''
PN_ORACLE_OS_USER = 'oracleOSUserName'  #:
''' Property name for the oracle os user, stored in userCreds.properties'''
PN_ORACLE_OS_PWD = 'oracleOSUserPassword'  #:
''' Property name for the oracle os user password, stored in userCreds.properties'''
PN_ORACLE_DIR_PREP = 'ORACLE_DIR'  #:
''' Property name for env.properties, contains location of "ORACLE". '''
PN_ORACLE_INSTANCE = 'oracleInstanceName'  #:
''' Property name for oracle instance name in oracleParameter.properties'''
PN_ORACLE_PORT = 'oraclePort'  #:
''' Property name for oracle port'''
PN_CREATE_MEDIA_ONLY = 'createMediaOnly'  #:
''' Property name for 'create media only' '''
PN_DEBUG_DO_NOT_INSTALL = 'debugDoNotInstall'  #:
''' Property name for 'do not install debug' '''
PN_SQLSERVER_INSTANCE = 'sqlserverInstanceName'  #:
''' Property name for SQL instance name in sqlserverParameter.properties'''
PN_SQLSERVER_COLLATION = 'sqlserverCollation'  #:
''' Property name for SQL collation in sqlserverParameter.properties'''
PN_SQLSERVER_PORT = 'sqlserverPort'  #:
''' Property name for SQL port in sqlserverParameter.properties'''
PN_SQLSERVER_SA_PASS = 'sqlSAPassword'  #:
''' Property name for SA password in in sqlserverParameter.properties'''
PN_SQL_DB_USER_PASS = 'sqlDBUsersPassword'  #:
''' Property name for DB user password in in sqlserverParameter.properties'''

PN_USER_OS_USER_NAME = 'OSUserName'  #:
''' Property name for OS user name '''
PN_USER_OS_USER_PASSWORD = 'OSUserPassword'  #:
''' Property name for OS user password '''
PN_USER_OS_USER_GROUP = 'OSUserGroup'  #:
''' Property name for OS user group '''
PN_USER_OS_USER_PRIMARY_GROUP = 'OSUserPrimaryGroup'  #:
''' Property name for OS user primary group '''

PN_USER_WAS_USER_NAME = 'WAS_USER_NAME'  #:
''' Property name for WAS Admin user name '''
PN_USER_WAS_USER_PASSWORD = 'WAS_PASSWORD'  #:
''' Property name for WAS Admin user password '''
PN_IHS_KEYSTORE_PASSWORD = 'IHSServerKeyStorePassword'  #:
''' Property name for IHS KEYSTORE_PASSWORD '''
PN_IHS_HOSTNAME = 'ihshost'  #:
''' Property name for IHS hostname'''
PN_LOCAL_APP_SERVER = 'appServer'  #:
''' App server name (liberty, was, notUsed) for this local instance '''
PN_LOCAL_DB_VENDOR = 'localDB_VENDOR'  #:
''' DB vendor for local apps, which overrides global DB_VENDOR' '''
PN_DB_VENDOR = 'DB_VENDOR'  #:
''' property name of the db vendor, this is from env.properties file'''
PN_DB_DRIVER_FILE = 'dbDriverFile'  #:
''' database driver file name, depends on the database version, the driver name may different'''

PN_USER_REGISTRY = 'userRegistryType'  #:
''' App server name (liberty, was, notUsed) for this local instance '''

PN_LIBERTY_DIR = 'LIBERTY_CONF_DIR'  #:
''' Property name for env.properties. '''

# env.properties
DEFAULT_ENV_PROPS_LOCATION = os.path.join(os.environ.get('QD_ENV_DIR') or
                                          qd.os_switch('/etc', r'C:\etc'), 'env.properties')  #:
''' Default location of environment properties.

.. warning:: THIS VALUE MUST BE SYNCED WITH UCD PROCESS DEFINITIONS
'''

DEFAULT_INST_DIR = qd.os_switch('/opt/IBM', r'C:\IBM')  #:
''' Default installation locations.  (Could go into the properties file in the future)

.. note::
   using default locations based on DEFAULT_INST_DIR are discouraged,
   better to use RELATIVE LOCATIONS based on 'QD_ROOT'

'''

# CLM_CACHE TODO not sure /opt/IBM/IMShared is correct
DEFAULT_IM_SHARED = qd.path_join(DEFAULT_INST_DIR, 'IMShared')  #:
''' Default Installation Manager shared location '''
DEFAULT_REL_IM_SHARED = 'IMShared'  #:
''' Default Installation manager relative shared location '''
DEFAULT_IM_INST_DIR = qd.path_join(DEFAULT_INST_DIR, 'InstallationManager')  #:
''' Default Installation Manager installation location '''
DEFAULT_REL_IM_INST_DIR = ('InstallationManager')  #:
''' Default Installation Manager relative installation location '''
# CAUTION: BE CAREFUL TO DISTINGUISH DEFAULT_IM_INST_DIR and DEFAULT_IM_ECLIPSE_DIR
DEFAULT_IM_ECLIPSE_DIR = qd.path_join(DEFAULT_IM_INST_DIR, 'eclipse')  #:
''' Default Installation Manager Eclipse location.

   .. warning:: Be careful to distinguish between DEFAULT_IM_INST_DIR and DEFAULT_IM_ECLIPSE_DIR '''
DEFAULT_REL_IM_ECLIPSE_DIR = qd.path_join(DEFAULT_REL_IM_INST_DIR, 'eclipse')  #:
''' Default Installation Manager relative eclipse location '''

DEFAULT_CLM_HOME = qd.path_join(DEFAULT_INST_DIR, 'JazzTeamServer')  #:
''' Default CLM/JTS location '''
DEFAULT_REL_CLM_HOME = 'JazzTeamServer'  #:
''' Default CLM/JTS relative location '''

DEFAULT_WAS_HOME = qd.path_join(DEFAULT_INST_DIR, 'WebSphere')  #:
''' Default WAS/WebSphere location '''
DEFAULT_REL_WAS_HOME = 'WebSphere'  #:
''' Default WAS/WebSphere relative location '''
DEFAULT_REL_WAS_ROOT = qd.path_join(DEFAULT_REL_WAS_HOME, 'AppServer')   #:
''' Default WAS/WebSphere/AppServer relative location '''
DEFAULT_REL_WAS_PROFILE_NAME = 'AppSrv01'  #:
''' Default profile name'''
DEFAULT_REL_WAS_PROFILE = qd.path_join(DEFAULT_REL_WAS_ROOT, 'profiles',
                                       DEFAULT_REL_WAS_PROFILE_NAME)  #:
''' Default WAS/WebSphere/AppServer/profiles/AppSrv01 relative location '''

DEFAULT_IHS_HOME = qd.path_join(DEFAULT_INST_DIR, 'HTTPServer')  #:
''' Default IHS/HTTPServer location '''
DEFAULT_REL_IHS_HOME = 'HTTPServer'  #:
''' Default IHS/HTTPServer relative location '''
DEFAULT_PLUGINS_HOME = qd.path_join(DEFAULT_INST_DIR, 'Plugins')  #:
''' Default plugins location '''
DEFAULT_REL_PLUGINS_HOME = 'Plugins'  #:
''' Default plugins relative location '''

DEFAULT_REL_TMP_DIR = 'qd_tmp'  #:
''' Temporary directory name '''
DEFAULT_LDAP_REL_PATH = 'ConfigureWAS/LDAP/LDAPSecurity.properties'  #:
''' LDAPSecurity.properties default location within the resources directory '''
DEFAULT_USER_CREDS_REL_PATH = 'AutomationPrep/UserParameters/userCreds.properties'  #:
''' userCreds.properties default location within the resources directory '''
DEFAULT_WSADMIN_EXE = qd.os_switch('wsadmin.sh', 'wsadmin.bat')  #:
''' Default wsadmin executable '''
DEFAULT_DW_FOLDER = 'DWFolder'  #:
''' Default data warehouse folder name '''

DEFAULT_REL_DB2FS_DIR = 'db2fs'
DEFAULT_REL_ORACLE_STORAGE_DIR = 'dbstorage'  #:
''' Default Oracle location '''
DEFAULT_REL_SQLSERVER_STORAGE_DIR = 'sqlserverinst'
''' Default SQL Instance location '''

# clients
DEFAULT_RTCC_HOME = qd.path_join(DEFAULT_INST_DIR, 'TeamConcert')  #:
''' Default RTCC/TeamConcert client location '''
DEFAULT_RDMIE_HOME = qd.path_join(DEFAULT_INST_DIR, 'RDMImportEngine')  #:
''' Default RDMImportEngine client location '''
DEFAULT_RPY_HOME = qd.path_join(DEFAULT_INST_DIR, 'Rhapsody')  #:
''' Default Rhapsody client location '''
DEFAULT_RLKS_HOME = qd.path_join(DEFAULT_INST_DIR, 'RLKS')  #:
''' Default RLKS (license key server) client location '''


##
# Get properties files
##
def get_default_env_props():
    '''
    Get instance of env.properties (aka /etc/env.properties)

    :return: env.properties instance
    '''
    return qd.QDEnv(DEFAULT_ENV_PROPS_LOCATION)


def get_user_props(env_props):
    '''
    Get user properties

    :param env_props: environment properties
    :return: userCreds.properties instance
    '''
    user_creds_file = resource.get_resource(DEFAULT_USER_CREDS_REL_PATH)
    user_env = qd.QDEnv(user_creds_file, env_props)
    return user_env


def get_media_props(env_props, error_if_not_exist=True):
    '''
    Get media properties

    :param env_props: environment properties
    :param error_if_not_exist: determine if an error should raised when the file does not exist
    :return: media.properties instance
    '''
    try:
        media_properties = \
            resource.get_resource('AutomationPrep/MediaParameters/media.properties')
    except IOError:
        media_properties = \
            resource.create_resource_path('AutomationPrep/MediaParameters/media.properties')
    media_env = qd.QDEnv(media_properties, env_props,
                         error_if_not_exist=error_if_not_exist)
    return media_env


def get_database_props(env_props):
    '''
    Get (db2, oracle, sqlserver) database properties; None for other database types.

    :param env_props: environment properties
    :return: DB2 or Oracle database properties; returns None for any other database
    '''
    supported_vendors = ['db2', 'oracle', 'sqlserver']
    db_vendor = env_props.get_values('DB_VENDOR')
    if db_vendor in supported_vendors:
        db_properties = resource.get_resource(
            'AutomationPrep/DatabaseParameters/%sParameter.properties' % db_vendor)

        db_env = qd.QDEnv(db_properties)
        qd.echo_info("database vendor [%s], properties file  [%s]"
                     % (db_vendor, (db_env.filename)))
    else:
        db_env = None
        qd.echo_info("database vendor [%s] does not require a properties file" % db_vendor)
    return db_env


def get_custom_ldap_props(env_props):
    '''
    Get custom LDAP properties

    :param env_props: environment properties
    :return: Custom_LDAPSecurity.properties instance
    '''
    ldap_props_file = \
        resource.get_resource('AutomationPrep/CustomLDAP/Custom_LDAPSecurity.properties')
    ldap_env = qd.QDEnv(ldap_props_file, env_props)
    return ldap_env


def get_ldap_props(env_props, was_common_dir=None):
    '''
    Get WAS LDAP properties

    :param env_props: environment properties
    :param was_common_dir: The directory path of the Websphere configuration scripts
    :return: LDAPSecurity.properties instance
    '''
    was_common_dir = was_common_dir or env_props.get_values(PN_WAS_COMMON_DIR)
    ldap_env = qd.QDEnv(resource.get_resource(DEFAULT_LDAP_REL_PATH), env_props)
    return ldap_env


def copy_ldap_property_file():
    '''
    Determine the correct LDAP property file to be used and copy it into the default location
    :return The generic LDAP file path
    '''
    env_props = get_default_env_props()
    default_absolute_ldap_file = resource.create_resource_path(DEFAULT_LDAP_REL_PATH)

    default_ldap_dir = qd.dirname(default_absolute_ldap_file)
    default_ldap_file_name = qd.basename(default_absolute_ldap_file)
    specific_ldap_file_name = env_props.get_values(PN_LDAP_PROPS) + '_LDAPSecurity.properties'

    # Copy CustomLDAP property file to LDAP folder
    if env_props.get_values(PN_LDAP_PROPS) == 'CustomLDAP':
        qd.check_and_create_folder(default_ldap_dir)
        source_ldap_template = resource.get_resource(
            'AutomationPrep/CustomLDAP/Custom_LDAPSecurity.properties')
        qd.copy_file_new_name(source_ldap_template, default_ldap_dir, specific_ldap_file_name)

    # Copy the specific LDAP template file to the default LDAP property file name
    specific_ldap_file_path = qd.path_join(default_ldap_dir, specific_ldap_file_name)
    qd.check_file_exists(specific_ldap_file_path)
    qd.duplicate_file(specific_ldap_file_path, default_ldap_file_name)
    qd.check_file_exists(default_absolute_ldap_file)
    return default_absolute_ldap_file


##
# Get properties values
##
def load_create_media_only(media_props):
    '''
     Get the create_media_only property value from media.properties

    :param media_props: media properties
    :return: create_media_only property (True or False)
    '''

    # validate execution options
    create_media_only = media_props.get_values(PN_CREATE_MEDIA_ONLY,
                                               **{PN_CREATE_MEDIA_ONLY: False})
    create_media_only = qd.parse_boolean(create_media_only, PN_CREATE_MEDIA_ONLY)
    if create_media_only:
        qd.echo_info("Adding application installer to media server")
    else:
        qd.echo_info("Installing applications")
    return create_media_only


def load_debug_do_not_install(media_props):
    '''
     Get the debug_do_not_install property value from media.properties

    :param media_props: media properties
    :return: debug_do_not_install property (True or False)
    '''

    debug_do_not_install = media_props.get_values(PN_DEBUG_DO_NOT_INSTALL,
                                                  **{PN_DEBUG_DO_NOT_INSTALL: False})
    debug_do_not_install = qd.parse_boolean(debug_do_not_install, PN_DEBUG_DO_NOT_INSTALL,
                                            none_value=False)
    if debug_do_not_install:
        qd.echo_debug("Will exit prior to installing applications")
    else:
        qd.echo_info("Installing applications")
    return debug_do_not_install


def get_os_users(user_props):
    '''
    .. warning::
       ONLY INCLUDED FOR BACKWARDS COMPATABILITY
       should really use get_new_os_users_creds_and_grp()

    Returns the name, password (decoded) and group for the OSUser
    as defined in userCreds.properties file

    :param user_props: user properties
    :return: tuple (os_user_name, os_user_password, os_user_group):
        * **os_user_name**: the name of the new user to be added
        * **os_user_password**: the new user's password (decoded)
        * **os_user_group**: the group the new user must belong to
    '''
    os_user_name, os_user_password, os_user_group = get_new_os_users_creds_and_grp(user_props)
    return os_user_name, os_user_password, os_user_group


def get_new_os_users_creds_and_grp(user_props):
    '''
    Returns the name, password (decoded) and group for the OSUser
    as defined in userCreds.properties file

    :param user_props: user properties
    :return: tuple (os_user_name, os_user_password, os_user_group):
        * **os_user_name**: the name of the new user to be added
        * **os_user_password**: the new user's password (decoded)
        * **os_user_group**: the group the new user must belong to
    '''
    os_user_name, os_user_password = get_os_users_creds(user_props)
    os_user_group = user_props.get_values(PN_USER_OS_USER_GROUP)
    qd.echo_info("os_user_group: [%s]" % os_user_group)

    return os_user_name, os_user_password, os_user_group


def get_os_users_creds_and_group(user_props):
    '''
    Returns the name, password (decoded) and primary group for the OSUser as held in {user_props}

    :param user_props: user properties
    :return: tuple (os_user_name, os_user_password, os_user_primary_group):
        * **os_user_name**: the name of the user
        * **os_user_password**: the user's password (decoded)
        * **os_user_primary_group**: the primary group the user belongs to
    '''
    os_user_name, os_user_password = get_os_users_creds(user_props)
    os_user_primary_group = user_props.get_values(PN_USER_OS_USER_PRIMARY_GROUP)
    qd.echo_info("os_user_group: [%s]" % os_user_primary_group)

    return os_user_name, os_user_password, os_user_primary_group


def get_os_users_creds(user_props):
    '''
    Returns the name, password (decoded) for the OSUser as held in {user_props}

    :param user_props: user properties
    :return: tuple (os_user_name, os_user_password):
        * **os_user_name**: the name of the user to be added
        * **os_user_password**: the user's password (decoded)
    '''
    os_user_name, os_user_password = user_props.get_values(PN_USER_OS_USER_NAME,
                                                           PN_USER_OS_USER_PASSWORD)
    os_user_password = qd.decode_base64(os_user_password)

    qd.echo_info("os_user_name: [%s]" % os_user_name)
    qd.echo_info("os_user_password: [%s]" % ('*' * len(os_user_password)))

    return os_user_name, os_user_password


def get_os_users_name_and_group(user_props):
    '''
    Returns the name and primary group for the OSUser as held in {user_props}

    :param user_props: user properties
    :return: tuple (os_user_name, os_user_primary_group):
        * **os_user_name**: the name of the user
        * **os_user_primary_group**: the primary group the user belongs to
    '''
    os_user_name, os_user_primary_group = user_props.get_values(PN_USER_OS_USER_NAME,
                                                                PN_USER_OS_USER_PRIMARY_GROUP)
    qd.echo_info("os_user_name: [%s]" % os_user_name)
    qd.echo_info("os_user_group: [%s]" % os_user_primary_group)

    return os_user_name, os_user_primary_group


def set_new_os_users_primary_group(user_props, value):
    '''
    Sets the OSUserPrimaryGroup {user_props} and updates userCreds.properties file

    :param user_props: user properties
    :param value: new value for OSUserPrimaryGroup
    '''
    user_props.set_value(PN_USER_OS_USER_PRIMARY_GROUP, value)


def get_was_admin_users(user_props):
    '''
    Get the WAS administrator user name and password, return decrypted password

    :param user_props: user properties
    :return: tuple (was_user_name, was_user_password):
        * **was_user_name**: WAS user name
        * **was_user_password**: decrypted WAS password
    '''
    was_user_name, was_user_password = user_props.get_values(PN_USER_WAS_USER_NAME,
                                                             PN_USER_WAS_USER_PASSWORD)
    was_user_password = qd.decode_base64(was_user_password)

    qd.echo_info("os_user_name: [%s]" % was_user_name)
    qd.echo_info("os_user_password: [%s]" % ('*' * len(was_user_password)))

    return was_user_name, was_user_password


def check_app_server(is_ihs=False):
    '''
    Verify if the script should run on the server
    Will run if it is a Liberty or WAS server and just some cases in IHS server
    It's intended to remove the use of AppServer TAG

    :param is_ihs: If true the script can run on IHS server
    :return: Return True if need to run on the server else return False
    '''
    env_props = get_default_env_props()
    try:
        if (env_props.get_values('WASRole') == 'was' or
                env_props.get_values('WASRole') == 'liberty'):

            # Added for evaluation topology support
            # If no IHS server it means that it's a evaluation topology
            env_props.get_values('ihshost')
            return True
        elif is_ihs:
            if env_props.get_values('WASRole') == 'ihs':
                return True
            else:
                qd.echo_info("No WAS or Liberty on this server")
                return False
        else:
            qd.echo_info("No WAS or Liberty on this server")
            return False
    except AssertionError:
        qd.echo_info("No WAS or Liberty on this server")
        return False


def append_to_env_prop(properties_file_to_append):
    '''
    Appends the specified file to the env.properties file on the system.

    :param properties_file_to_append: full path to file to be appended
    '''
    env_props = get_default_env_props().filename

    with open(env_props, 'at') as prop_file:
        prop_file.write('\n')  # Add line feed in case file doesn't end with one
        prop_file.write(open(properties_file_to_append, 'r').read())
