# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Provides functions to simplify finding non-Python resources.
'''
import os
import errno
import qdlib.qd as qd

QD_RESOURCES_DIRECTORY_ENV_VAR = "QD_RESOURCES_DIR"

DEFAULT_RESOURCES_DIR_NAME = 'resources'


def get_resources_directory():
    '''
    Returns the path to the resources directory.  This is usually the QuickDeployer/resources
    directory, but the value can be overridden by setting the QD_RESOURCES_DIR environment
    variable.  This function should not be used to find resources (use get_resource() for that,
    but this function may be used to find a location where resources may be written.

    :return: resources directory path
    '''
    resources_dir = os.environ.get(QD_RESOURCES_DIRECTORY_ENV_VAR, None)

    if not resources_dir:
        script_dir = os.path.dirname(os.path.abspath(__file__))
        resources_dir = os.path.join(script_dir, '..', '..', DEFAULT_RESOURCES_DIR_NAME)
        resources_dir = os.path.abspath(resources_dir)

    if not os.path.isdir(resources_dir):
        raise IOError(errno.ENOENT, os.strerror(errno.ENOENT), resources_dir)

    return resources_dir


def get_resource(resource_path):
    '''
    Returns the full path to the specified resource.  Paths should use forward slashes regardless
    of operating system.  If the path contains windows/linux directories, the os-specific
    directory should be omitted and the function will automatically return the correct resource.

    :param resource_path: relative path to resource (using forward slashes)
    :return: full path to resource
    '''
    full_path = os.path.abspath(os.path.join(get_resources_directory(), resource_path))

    parent_path = os.path.abspath(os.path.join(full_path, os.pardir))
    file_name = os.path.basename(full_path)
    os_dir_name = "windows" if qd.is_windows() else "linux"

    os_full_path = os.path.abspath(os.path.join(parent_path, os_dir_name, file_name))

    if os.path.isfile(os_full_path):
        return os_full_path

    if not os.path.exists(full_path):
        raise IOError(errno.ENOENT, os.strerror(errno.ENOENT), full_path)

    return full_path


def get_escaped_resource(resource_path):
    '''
    Returns the escaped full path to the specified resource.  The escaped path is useful when the
    path is passed as an argument via command line.  See get_resource for details for how the
    resource is located..

    :param resource_path: relative path to resource (using forward slashes)
    :return: escaped full path to resource
    '''
    full_path = get_resource(resource_path)
    return qd.escape_path(full_path)


def get_relative_path(full_path):
    '''
    Converts the full path of a resource into a path relative to the resources directory.  The
    relative path returned will always use forward slashes regardless of operating system.  The
    full path must include a resource directory or the function will fail.

    :param full_path: an absolute Python path that includes the resources directory
    :return: path relative to resources directory
    '''
    working_path = full_path.replace('\\', '/')

    if DEFAULT_RESOURCES_DIR_NAME + '/' in working_path:
        res_pos = working_path.find(DEFAULT_RESOURCES_DIR_NAME + '/')
        working_path = working_path[res_pos + len(DEFAULT_RESOURCES_DIR_NAME) + 1:]
        return working_path

    resources_dir_name = os.path.basename(get_resources_directory())
    if resources_dir_name + '/' in working_path:
        res_pos = working_path.find(resources_dir_name + '/')
        working_path = working_path[res_pos + len(resources_dir_name) + 1:]
        return working_path

    raise AssertionError("Unable to find resource directory path in %s" % full_path)


def create_resource_path(resource_path):
    '''
    Creates the path to a location in the resources directory for the specified resource path.
    Paths should use forward slashes regardless of operating system.  If the path already exists,
    this will throw an IOError.

    :param resource_path: relative path to resource (using forward slashes)
    '''
    full_path = os.path.abspath(os.path.join(get_resources_directory(), resource_path))
    if os.path.exists(full_path):
        raise IOError('Unable to create resource path %s.  It already exists.' % full_path)
    return full_path
